/*
 * Access a PWM server to update a display panel backlight.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/cxx/ipc_server>
#include <l4/re/env>
#include <l4/re/util/object_registry>
#include <l4/sys/capability>

#include <stdlib.h>

#include <l4/devices/pwm-client.h>
#include "backlight-ops.h"

/* PWM access abstractions. */

static L4::Cap<Pwm_device_interface> pwm_device;



/* Backlight device. */

class Backlight_device_server : public L4::Server_object_t<L4::Kobject>
{
  int _min, _max;

public:
  explicit Backlight_device_server(int min, int max)
  : _min(min), _max(max)
  {
  }

  /* Dispatch incoming requests. */

  int dispatch(l4_umword_t obj, L4::Ipc::Iostream &ios)
  {
    l4_msgtag_t tag;
    int arg;

    (void) obj;
    ios >> tag;

    switch (tag.label())
    {
      case Backlight_op_disable:
        disable();
        return L4_EOK;

      case Backlight_op_enable:
        enable();
        return L4_EOK;

      case Backlight_op_set_brightness:
        ios >> arg;
        set_brightness(arg);
        return L4_EOK;

      default:
        return -L4_EBADPROTO;
    }
  }

  void disable()
  {
    pwm_device->disable();
  }

  void enable()
  {
    pwm_device->enable();
  }

  /* Use the PWM device to update the backlight level. */

  void set_brightness(int level)
  {
    pwm_device->set_duty(level);
  }
};

static L4Re::Util::Registry_server<> server;



/* Arguments: <minimum level> <maximum level> */

int main(int argc, char *argv[])
{
  int min, max;

  if (argc < 3)
    return 1;

  /* Obtain a reference to the PWM device. */

  pwm_device = L4Re::Env::env()->get_cap<Pwm_device_interface>("pwm");
  if (!pwm_device.is_valid()) return 1;

  /* Initialise and register a new server object. */

  min = atoi(argv[1]);
  max = atoi(argv[2]);

  Backlight_device_server server_obj(min, max);
  server.registry()->register_obj(&server_obj, "backlight");

  /* Enter the IPC server loop. */

  server.loop();
  return 0;
}
