/*
 * Export Letux 400 display operations as a server.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4730.h>
#include <l4/devices/backlight-client.h>
#include <l4/devices/memory.h>

#include <l4/cxx/ipc_server>
#include <l4/re/env>
#include <l4/re/util/object_registry>
#include <l4/sys/capability>

#include "display-ops.h"

/* Virtual address for the GPIO register block. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;

/* GPIO and backlight device abstractions. */

static Gpio_jz4730_chip *gpio_port_b = 0, *gpio_port_c = 0;
static L4::Cap<Backlight_device_interface> backlight_device;



/* GPIO pin definitions. */

enum Jz4730_lcd_gpio_mask
{
  Jz4730_lcd_gpio_func1 = 0xf8ffff00,
  Jz4730_lcd_gpio_func2 = 0x07000000,
};

enum Jz4730_lcd_gpio
{
  Jz4730_lcd_gpio_led_enable     = 28,
  Jz4730_lcd_gpio_display_enable = 29,
};



static int setup_memory(void)
{
  if (get_memory("jz4730-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  return 0;
}



/* Display device. */

class Display_device_server : public L4::Server_object_t<L4::Kobject>
{
  Pin_slice lcd_mask1 = {.offset=0, .mask=Jz4730_lcd_gpio_func1};
  Pin_slice lcd_mask2 = {.offset=0, .mask=Jz4730_lcd_gpio_func2};

public:
  explicit Display_device_server()
  {
    /* Set functions for the LCD pins. */

    gpio_port_b->multi_config_pad(lcd_mask1, Hw::Gpio_chip::Function_alt, 1);
    gpio_port_b->multi_config_pad(lcd_mask2, Hw::Gpio_chip::Function_alt, 2);

    /* Enable LED, disable display. */

    gpio_port_c->setup(Jz4730_lcd_gpio_led_enable, Hw::Gpio_chip::Output, 1);
    gpio_port_c->setup(Jz4730_lcd_gpio_display_enable, Hw::Gpio_chip::Output, 0);
  }

  /* Dispatch incoming requests. */

  int dispatch(l4_umword_t obj, L4::Ipc::Iostream &ios)
  {
    l4_msgtag_t tag;

    (void) obj;
    ios >> tag;

    switch (tag.label())
    {
      case Display_op_disable:
        disable();
        return L4_EOK;

      case Display_op_enable:
        enable();
        return L4_EOK;

      default:
        return -L4_EBADPROTO;
    }
  }

  /* Switch the display off. */

  void disable(void)
  {
    gpio_port_c->setup(Jz4730_lcd_gpio_display_enable, Hw::Gpio_chip::Output, 0);
    backlight_device->disable();
  }

  /* Switch the display on. */

  void enable(void)
  {
    gpio_port_c->setup(Jz4730_lcd_gpio_display_enable, Hw::Gpio_chip::Output, 1);
    backlight_device->enable();
  }
};

static L4Re::Util::Registry_server<> server;



int main(void)
{
  if (setup_memory()) return 1;

  /* Initialise the GPIO abstractions. */

  Gpio_jz4730_chip gpb(gpio_virt_base + (1 * 0x30), gpio_virt_base + (2 * 0x30), 32);
  Gpio_jz4730_chip gpc(gpio_virt_base + (2 * 0x30), gpio_virt_base + (3 * 0x30), 32);

  gpio_port_b = &gpb;
  gpio_port_c = &gpc;

  /* Obtain a reference to the backlight device. */

  backlight_device = L4Re::Env::env()->get_cap<Backlight_device_interface>("backlight");
  if (!backlight_device.is_valid()) return 1;

  /* Initialise and register a new server object. */

  Display_device_server server_obj;
  server.registry()->register_obj(&server_obj, "display");

  /* Enter the IPC server loop. */

  server.loop();
  return 0;
}
