/*
 * Export Ben NanoNote display operations as a server.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4740.h>
#include <l4/devices/backlight-client.h>
#include <l4/devices/lcd-jz4740-config.h>
#include <l4/devices/lcd-jz4740-panel.h>
#include <l4/devices/memory.h>

#include <l4/cxx/ipc_server>
#include <l4/re/env>
#include <l4/re/util/object_registry>
#include <l4/sys/capability>

#include "display-ops.h"

/* Virtual address for the GPIO register block. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;

/* GPIO and backlight device abstractions. */

static Gpio_jz4740_chip *gpio_port_c = 0;
static L4::Cap<Backlight_device_interface> backlight_device;



/* GPIO pin definitions. */

enum Jz4740_lcd_gpio_mask
{
  Jz4740_lcd_gpio_d0_d7 = 0xff,
};

enum Jz4740_lcd_gpio
{
  Jz4740_lcd_gpio_clk = 18, /* pixel clock */
  Jz4740_lcd_gpio_rs  = 19, /* hsync; command/data select */
  Jz4740_lcd_gpio_cs  = 20, /* vsync; chip select */
};



static int setup_memory(void)
{
  if (get_memory("jz4740-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  return 0;
}



/* Display device. */

class Display_device_server : public L4::Server_object_t<L4::Kobject>
{
  Pin_slice slcd8_mask = {.offset=0, .mask=(1 << Jz4740_lcd_gpio_cs) | (1 << Jz4740_lcd_gpio_rs) |
                                           (1 << Jz4740_lcd_gpio_clk) | Jz4740_lcd_gpio_d0_d7};

public:
  /* Dispatch incoming requests. */

  int dispatch(l4_umword_t obj, L4::Ipc::Iostream &ios)
  {
    l4_msgtag_t tag;

    (void) obj;
    ios >> tag;

    switch (tag.label())
    {
      case Display_op_disable:
        disable();
        return L4_EOK;

      case Display_op_enable:
        enable();
        return L4_EOK;

      default:
        return -L4_EBADPROTO;
    }
  }

  /* Switch the display off. */

  void disable(void)
  {
    /* Configure SLCD8 pins. */

    gpio_port_c->multi_setup(slcd8_mask, Hw::Gpio_chip::Input, 0);
    backlight_device->disable();
  }

  /* Switch the display on. */

  void enable(void)
  {
    /* Configure SLCD8 pins. */

    gpio_port_c->multi_config_pad(slcd8_mask, Hw::Gpio_chip::Function_alt, 0);
    backlight_device->enable();
  }
};

static L4Re::Util::Registry_server<> server;



int main(void)
{
  if (setup_memory()) return 1;

  /* Initialise the GPIO abstraction. */

  Gpio_jz4740_chip gpio_port(gpio_virt_base + 0x200, gpio_virt_base + 0x300, 32);

  gpio_port_c = &gpio_port;

  /* Obtain a reference to the backlight device. */

  backlight_device = L4Re::Env::env()->get_cap<Backlight_device_interface>("backlight");
  if (!backlight_device.is_valid()) return 1;

  /* Initialise and register a new server object. */

  Display_device_server server_obj;
  server.registry()->register_obj(&server_obj, "display");

  /* Enter the IPC server loop. */

  server.loop();
  return 0;
}
