/*
 * Export the keypad GPIOs on the Letux 400 as a data space accessible via the
 * "keypad" capability.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4730.h>
#include <l4/devices/dataspace.h>
#include <l4/devices/memory.h>
#include "keypad-server.h"
#include "keypad-event-loop.h"

#include <l4/re/dataspace>
#include <l4/re/env>
#include <l4/re/util/object_registry>

#include <l4/sys/cache.h>

#include <stdint.h>

enum Jz4730_keypad_gpio
{
  Jz4730_keypad_gpio_inputs_count = 8,
  Jz4730_keypad_gpio_outputs_count = 17,
};

/* Port A input pins. */

const uint8_t Jz4730_keypad_inputs[Jz4730_keypad_gpio_inputs_count] = {
  0, 1, 2, 3, 4, 5, 6, 7
};

const Pin_slice Jz4730_keypad_inputs_mask = {0x000000ff, 0};

/* Port D output pins. */

const uint8_t Jz4730_keypad_outputs[Jz4730_keypad_gpio_outputs_count] = {
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 29
};

const Pin_slice Jz4730_keypad_outputs_mask = {0x2000ffff, 0};

/* Peripheral memory regions. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;

/* GPIO abstractions. */

static void *gpio_port_a, *gpio_port_d;

/* Keypad status: an array of keypad column values. */

uint32_t *keypad = 0;

/* Imported keypad memory referenced by the array. */

void *keymem = 0;



/* Set up access to memory. */

static int setup_memory()
{
  if (get_memory("jz4730-gpio", &gpio_virt_base, &gpio_virt_base_end) < 0)
    return 1;

  gpio_port_a = jz4730_gpio_init(gpio_virt_base, gpio_virt_base + 0x30, 32);
  gpio_port_d = jz4730_gpio_init(gpio_virt_base + 0x90, gpio_virt_base + 0xc0, 32);

  return 0;
}



/* Initialise the pins for scanning the keypad. */

static void init_keyscan()
{
  jz4730_gpio_multi_setup(gpio_port_a, &Jz4730_keypad_inputs_mask, Hw::Gpio_chip::Input, 0);
  jz4730_gpio_multi_config_pull(gpio_port_a, &Jz4730_keypad_inputs_mask, Hw::Gpio_chip::Pull_up);
  jz4730_gpio_multi_setup(gpio_port_d, &Jz4730_keypad_outputs_mask, Hw::Gpio_chip::Input, 0);
}

/*
Scan the keypad by enabling each output column and inspecting each input row.
Store each column bitmap in the keypad array.
*/

static void scan_keypad()
{
  uint8_t column, row, value;

  for (column = 0; column < Jz4730_keypad_gpio_outputs_count; column++)
  {
    jz4730_gpio_setup(gpio_port_d, Jz4730_keypad_outputs[column], Hw::Gpio_chip::Output, 0);
    l4_sleep(1);

    value = 0;

    for (row = 0; row < Jz4730_keypad_gpio_inputs_count; row++)
      value = (value << 1) | (jz4730_gpio_get(gpio_port_a, Jz4730_keypad_inputs[row]) ? 0 : 1);

    keypad[column] = value;

    jz4730_gpio_setup(gpio_port_d, Jz4730_keypad_outputs[column], Hw::Gpio_chip::Input, 0);
  }

  l4_cache_clean_data((unsigned long) keypad,
                      (unsigned long) keypad + Jz4730_keypad_gpio_outputs_count);
}



static L4Re::Util::Registry_server<> server;



/* Main program. */

int main(void)
{
  /* Memory allocation capability for the keypad data. */

  L4::Cap<L4Re::Dataspace> mem;
  l4_size_t mem_size = Jz4730_keypad_gpio_outputs_count * sizeof(uint32_t);

  if (setup_memory()) return 1;

  mem = allocate_data(mem_size, &keymem);

  if (!mem.is_valid()) return 1;

  keypad = (uint32_t *) keymem;

  /* Set up keypad access and start scanning. */

  init_keyscan();

  /* Set up a thread to scan the keypad concurrently with the server loop. */

  Keypad_event_loop loop;
  loop.attach(scan_keypad);
  loop.start();

  /* Initialise and register a server object. */

  Keypad_server server_obj(mem);
  server.registry()->register_obj(&server_obj, "keypad");

  /* Enter the IPC server loop. */

  server.loop();
  return 0;
}
