/*
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include "cpm.h"

#include <l4/sys/types.h>
#include <stdint.h>



#ifdef __cplusplus

#include <l4/devices/hw_register_block.h>

/* A simple abstraction for accessing the CPM registers.
 * A proper device could inherit from Hw::Device and use an
 * Int_property for _exclk_freq and _rtclk_freq. */

class Cpm_jz4780_chip : public Cpm_chip
{
private:
  Hw::Register_block<32> _regs;
  uint32_t _exclk_freq, _rtclk_freq;

  // Utility methods.

  uint32_t get_field(uint32_t reg, uint32_t mask, uint8_t shift);
  void set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value);
  uint8_t _get_divider(uint32_t reg, uint32_t mask, uint8_t shift);

  // PLL control.

  int have_pll(uint32_t pll_reg);
  int pll_enabled(uint32_t pll_reg);
  int pll_bypassed(uint32_t pll_reg);

  // General frequency modifiers.

  uint16_t get_multiplier(uint32_t pll_reg);
  void set_multiplier(uint32_t pll_reg, uint16_t multiplier);
  uint8_t get_input_division(uint32_t pll_reg);
  void set_input_division(uint32_t pll_reg, uint8_t divider);
  uint8_t get_output_division(uint32_t pll_reg);
  void set_output_division(uint32_t pll_reg, uint8_t divider);

  // Clock dividers.

  void set_lcd_pixel_divider(uint16_t division);

  // Clock control.

  void _update_i2c(uint8_t channel, int enable);

  // Input frequencies.

  uint32_t get_pll_frequency(uint32_t pll_reg);

  // Clock sources.

  void set_hclock2_source(uint8_t source);
  void set_lcd_source(uint8_t source);

public:
  void set_pclock_source(uint8_t source);
  Cpm_jz4780_chip(l4_addr_t addr, uint32_t exclk_freq, uint32_t rtclk_freq);

  int have_clock();
  void start_clock();

  // Clock divider values.

  uint8_t get_cpu_divider();
  uint8_t get_hclock0_divider();
  uint8_t get_hclock2_divider();
  uint8_t get_pclock_divider();
  uint8_t get_lcd_pixel_divider();
  uint8_t get_memory_divider();

  // Clock control.

  void start_lcd();
  void stop_lcd();

  void start_i2c(uint8_t channel);
  void stop_i2c(uint8_t channel);

  // Input frequencies.

  uint8_t get_main_source();
  uint32_t get_main_frequency();

  // Clock sources, providing the input frequency.

  uint8_t get_cpu_source();
  uint8_t get_hclock0_source();
  uint8_t get_hclock2_source();
  uint8_t get_lcd_source();
  uint8_t get_memory_source();
  uint8_t get_pclock_source();

  uint32_t get_cpu_source_frequency();
  uint32_t get_hclock0_source_frequency();
  uint32_t get_hclock2_source_frequency();
  uint32_t get_lcd_source_frequency();
  uint32_t get_memory_source_frequency();
  uint32_t get_pclock_source_frequency();

  // Final, calculated frequencies.

  uint32_t get_cpu_frequency();
  uint32_t get_hclock0_frequency();
  uint32_t get_hclock2_frequency();
  uint32_t get_lcd_pixel_frequency();
  uint32_t get_memory_frequency();
  uint32_t get_pclock_frequency();

  uint32_t get_apll_frequency();
  uint32_t get_epll_frequency();
  uint32_t get_mpll_frequency();
  uint32_t get_vpll_frequency();

  void set_lcd_pixel_frequency(uint32_t pclk);
  void set_lcd_frequencies(uint32_t pclk, uint8_t multiplier);
  void set_pll_parameters(uint32_t pll_reg, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider);

  void update_output_frequency();
};

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *jz4780_cpm_init(l4_addr_t cpm_base);

int jz4780_cpm_have_clock(void *cpm);
void jz4780_cpm_start_clock(void *cpm);

void jz4780_cpm_start_lcd(void *cpm);
void jz4780_cpm_stop_lcd(void *cpm);

uint8_t jz4780_cpm_get_cpu_divider(void *cpm);
uint8_t jz4780_cpm_get_hclock0_divider(void *cpm);
uint8_t jz4780_cpm_get_hclock2_divider(void *cpm);
uint8_t jz4780_cpm_get_lcd_pixel_divider(void *cpm);
uint8_t jz4780_cpm_get_memory_divider(void *cpm);
uint8_t jz4780_cpm_get_pclock_divider(void *cpm);

uint8_t jz4780_cpm_get_hclock0_source(void *cpm);
uint8_t jz4780_cpm_get_hclock2_source(void *cpm);
uint8_t jz4780_cpm_get_lcd_source(void *cpm);
uint8_t jz4780_cpm_get_memory_source(void *cpm);
uint8_t jz4780_cpm_get_pclock_source(void *cpm);

uint32_t jz4780_cpm_get_hclock0_source_frequency(void *cpm);
uint32_t jz4780_cpm_get_hclock2_source_frequency(void *cpm);
uint32_t jz4780_cpm_get_lcd_source_frequency(void *cpm);
uint32_t jz4780_cpm_get_memory_source_frequency(void *cpm);
uint32_t jz4780_cpm_get_pclock_source_frequency(void *cpm);

void jz4780_cpm_set_pclock_source(void *cpm, uint8_t source);

uint8_t jz4780_cpm_get_main_source(void *cpm);
uint32_t jz4780_cpm_get_main_frequency(void *cpm);

uint32_t jz4780_cpm_get_cpu_frequency(void *cpm);
uint32_t jz4780_cpm_get_hclock0_frequency(void *cpm);
uint32_t jz4780_cpm_get_hclock2_frequency(void *cpm);
uint32_t jz4780_cpm_get_lcd_pixel_frequency(void *cpm);
uint32_t jz4780_cpm_get_memory_frequency(void *cpm);
uint32_t jz4780_cpm_get_pclock_frequency(void *cpm);

uint32_t jz4780_cpm_get_apll_frequency(void *cpm);
uint32_t jz4780_cpm_get_epll_frequency(void *cpm);
uint32_t jz4780_cpm_get_mpll_frequency(void *cpm);
uint32_t jz4780_cpm_get_vpll_frequency(void *cpm);

void jz4780_cpm_set_lcd_pixel_frequency(void *cpm, uint32_t pclk);
void jz4780_cpm_set_mpll_parameters(void *cpm, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider);

EXTERN_C_END
