/*
 * Clock and power management. This exposes the combined functionality
 * provided by the jz4730. The power management functionality could be exposed
 * using a separate driver.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-jz4730.h"



enum Regs : unsigned
{
  Clock_control       = 0x000,  // CFCR (CPCCR in JZ4740)
  Pll_control         = 0x010,  // PLCR1 (CPPCR in JZ4740)
  Clock_gate          = 0x020,  // MSCR (CLKGR in JZ4740)
  Sleep_control       = 0x024,  // SCR
  Lcd_divider         = 0x060,  // CFCR2
};

enum Clock_bits : unsigned
{
  Clock_enable          = 20, // UPE
  Clock_memory_divider  = 16, // MFR
  Clock_lcd_divider     = 12, // LFR
  Clock_pclock_divider  = 8,  // PFR (slow APB peripherals)
  Clock_hclock_divider  = 4,  // SFR (fast AHB peripherals)
  Clock_cpu_divider     = 0,  // IFR
};

enum Pll_bits : unsigned
{
  Pll_multiplier        = 23, // PLL1FD
  Pll_input_division    = 18, // PLL1RD
  Pll_output_division   = 16, // PLL1OD
  Pll_stable            = 10, // PLL1S
  Pll_bypassed          = 9,  // PLL1BP
  Pll_enabled           = 8,  // PLL1EN
};

enum Clock_gate_bits : unsigned
{
  Clock_gate_lcd        = 7,  // LCD
  Clock_gate_timer      = 3,  // OST
};

enum Lcd_divider_bits : unsigned
{
  Lcd_divider_value     = 0,  // PIXFR (in CFCR2)
};



// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Cpm_jz4730_chip::Cpm_jz4730_chip(l4_addr_t addr, uint32_t exclk_freq)
: _exclk_freq(exclk_freq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);

  // add_cid("cpm");
  // add_cid("cpm-jz4730");
  // register_property("exclk_freq", &_exclk_freq);
}

// Clock/timer control.

int
Cpm_jz4730_chip::have_clock()
{
  return !(_regs[Clock_gate] & (1 << Clock_gate_timer));
}

void
Cpm_jz4730_chip::start_clock()
{
  _regs[Clock_gate] = _regs[Clock_gate] & ~(1 << Clock_gate_timer);
}

// PLL control.

// Return whether the PLL is stable.

int
Cpm_jz4730_chip::have_pll()
{
  return _regs[Pll_control] & (1 << Pll_stable);
}

int
Cpm_jz4730_chip::pll_enabled()
{
  return _regs[Pll_control] & (1 << Pll_enabled);
}

int
Cpm_jz4730_chip::pll_bypassed()
{
  return _regs[Pll_control] & (1 << Pll_bypassed);
}

// Feedback (9-bit) multiplier.

uint16_t
Cpm_jz4730_chip::get_multiplier()
{
  return ((_regs[Pll_control] & (0x1ff << Pll_multiplier)) >> Pll_multiplier) + 2;
}

// Input (5-bit) divider.

uint8_t
Cpm_jz4730_chip::get_input_division()
{
  return ((_regs[Pll_control] & (0x1f << Pll_input_division)) >> Pll_input_division) + 2;
}

// Output divider.

static uint8_t od[4] = {1, 2, 2, 4};

uint8_t
Cpm_jz4730_chip::get_output_division()
{
  return od[(_regs[Pll_control] & (0x03 << Pll_output_division)) >> Pll_output_division];
}

// General clock divider.

static uint8_t cd[] = {1, 2, 3, 4, 6, 8, 12, 16, 24, 32};

uint8_t
Cpm_jz4730_chip::_get_divider(uint32_t reg, uint32_t mask, uint8_t shift)
{
  uint8_t d = (_regs[reg] & mask) >> shift;
  return (d < 10) ? cd[d] : 1;
}

// CPU clock (CCLK) divider.

uint8_t
Cpm_jz4730_chip::get_cpu_divider()
{
  return _get_divider(Clock_control, 0xf << Clock_cpu_divider, Clock_cpu_divider);
}

// Fast peripheral clock (HCLK) divider.

uint8_t
Cpm_jz4730_chip::get_hclock_divider()
{
  return _get_divider(Clock_control, 0xf << Clock_hclock_divider, Clock_hclock_divider);
}

// Slow peripheral clock (PCLK) divider.

uint8_t
Cpm_jz4730_chip::get_pclock_divider()
{
  return _get_divider(Clock_control, 0xf << Clock_pclock_divider, Clock_pclock_divider);
}

// Memory clock (MCLK) divider.

uint8_t
Cpm_jz4730_chip::get_memory_divider()
{
  return _get_divider(Clock_control, 0xf << Clock_memory_divider, Clock_memory_divider);
}

// Clock source divider for MSC, I2S, LCD and USB.

uint8_t
Cpm_jz4730_chip::get_source_divider()
{
  return 1;
}

// LCD device clock divider.

void
Cpm_jz4730_chip::set_lcd_device_divider(uint8_t division)
{
  if (division == 0)
    division = 1;

  // NOTE: The vendor code (clock.c) bounds the divider at 16, but maybe this is
  // NOTE: confused with the width of the bitfield which actually contains an
  // NOTE: index for the clock divider value array (cd).

  else if (division > 16)
    division = 16;

  // Obtain the divider or closest higher divider.

  int i;

  for (i = 0; i < 10; i++)
    if (cd[i] >= division) break;

  _regs[Clock_control] = (_regs[Clock_control] & ~(0xf << Clock_lcd_divider)) |
                         (cd[i] << Clock_lcd_divider);
}

// LCD pixel clock divider.

uint16_t
Cpm_jz4730_chip::get_lcd_pixel_divider()
{
  return (_regs[Lcd_divider] >> Lcd_divider_value) + 1;
}

void
Cpm_jz4730_chip::set_lcd_pixel_divider(uint16_t division)
{
  if (division == 0)
    division = 1;
  else if (division > 512)
    division = 512;

  _regs[Lcd_divider] = (_regs[Lcd_divider] & ~(0x1ff << Lcd_divider_value)) |
                       ((division - 1) << Lcd_divider_value);
}



uint32_t
Cpm_jz4730_chip::get_lcd_pixel_frequency()
{
  return get_output_frequency() / get_lcd_pixel_divider();
}

// Set the device and pixel frequencies, indicating the latter and
// providing the device:pixel frequency ratio.

void
Cpm_jz4730_chip::set_lcd_frequencies(uint32_t pclk, uint8_t ratio)
{
  uint32_t out = get_output_frequency(),
           lcd = pclk * ratio;

  set_lcd_pixel_divider(out / pclk);

  // Limit the device frequency to 150MHz.

  if (lcd > 150000000) lcd = 150000000;

  set_lcd_device_divider(out / lcd);
}



// LCD clock control.

void
Cpm_jz4730_chip::start_lcd()
{
  _regs[Clock_gate] = _regs[Clock_gate] & ~(1 << Clock_gate_lcd);
}

void
Cpm_jz4730_chip::stop_lcd()
{
  _regs[Clock_gate] = _regs[Clock_gate] | (1 << Clock_gate_lcd);
}



uint32_t
Cpm_jz4730_chip::get_pll_frequency()
{
  // Test for PLL enable and not PLL bypass.

  if (pll_enabled() && !pll_bypassed())
    return (_exclk_freq * get_multiplier()) /
           (get_input_division() * get_output_division());
  else
    return _exclk_freq;
}

// Clock frequency for MSC, I2S, LCD and USB.

uint32_t
Cpm_jz4730_chip::get_output_frequency()
{
  return get_pll_frequency() / get_source_divider();
}

void
Cpm_jz4730_chip::update_output_frequency()
{
  _regs[Clock_control] = _regs[Clock_control] | (1 << Clock_enable);
}

// Clock frequency for the CPU.

uint32_t Cpm_jz4730_chip::get_cpu_frequency()
{
  if (pll_enabled())
    return get_pll_frequency() / get_cpu_divider();
  else
    return _exclk_freq;
}

// Clock frequency for fast peripherals.

uint32_t
Cpm_jz4730_chip::get_hclock_frequency()
{
  if (pll_enabled())
    return get_pll_frequency() / get_hclock_divider();
  else
    return _exclk_freq;
}

// Clock frequency for slow peripherals.

uint32_t
Cpm_jz4730_chip::get_pclock_frequency()
{
  if (pll_enabled())
    return get_pll_frequency() / get_pclock_divider();
  else
    return _exclk_freq;
}

// Clock frequency for the memory.

uint32_t
Cpm_jz4730_chip::get_memory_frequency()
{
  if (pll_enabled())
    return get_pll_frequency() / get_memory_divider();
  else
    return _exclk_freq;
}



// C language interface functions.

void
*jz4730_cpm_init(l4_addr_t cpm_base)
{
  /* Initialise the clock and power management peripheral with the
     register memory region and a 3.6864 MHz EXCLK frequency. */

  return (void *) new Cpm_jz4730_chip(cpm_base, 3686400);
}

int
jz4730_cpm_have_clock(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->have_clock();
}

void
jz4730_cpm_start_clock(void *cpm)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->start_clock();
}

void
jz4730_cpm_start_lcd(void *cpm)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->start_lcd();
}

void
jz4730_cpm_stop_lcd(void *cpm)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->stop_lcd();
}

uint32_t
jz4730_cpm_get_cpu_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_cpu_frequency();
}

uint32_t
jz4730_cpm_get_hclock_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_hclock_frequency();
}

uint32_t
jz4730_cpm_get_output_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_output_frequency();
}

uint32_t
jz4730_cpm_get_pclock_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_pclock_frequency();
}

uint32_t
jz4730_cpm_get_memory_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_memory_frequency();
}

uint16_t
jz4730_cpm_get_lcd_pixel_divider(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_lcd_pixel_divider();
}

uint32_t
jz4730_cpm_get_lcd_pixel_frequency(void *cpm)
{
  return static_cast<Cpm_jz4730_chip *>(cpm)->get_lcd_pixel_frequency();
}

void
jz4730_cpm_set_lcd_frequencies(void *cpm, uint32_t pclk, uint8_t ratio)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->set_lcd_frequencies(pclk, ratio);
}

void
jz4730_cpm_update_output_frequency(void *cpm)
{
  static_cast<Cpm_jz4730_chip *>(cpm)->update_output_frequency();
}
