/*
 * Clock and power management. This exposes the combined functionality
 * provided by the jz4780 and related SoCs. The power management
 * functionality could be exposed using a separate driver.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-jz4780.h"



enum Regs : unsigned
{
  Clock_control       = 0x000,  // CPCCR
  Low_power_control   = 0x004,  // LCR
  Pll_control         = 0x00c,  // CPPCR
  Pll_control_A       = 0x010,  // CPAPCR
  Pll_control_M       = 0x014,  // CPMPCR
  Pll_control_E       = 0x018,  // CPEPCR
  Pll_control_V       = 0x01c,  // CPVPCR
  Clock_gate0         = 0x020,  // CLKGR0
  Clock_gate1         = 0x028,  // CLKGR1
  Sleep_control       = 0x024,  // OPCR (oscillator and power control)
  Ddr_divider         = 0x02c,  // DDRCDR
  I2s_divider0        = 0x060,  // I2SCDR
  I2s_divider1        = 0x0a0,  // I2S1CDR
  Lcd_divider0        = 0x054,  // LP0CDR
  Lcd_divider1        = 0x064,  // LP1CDR
  Msc_divider0        = 0x068,  // MSC0CDR
  Msc_divider1        = 0x0a4,  // MSC1CDR
  Msc_divider2        = 0x0a8,  // MSC2CDR
  Uhc_divider         = 0x06c,  // UHCCDR
  Ssi_divider         = 0x074,  // SSICDR
};

enum Clock_bits : unsigned
{
  Clock_enable          = 22, // CE_CPU
  Clock_pclock_divider  = 16, // PDIV (slow APB peripherals)
  Clock_hclock2_divider = 12, // H2DIV (fast AHB peripherals)
  Clock_hclock0_divider = 8,  // H0DIV (fast AHB peripherals)
  Clock_cpu_divider     = 0,  // CDIV
};

enum Pll_bits : unsigned
{
  Pll_multiplier        = 19, // xPLLM
  Pll_input_division    = 13, // xPLLN
  Pll_output_division   = 9,  // xPLLOD
  Pll_stable            = 4,  // xPLL_ON
  Pll_bypassed          = 1,  // xPLLBP
  Pll_enabled           = 0,  // xPLLEN
};

enum Clock_source_bits : unsigned
{
  Clock_source_main     = 30, // SEL_SRC (output to SCLK_A)
  Clock_source_cpu      = 28, // SEL_CPLL (output to CCLK)
  Clock_source_hclock0  = 26, // SEL_H0PLL (output to AHB0)
  Clock_source_hclock2  = 24, // SEL_H2PLL (output to AHB2)
  Clock_source_ddr      = 30, // DCS
  Clock_source_i2s      = 31, // I2CS
  Clock_source_lcd      = 30, // LPCS
};

enum Clock_sources : unsigned
{
  Source_pll_A          = 1,  // APLL
  Source_external       = 2,  // EXCLK
  Source_realtime       = 3,  // RTCLK
  Source_main_frequency = 1,  // SCLK_A
  Source_pll_M          = 2,  // MPLL
  Source_pll_E          = 3,  // EPLL
  Source_pll_V          = 3,  // VPLL
};

enum Clock_gate_bits : unsigned
{
  Clock_gate_lcd        = 28, // LCD (in CLKGR0)
  Clock_gate_tve        = 27, // TVE (in CLKGR0)
  Clock_gate_smb4       = 12, // SMB4 (in CLKGR1)
  Clock_gate_smb3       = 0,  // SMB3 (in CLKGR1)
  Clock_gate_smb2       = 25, // SMB2 (in CLKGR0)
  Clock_gate_smb1       = 6,  // SMB2 (in CLKGR0)
  Clock_gate_smb0       = 5,  // SMB2 (in CLKGR0)
};

enum Clock_gate_regs : unsigned
{
  Clock_gate_reg_smb4   = Clock_gate1,
  Clock_gate_reg_smb3   = Clock_gate1,
  Clock_gate_reg_smb2   = Clock_gate0,
  Clock_gate_reg_smb1   = Clock_gate0,
  Clock_gate_reg_smb0   = Clock_gate0,
};

enum Divider_bits : unsigned
{
  Ddr_divider_value     = 0,  // DDRCDR
  Lcd_divider_value     = 0,  // LPCDR
};

enum Lcd_clock_bits : unsigned
{
  Lcd_change_enable     = 0x1000, // CE_LCD
  Lcd_change_busy       = 0x0800, // LCD_BUSY
  Lcd_clock_stop        = 0x0400, // LCD_STOP
};



// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Cpm_jz4780_chip::Cpm_jz4780_chip(l4_addr_t addr, uint32_t exclk_freq, uint32_t rtclk_freq)
: _exclk_freq(exclk_freq), _rtclk_freq(rtclk_freq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);

  // add_cid("cpm");
  // add_cid("cpm-jz4780");
  // register_property("exclk_freq", &_exclk_freq);
}

// Clock/timer control.
// NOTE: For the time being, assume that the system is configured.

int
Cpm_jz4780_chip::have_clock()
{
  // return !(_regs[Clock_gate] & (1 << Clock_gate_timer));
  return 1;
}

void
Cpm_jz4780_chip::start_clock()
{
  // _regs[Clock_gate] = _regs[Clock_gate] & ~(1 << Clock_gate_timer);
}



// Utility methods.

uint32_t
Cpm_jz4780_chip::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Cpm_jz4780_chip::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}

// General clock divider access.

uint8_t
Cpm_jz4780_chip::_get_divider(uint32_t reg, uint32_t mask, uint8_t shift)
{
  uint8_t d = get_field(reg, mask, shift);

	// NOTE: Value 14 stops the clock, 15 presumably resumes the clock.

  return (d < 14) ? d + 1 : 1;
}



// PLL control.

// Return whether the PLL is stable.

int
Cpm_jz4780_chip::have_pll(uint32_t pll_reg)
{
  return _regs[pll_reg] & (1 << Pll_stable);
}

int
Cpm_jz4780_chip::pll_enabled(uint32_t pll_reg)
{
  return _regs[pll_reg] & (1 << Pll_enabled);
}

int
Cpm_jz4780_chip::pll_bypassed(uint32_t pll_reg)
{
  return _regs[pll_reg] & (1 << Pll_bypassed);
}

// Feedback (13-bit) multiplier.

uint16_t
Cpm_jz4780_chip::get_multiplier(uint32_t pll_reg)
{
  return get_field(pll_reg, 0x1fff, Pll_multiplier) + 1;
}

void
Cpm_jz4780_chip::set_multiplier(uint32_t pll_reg, uint16_t multiplier)
{
  set_field(pll_reg, 0x1fff, Pll_multiplier, multiplier - 1);
}

// Input (6-bit) divider.

uint8_t
Cpm_jz4780_chip::get_input_division(uint32_t pll_reg)
{
  return get_field(pll_reg, 0x3f, Pll_input_division) + 1;
}

void
Cpm_jz4780_chip::set_input_division(uint32_t pll_reg, uint8_t divider)
{
  set_field(pll_reg, 0x3f, Pll_input_division, divider - 1);
}

// Output divider.

uint8_t
Cpm_jz4780_chip::get_output_division(uint32_t pll_reg)
{
  uint8_t d = get_field(pll_reg, 0x0f, Pll_output_division);

  // Zero yields a division of one. Otherwise enforce even results.

  return d == 0 ? 1 : (d + 1) & 0x0e;
}

void
Cpm_jz4780_chip::set_output_division(uint32_t pll_reg, uint8_t divider)
{
  uint8_t d = divider <= 1 ? 0 : (divider & 0x0e) - 1;

  set_field(pll_reg, 0x0f, Pll_output_division, d);
}

uint32_t
Cpm_jz4780_chip::get_pll_frequency(uint32_t pll_reg)
{
  // Test for PLL enable and not PLL bypass.

  if (pll_enabled(pll_reg) && !pll_bypassed(pll_reg))
    return (_exclk_freq * get_multiplier(pll_reg)) /
           (get_input_division(pll_reg) * get_output_division(pll_reg));
  else
    return _exclk_freq;
}

void
Cpm_jz4780_chip::set_pll_parameters(uint32_t pll_reg, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  set_multiplier(pll_reg, multiplier);
  set_input_division(pll_reg, in_divider);
  set_output_division(pll_reg, out_divider);

  if (pll_enabled(pll_reg) && !pll_bypassed(pll_reg))
    while (!have_pll(pll_reg));
}



// CPU clock (CCLK) divider.

uint8_t
Cpm_jz4780_chip::get_cpu_divider()
{
  return _get_divider(Clock_control, 0xf, Clock_cpu_divider);
}

// Fast peripheral clock (H0CLK) divider.

uint8_t
Cpm_jz4780_chip::get_hclock0_divider()
{
  return _get_divider(Clock_control, 0xf, Clock_hclock0_divider);
}

// Fast peripheral clock (H2CLK) divider.

uint8_t
Cpm_jz4780_chip::get_hclock2_divider()
{
  return _get_divider(Clock_control, 0xf, Clock_hclock2_divider);
}

// Slow peripheral clock (PCLK) divider.

uint8_t
Cpm_jz4780_chip::get_pclock_divider()
{
  return _get_divider(Clock_control, 0xf, Clock_pclock_divider);
}

// LCD clock (LPCLK) divider for LCD0 pixel clock.

uint8_t
Cpm_jz4780_chip::get_lcd_pixel_divider()
{
  return get_field(Lcd_divider0, 0xff, Lcd_divider_value) + 1;
}

// Memory clock (DDR_CLK) divider.

uint8_t
Cpm_jz4780_chip::get_memory_divider()
{
  return _get_divider(Ddr_divider, 0xf, Ddr_divider_value);
}

// LCD pixel clock divider.
// NOTE: This only supports the first LCD peripheral.

void
Cpm_jz4780_chip::set_lcd_pixel_divider(uint16_t division)
{
  if ((division < 1) || (division > 256))
    return;

  // Enable change.

  _regs[Lcd_divider0] = _regs[Lcd_divider0] | Lcd_change_enable;

  // Set the divider.

  set_field(Lcd_divider0, 0xff, Lcd_divider_value, division - 1);

  // Restart clock and disable change.

  while (_regs[Lcd_divider0] & Lcd_change_busy);
  _regs[Lcd_divider0] = _regs[Lcd_divider0] & ~Lcd_change_enable;
}



// Clock gating control.

void
Cpm_jz4780_chip::start_lcd()
{
  _regs[Clock_gate0] = _regs[Clock_gate0] & ~(1 << Clock_gate_lcd);
}

void
Cpm_jz4780_chip::stop_lcd()
{
  _regs[Clock_gate0] = _regs[Clock_gate0] | (1 << Clock_gate_lcd);
}

void
Cpm_jz4780_chip::_update_i2c(uint8_t channel, int enable)
{
  uint8_t bit;
  uint32_t reg;
  switch (channel)
  {
    case 0: bit = Clock_gate_smb0; reg = Clock_gate_reg_smb0; break;
    case 1: bit = Clock_gate_smb1; reg = Clock_gate_reg_smb1; break;
    case 2: bit = Clock_gate_smb2; reg = Clock_gate_reg_smb2; break;
    case 3: bit = Clock_gate_smb3; reg = Clock_gate_reg_smb3; break;
    case 4: bit = Clock_gate_smb4; reg = Clock_gate_reg_smb4; break;
    default: return;
  }
  if (enable)
    _regs[reg] = _regs[reg] & ~(1 << bit);
  else
    _regs[reg] = _regs[reg] | (1 << bit);
}

void
Cpm_jz4780_chip::start_i2c(uint8_t channel)
{
  _update_i2c(channel, 1);
}

void
Cpm_jz4780_chip::stop_i2c(uint8_t channel)
{
  _update_i2c(channel, 0);
}



// Clock sources.

uint8_t
Cpm_jz4780_chip::get_memory_source()
{
  return get_field(Ddr_divider, 0x3, Clock_source_ddr);
}

uint32_t
Cpm_jz4780_chip::get_memory_source_frequency()
{
  switch (get_memory_source())
  {
    case Source_main_frequency:
    return get_main_frequency();
    case Source_pll_M:
    return get_pll_frequency(Pll_control_M);
    default:
    return 0;
  }
}

uint8_t
Cpm_jz4780_chip::get_cpu_source()
{
  return get_field(Clock_control, 0x3, Clock_source_cpu);
}

uint32_t
Cpm_jz4780_chip::get_cpu_source_frequency()
{
  switch (get_cpu_source())
  {
    case Source_main_frequency:
    return get_main_frequency();
    case Source_pll_M:
    return get_pll_frequency(Pll_control_M);
    case Source_pll_E:
    return get_pll_frequency(Pll_control_E);
    default:
    return 0;
  }
}

uint8_t
Cpm_jz4780_chip::get_hclock0_source()
{
  return get_field(Clock_control, 0x3, Clock_source_hclock0);
}

uint32_t
Cpm_jz4780_chip::get_hclock0_source_frequency()
{
  switch (get_hclock0_source())
  {
    case Source_main_frequency:
    return get_main_frequency();
    case Source_pll_M:
    return get_pll_frequency(Pll_control_M);
    case Source_pll_E:
    return get_pll_frequency(Pll_control_E);
    default:
    return 0;
  }
}

uint8_t
Cpm_jz4780_chip::get_hclock2_source()
{
  return get_field(Clock_control, 0x3, Clock_source_hclock2);
}

uint32_t
Cpm_jz4780_chip::get_hclock2_source_frequency()
{
  switch (get_hclock2_source())
  {
    case Source_main_frequency:
    return get_main_frequency();
    case Source_pll_M:
    return get_pll_frequency(Pll_control_M);
    case Source_realtime:
    return _rtclk_freq; // "TCK" in the manual, RTCLK in Linux driver code
    default:
    return 0;
  }
}

void
Cpm_jz4780_chip::set_hclock2_source(uint8_t source)
{
  set_field(Clock_control, 0x3, Clock_source_hclock2, source);
}

uint8_t
Cpm_jz4780_chip::get_lcd_source()
{
  return get_field(Lcd_divider0, 0x3, Clock_source_lcd);
}

uint32_t
Cpm_jz4780_chip::get_lcd_source_frequency()
{
  switch (get_lcd_source())
  {
    case Source_main_frequency:
    return get_main_frequency();
    case Source_pll_M:
    return get_pll_frequency(Pll_control_M);
    case Source_pll_V:
    return get_pll_frequency(Pll_control_V);
    default:
    return 0;
  }
}

void
Cpm_jz4780_chip::set_lcd_source(uint8_t source)
{
  // Stop clock and enable change.

  _regs[Lcd_divider0] = _regs[Lcd_divider0] | Lcd_change_enable | Lcd_clock_stop;

  // Set the source.

  set_field(Lcd_divider0, 0x03, Clock_source_lcd, source);

  // Restart clock and disable change.

  while (_regs[Lcd_divider0] & Lcd_change_busy);
  _regs[Lcd_divider0] = _regs[Lcd_divider0] & ~(Lcd_change_enable | Lcd_clock_stop);
}

uint8_t
Cpm_jz4780_chip::get_pclock_source()
{
  return get_hclock2_source();
}

uint32_t
Cpm_jz4780_chip::get_pclock_source_frequency()
{
  return get_hclock2_source_frequency();
}

void
Cpm_jz4780_chip::set_pclock_source(uint8_t source)
{
  set_hclock2_source(source);
}



// Source frequency, used by various clock sources.

uint8_t
Cpm_jz4780_chip::get_main_source()
{
  return get_field(Clock_control, 0x3, Clock_source_main);
}

uint32_t
Cpm_jz4780_chip::get_main_frequency()
{
  switch (get_main_source())
  {
    case Source_pll_A:
    return get_pll_frequency(Pll_control_A);
    case Source_external:
    return _exclk_freq;
    case Source_realtime:
    return _rtclk_freq;
    default:
    return 0;
  }
}

// Clock frequency for the CPU.

uint32_t
Cpm_jz4780_chip::get_cpu_frequency()
{       
  return get_cpu_source_frequency() / get_cpu_divider();
}

// Clock frequency for fast peripherals.

uint32_t
Cpm_jz4780_chip::get_hclock0_frequency()
{
  return get_hclock0_source_frequency() / get_hclock0_divider();
}

// Clock frequency for fast peripherals.

uint32_t
Cpm_jz4780_chip::get_hclock2_frequency()
{
  return get_hclock2_source_frequency() / get_hclock2_divider();
}

// Clock frequency for slow peripherals.

uint32_t
Cpm_jz4780_chip::get_pclock_frequency()
{
  return get_pclock_source_frequency() / get_pclock_divider();
}

// Clock frequency for the LCD0 controller.

uint32_t
Cpm_jz4780_chip::get_lcd_pixel_frequency()
{
  return get_lcd_source_frequency() / get_lcd_pixel_divider();
}

// Clock frequency for the memory.

uint32_t
Cpm_jz4780_chip::get_memory_frequency()
{
  return get_memory_source_frequency() / get_memory_divider();
}

uint32_t
Cpm_jz4780_chip::get_apll_frequency()
{
  return get_pll_frequency(Pll_control_A);
}

uint32_t
Cpm_jz4780_chip::get_epll_frequency()
{
  return get_pll_frequency(Pll_control_E);
}

uint32_t
Cpm_jz4780_chip::get_mpll_frequency()
{
  return get_pll_frequency(Pll_control_M);
}

uint32_t
Cpm_jz4780_chip::get_vpll_frequency()
{
  return get_pll_frequency(Pll_control_V);
}



// Set the pixel frequency.
// Unlike the jz4740, HCLK/AHB0 is used as the device frequency, with the pixel
// frequency being based on the selected clock source (SCLK_A, MPLL or VPLL).

void
Cpm_jz4780_chip::set_lcd_pixel_frequency(uint32_t pclk)
{
  // Switch to the video PLL and attempt to set the divider.

  set_lcd_source(Source_pll_V);
  set_lcd_pixel_divider(get_lcd_source_frequency() / pclk);
}

// NOTE: Compatibility method. Probably needs reviewing.

void
Cpm_jz4780_chip::set_lcd_frequencies(uint32_t pclk, uint8_t ratio)
{
  (void) ratio;
  set_lcd_pixel_frequency(pclk);
}

// NOTE: Empty method for compatibility.

void
Cpm_jz4780_chip::update_output_frequency()
{
}



// C language interface functions.

void
*jz4780_cpm_init(l4_addr_t cpm_base)
{
  /* Initialise the clock and power management peripheral with the
     register memory region and a 48MHz EXCLK frequency. */

  return (void *) new Cpm_jz4780_chip(cpm_base, 48000000, 32768);
}

int
jz4780_cpm_have_clock(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->have_clock();
}

void
jz4780_cpm_start_clock(void *cpm)
{
  static_cast<Cpm_jz4780_chip *>(cpm)->start_clock();
}



void
jz4780_cpm_start_lcd(void *cpm)
{
  static_cast<Cpm_jz4780_chip *>(cpm)->start_lcd();
}

void
jz4780_cpm_stop_lcd(void *cpm)
{
  static_cast<Cpm_jz4780_chip *>(cpm)->stop_lcd();
}



uint8_t
jz4780_cpm_get_cpu_divider(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_cpu_divider();
}

uint8_t
jz4780_cpm_get_hclock0_divider(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_hclock0_divider();
}

uint8_t
jz4780_cpm_get_hclock2_divider(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_hclock2_divider();
}

uint8_t
jz4780_cpm_get_lcd_pixel_divider(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_lcd_pixel_divider();
}

uint8_t
jz4780_cpm_get_memory_divider(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_memory_divider();
}

uint8_t
jz4780_cpm_get_pclock_divider(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_pclock_divider();
}



uint8_t
jz4780_cpm_get_hclock0_source(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_hclock0_source();
}

uint8_t
jz4780_cpm_get_hclock2_source(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_hclock2_source();
}

uint8_t
jz4780_cpm_get_lcd_source(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_lcd_source();
}

uint8_t
jz4780_cpm_get_memory_source(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_memory_source();
}

uint8_t
jz4780_cpm_get_pclock_source(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_pclock_source();
}

void
jz4780_cpm_set_pclock_source(void *cpm, uint8_t source)
{       
  static_cast<Cpm_jz4780_chip *>(cpm)->set_pclock_source(source);
}



uint32_t
jz4780_cpm_get_hclock0_source_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_hclock0_source_frequency();
}

uint32_t
jz4780_cpm_get_hclock2_source_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_hclock2_source_frequency();
}

uint32_t
jz4780_cpm_get_lcd_source_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_lcd_source_frequency();
}

uint32_t
jz4780_cpm_get_memory_source_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_memory_source_frequency();
}

uint32_t
jz4780_cpm_get_pclock_source_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_pclock_source_frequency();
}



uint8_t
jz4780_cpm_get_main_source(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_main_source();
}

uint32_t
jz4780_cpm_get_main_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_main_frequency();
}

uint32_t
jz4780_cpm_get_cpu_frequency(void *cpm)
{       
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_cpu_frequency();
}

uint32_t
jz4780_cpm_get_hclock0_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_hclock0_frequency();
}

uint32_t
jz4780_cpm_get_hclock2_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_hclock2_frequency();
}

uint32_t
jz4780_cpm_get_lcd_pixel_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_lcd_pixel_frequency();
}

uint32_t
jz4780_cpm_get_memory_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_memory_frequency();
}

uint32_t
jz4780_cpm_get_pclock_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_pclock_frequency();
}

uint32_t
jz4780_cpm_get_apll_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_apll_frequency();
}

uint32_t
jz4780_cpm_get_epll_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_epll_frequency();
}

uint32_t
jz4780_cpm_get_mpll_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_mpll_frequency();
}

uint32_t
jz4780_cpm_get_vpll_frequency(void *cpm)
{
  return static_cast<Cpm_jz4780_chip *>(cpm)->get_vpll_frequency();
}



void
jz4780_cpm_set_lcd_pixel_frequency(void *cpm, uint32_t pclk)
{
  static_cast<Cpm_jz4780_chip *>(cpm)->set_lcd_pixel_frequency(pclk);
}

void
jz4780_cpm_set_mpll_parameters(void *cpm, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  static_cast<Cpm_jz4780_chip *>(cpm)->set_pll_parameters(Pll_control_M, multiplier, in_divider, out_divider);
}
