/*
 * GPIO driver for Ingenic JZ4740.
 * (See below for additional copyright and licensing notices.)
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 *
 *
 * Subject to other copyrights, being derived from the bcm2835.cc and
 * omap.cc GPIO driver implementations.
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */

#pragma once

#include <l4/sys/err.h>
#include <l4/sys/types.h>
#include <stdint.h>
#include "gpio.h"



#ifdef __cplusplus

#include <l4/devices/hw_mmio_register_block.h>

// GPIO device control.

class Gpio_jz4740_irq_pin : public Hw::Gpio_irq_pin
{
  unsigned _pin;
  Hw::Register_block<32> _regs;

  // Convenience method for obtaining the bit corresponding to a pin.

  l4_uint32_t _pin_bit(unsigned pin)
  { return 1 << (pin & 31); }

  void write_reg_pin(unsigned reg);

public:
  Gpio_jz4740_irq_pin(unsigned pin, Hw::Register_block<32> const &regs);

  void do_mask();
  void do_unmask();
  bool do_set_mode(unsigned mode);
  int clear();
  bool enabled();
};

class Gpio_jz4740_chip : public Hw::Gpio_chip
{
private:
  Hw::Register_block<32> _regs;

  l4_addr_t _start, _end;
  unsigned _nr_pins;

  // Convenience method for obtaining the bit corresponding to a pin.

  l4_uint32_t _pin_bit(unsigned pin)
  { return 1 << (pin & 31); }

  // Convenience method for obtaining the bit position of a pin.

  unsigned _pin_shift(unsigned pin)
  { return pin % 32; }

  // Permit only "aligned" accesses to registers.

  unsigned _reg_offset_check(unsigned pin_offset) const
  {
    switch (pin_offset)
      {
      case 0:
        return 0;

      default:
        throw -L4_EINVAL;
      }
  }

  // General configuration register updates.

  void _config(unsigned bitmap, unsigned mode);

  // General pull-up register updates.

  void _config_pull(unsigned bitmap, unsigned mode);

  // General pad register updates.

  void _config_pad(unsigned bitmap, unsigned func, unsigned value);

public:
  Gpio_jz4740_chip(l4_addr_t start, l4_addr_t end,
                   unsigned nr_pins);

  // Obtain the number of pins.

  unsigned nr_pins() const { return _nr_pins; }

  // Unnecessary operations.

  void request(unsigned) {}
  void free(unsigned) {}

  // Configuration methods.

  void setup(unsigned pin, unsigned mode, int value = 0);
  void config_pull(unsigned pin, unsigned mode);
  void config_pad(unsigned pin, unsigned func, unsigned value);
  void config_get(unsigned pin, unsigned reg, unsigned *value);

  // Multiple pin configuration methods.

  void multi_setup(Pin_slice const &mask, unsigned mode, unsigned outvalues = 0);
  void multi_config_pull(Pin_slice const &mask, unsigned mode);
  void multi_config_pad(Pin_slice const &mask, unsigned func, unsigned value = 0);
  void multi_set(Pin_slice const &mask, unsigned data);
  unsigned multi_get(unsigned offset);

  // IRQ pin configuration.

  Hw::Gpio_irq_pin *get_irq(unsigned pin);

  // Pin/port data methods.

  int get(unsigned pin);
  void set(unsigned pin, int value);

private:
  void config(unsigned pin, unsigned mode);
};

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *jz4740_gpio_init(l4_addr_t start, l4_addr_t end, unsigned pins);

void jz4740_gpio_setup(void *gpio, unsigned pin, unsigned mode, int value);
void jz4740_gpio_config_pull(void *gpio, unsigned pin, unsigned mode);
void jz4740_gpio_config_pad(void *gpio, unsigned pin, unsigned func, unsigned value);
void jz4740_gpio_config_get(void *gpio, unsigned pin, unsigned reg, unsigned *value);

void jz4740_gpio_multi_setup(void *gpio, Pin_slice const *mask, unsigned mode, unsigned outvalues);
void jz4740_gpio_multi_config_pull(void *gpio, Pin_slice const *mask, unsigned mode);
void jz4740_gpio_multi_config_pad(void *gpio, Pin_slice const *mask, unsigned func, unsigned value);
void jz4740_gpio_multi_set(void *gpio, Pin_slice const *mask, unsigned data);
unsigned jz4740_gpio_multi_get(void *gpio, unsigned offset);

int jz4740_gpio_get(void *gpio, unsigned pin);
void jz4740_gpio_set(void *gpio, unsigned pin, int value);

void *jz4740_gpio_get_irq(void *gpio, unsigned pin);
bool jz4740_gpio_irq_set_mode(void *gpio_irq, unsigned mode);

EXTERN_C_END
