/*
 * GPIO driver definitions.
 * (See below for additional copyright and licensing notices.)
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 *
 *
 * Subject to other copyrights, being derived from the gpio headers in the io
 * package.
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */

#pragma once

#include <l4/vbus/vbus_gpio.h>

typedef struct Pin_slice
{
  unsigned offset;
  unsigned mask;
} Pin_slice;

#ifdef __cplusplus

namespace Hw {

// IRQ control for each GPIO pin.

class Gpio_irq_pin
{
public:
  virtual void do_mask() = 0;
  virtual void do_unmask() = 0;
  virtual bool do_set_mode(unsigned mode) = 0;
  virtual int clear() = 0;
  virtual bool enabled() = 0;
};

// GPIO peripheral control.

class Gpio_chip
{
public:
  /// Modes used for setup()
  enum Fix_mode
  {
    Input  = L4VBUS_GPIO_SETUP_INPUT,  ///< Configure as input pin
    Output = L4VBUS_GPIO_SETUP_OUTPUT, ///< Configure as output pin
    Irq    = L4VBUS_GPIO_SETUP_IRQ,    ///< Configure as IRQ source
  };

  /// Modes used for pull up / pull down (config_pull())
  enum Pull_mode
  {
    Pull_none  = L4VBUS_GPIO_PIN_PULL_NONE,  ///< No pull up or pull down resistors
    Pull_up    = L4VBUS_GPIO_PIN_PULL_UP,    ///< Enable pull up resistor
    Pull_down  = L4VBUS_GPIO_PIN_PULL_DOWN,  ///< Enable pull down resistor
  };

  // Generic function options
  enum Function_levels
  {
    Function_gpio,
    Function_alt,
  };

  virtual void request(unsigned pin) = 0;
  virtual void free(unsigned pin) = 0;

  /**
   * \brief Request number of pins from GPIO chip
   * \return Number of pins of this GPIO chip
   */
  virtual unsigned nr_pins() const = 0;

  /**
   * \brief Generic (platform independent) setup for a pin.
   * \param pin the pin number to configure.
   * \param mode the mode for the pin (see Fix_mode).
   * \param value the value if the pin is configured as output.
   */
  virtual void setup(unsigned pin, unsigned mode, int value = 0) = 0;

  /**
   * \brief Generic (platform independet) function to set a pin's pull up/down mode
   * \param pin The pin number to configure.
   * \param mode The pull up/down mode (see Pull_mode).
   */
  virtual void config_pull(unsigned pin, unsigned mode) = 0;

  /**
   * \brief Set platform specific pad configuration.
   * \param pin the pin to configure.
   * \param func a platform specific sub-function of a pad to be configured
   * \param value a platform specific value for the given sub-function.
   */
  virtual void config_pad(unsigned pin, unsigned func, unsigned value) = 0;

  /**
   * \brief Get Platform specific pad configuration.
   * \param pin the pin to configure.
   * \param func a platform specific sub-function of a pad to be configured
   * \retparam value a platform specific value for the given sub-function.
   */
  virtual void config_get(unsigned pin, unsigned func, unsigned *value) = 0;

  /**
   * \brief Get the value of the given pin (generic API).
   * \param pin the pin to read the value from.
   */
  virtual int get(unsigned pin) = 0;

  /**
   * \brief Set the value of the given pin (generic API).
   * \pre The pin has to be configured as output before using this function,
   *      otherwise this call will be ignored.
   * \param pin the pin to write the value to.
   * \param value the value to program for output (0 or 1).
   */
  virtual void set(unsigned pin, int value) = 0;

  /**
   * \brief Setup multiple pins (generic API)
   * \param mask the pins to actually set up.
   * \param mode the mode for the pins (see Fix_mode).
   * \param outvalue the value if configured as output.
   */
  virtual void multi_setup(Pin_slice const &mask, unsigned mode, unsigned outvalue = 0) = 0;

  /**
   * \brief Configure multiple pads at once (platform specific API).
   * \param mask the pads to configure.
   * \param func the platform-specific sub-function to configure.
   * \param value the platform-specific value to set for the sub-function.
   * \see config_pad()
   */
  virtual void multi_config_pad(Pin_slice const &mask, unsigned func, unsigned value) = 0;

  /**
   * \brief Set the value for multiple output pins at once.
   * \param mask the pins that shall actually be set to the given values.
   * \param data the bit-wise value for each pin to set (according to mask).
   */
  virtual void multi_set(Pin_slice const &mask, unsigned data) = 0;

  /**
   * \brief Get the value for all pins at once.
   */
  virtual unsigned multi_get(unsigned offset) = 0;


  void output(unsigned pin, int value)
  { setup(pin, Output, value); }

  void input(unsigned pin)
  { setup(pin, Input); }

  void multi_output(Pin_slice const &mask, unsigned data)
  { multi_setup(mask, mask.mask, data); }

  void multi_input(Pin_slice const &mask)
  { multi_setup(mask, ~mask.mask); }
};

}

#else

typedef enum {
  false = 0,
  true = 1
} bool;

/// Modes used for setup()
enum Fix_mode
{
  Fix_input  = L4VBUS_GPIO_SETUP_INPUT,  ///< Configure as input pin
  Fix_output = L4VBUS_GPIO_SETUP_OUTPUT, ///< Configure as output pin
  Fix_irq    = L4VBUS_GPIO_SETUP_IRQ,    ///< Configure as IRQ source
};

/// Modes used for pull up / pull down (config_pull())
enum Pull_mode
{
  Pull_none  = L4VBUS_GPIO_PIN_PULL_NONE,  ///< No pull up or pull down resistors
  Pull_up    = L4VBUS_GPIO_PIN_PULL_UP,    ///< Enable pull up resistor
  Pull_down  = L4VBUS_GPIO_PIN_PULL_DOWN,  ///< Enable pull down resistor
};

// Generic function options
enum Function_levels
{
  Function_gpio,
  Function_alt,
};

#endif /* __cplusplus */
