/*
 * GPIO driver for Ingenic JZ4730.
 * (See below for additional copyright and licensing notices.)
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 *
 *
 * Subject to other copyrights, being derived from the bcm2835.cc and
 * omap.cc GPIO driver implementations.
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */

#include <l4/sys/icu.h>
#include <l4/util/util.h>
#include <l4/devices/hw_mmio_register_block.h>

#include "gpio-jz4730.h"

// GPIO register offsets (x in A..D).

enum Regs
{
  Port_data           = 0x000,  // PxGPDR
  Port_direction      = 0x004,  // PxGPDIR
  Pull_enable         = 0x00c,  // PxGPPUR
  Port_function_lower = 0x010,  // PxGPALR
  Port_function_upper = 0x014,  // PxGPAUR
  Irq_detect_lower    = 0x018,  // PxGPDLR
  Irq_detect_upper    = 0x01c,  // PxGPDUR
  Irq_enable          = 0x020,  // PxGPIER
  Irq_flag            = 0x028,  // PxGPFR
};



// Select the appropriate register and pin where two bits are assigned per pin,
// thus requiring two 32-bit registers to hold the configuration of 32 pins.

static
uint32_t
_select_bank_for_pin(uint32_t reg_upper, uint32_t reg_lower, uint8_t pin, uint8_t *pin_out)
{
  if (pin < 16)
  {
    *pin_out = pin;
    return reg_lower;
  }
  else
  {
    *pin_out = pin - 16;
    return reg_upper;
  }
}



// IRQ control for each GPIO pin.

Gpio_jz4730_irq_pin::Gpio_jz4730_irq_pin(unsigned pin, Hw::Register_block<32> const &regs)
: _pin(pin), _regs(regs)
{}

void
Gpio_jz4730_irq_pin::write_reg_pin(unsigned reg)
{
  // Write the pin bit to the register, setting or clearing the pin
  // depending on the register chosen.

  _regs[reg] = _pin_bit(_pin);
}

void
Gpio_jz4730_irq_pin::clear_reg_pin(unsigned reg)
{
  // Clear the pin bit in the register.

  _regs[reg] = _regs[reg] & ~(_pin_bit(_pin));
}

void
Gpio_jz4730_irq_pin::set_reg_pin(unsigned reg)
{
  // Set the pin bit in the register.

  _regs[reg] = _regs[reg] | _pin_bit(_pin);
}

// Clear the pin bits in the appropriate register.

void
Gpio_jz4730_irq_pin::clear_reg_pins(uint32_t reg_upper, uint32_t reg_lower)
{
  uint8_t pin_out;
  uint32_t reg = _select_bank_for_pin(reg_upper, reg_lower, _pin, &pin_out);

  _regs[reg] = _regs[reg] & ~(3 << (pin_out << 1));
}

// Clear and set the given value in the pin bits of the appropriate register.

void
Gpio_jz4730_irq_pin::set_reg_pins(uint32_t reg_upper, uint32_t reg_lower, uint8_t value)
{
  uint8_t pin_out;
  uint32_t reg = _select_bank_for_pin(reg_upper, reg_lower, _pin, &pin_out);

  _regs[reg] = (_regs[reg] & ~(3 << (pin_out << 1))) | (value << (pin_out << 1));
}

void Gpio_jz4730_irq_pin::do_mask()
{
  // Set the interrupt bit in the PxGPIER register.

  clear_reg_pin(Irq_enable);
}

void Gpio_jz4730_irq_pin::do_unmask()
{
  // Set the interrupt bit in the PxGPIER register, first also clearing the
  // flag bit in the PxGPFR register to allow interrupts to be delivered.

  clear_reg_pin(Irq_flag);
  set_reg_pin(Irq_enable);
}

bool Gpio_jz4730_irq_pin::do_set_mode(unsigned mode)
{
  // Standard comment found for this method:
  // this operation touches multiple mmio registers and is thus
  // not atomic, that's why we first mask the IRQ and if it was
  // enabled we unmask it after we have changed the mode

  if (enabled())
    do_mask();

  // Do the PxGPDUR/PxGPDLR configuration.

  switch(mode)
  {
    case L4_IRQ_F_LEVEL_HIGH:
      set_reg_pins(Irq_detect_upper, Irq_detect_lower, 1);
      break;
    case L4_IRQ_F_LEVEL_LOW:
      set_reg_pins(Irq_detect_upper, Irq_detect_lower, 0);
      break;
    case L4_IRQ_F_POS_EDGE:
      set_reg_pins(Irq_detect_upper, Irq_detect_lower, 3);
      break;
    case L4_IRQ_F_NEG_EDGE:
      set_reg_pins(Irq_detect_upper, Irq_detect_lower, 2);
      break;

    default:
      return false;
  }

  // Do the PxGPDIR, PxGPAUR/PxGPALR configuration.

  clear_reg_pin(Port_direction);
  clear_reg_pins(Port_function_upper, Port_function_lower);

  if (enabled())
    do_unmask();

  return true;
}

int Gpio_jz4730_irq_pin::clear()
{
  // Obtain the flag status for the pin, clearing it if set.

  l4_uint32_t e = _regs[Irq_flag] & (1UL << _pin);
  if (e)
    clear_reg_pin(Irq_flag);

  return (e >> _pin);
}

bool Gpio_jz4730_irq_pin::enabled()
{
  return true;
}



// Return two bitmaps for the given bitmap and value.

static
void
_get_bitmaps(uint32_t bitmap, uint8_t value, uint32_t *upper, uint32_t *lower)
{
  uint32_t mask = 0x80000000;

  *upper = 0; *lower = 0;

  while (mask != 0x8000)
  {
    *upper = (*upper << 2) | (bitmap & mask ? value : 0);
    mask >>= 1;
  }

  while (mask != 0)
  {
    *lower = (*lower << 2) | (bitmap & mask ? value : 0);
    mask >>= 1;
  }
}



// Initialise the GPIO controller.

Gpio_jz4730_chip::Gpio_jz4730_chip(l4_addr_t start, l4_addr_t end,
                                   unsigned nr_pins)
: _start(start), _end(end),
  _nr_pins(nr_pins)
{
  _regs = new Hw::Mmio_register_block<32>(_start);
}

// Return the value of a pin.

int
Gpio_jz4730_chip::get(unsigned pin)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  l4_uint32_t val = _regs[Port_data];
  return (val >> _pin_shift(pin)) & 1;
}

// Return multiple pin values.

unsigned
Gpio_jz4730_chip::multi_get(unsigned offset)
{
  _reg_offset_check(offset);
  return _regs[Port_data];
}

// Set the value of a pin.

void
Gpio_jz4730_chip::set(unsigned pin, int value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  if (value)
    _regs[Port_data] = _regs[Port_data] | _pin_bit(pin);
  else
    _regs[Port_data] = _regs[Port_data] & ~_pin_bit(pin);
}

// Set multiple pin values.

void
Gpio_jz4730_chip::multi_set(Pin_slice const &mask, unsigned data)
{
  _reg_offset_check(mask.offset);
  _regs[Port_data] = (_regs[Port_data] & ~(mask.mask)) | data;
}

// Set a pin up with the given mode and value (if appropriate).

void
Gpio_jz4730_chip::setup(unsigned pin, unsigned mode, int value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  config(pin, mode);

  if (mode == Output)
    set(pin, value);
}

// Configuration of a pin using the generic input/output/IRQ mode.

void
Gpio_jz4730_chip::config(unsigned pin, unsigned mode)
{
  _config(_pin_bit(pin), mode);
}

void
Gpio_jz4730_chip::_config(unsigned bitmap, unsigned mode)
{
  uint32_t upper_mask, lower_mask;
  unsigned bitmap_values = bitmap;

  switch (mode)
  {
    case Input:
    case Irq:
      // Clear the direction flags below.
      bitmap_values = 0;
      // Fall through to the next case to complete the operation.

    case Output:
      // Clear the flags if set immediately above; otherwise, set them.
      _regs[Port_direction] = (_regs[Port_direction] & ~bitmap) | bitmap_values;

      // Clear the port function for the bits.
      _get_bitmaps(bitmap, 3, &upper_mask, &lower_mask);
      _regs[Port_function_upper] = (_regs[Port_function_upper] & ~upper_mask);
      _regs[Port_function_lower] = (_regs[Port_function_lower] & ~lower_mask);
      break;

    default:
      break;
  }
}

// Pull-up configuration for a pin.

void
Gpio_jz4730_chip::config_pull(unsigned pin, unsigned mode)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  _config_pull(_pin_bit(pin), mode);
}

void
Gpio_jz4730_chip::_config_pull(unsigned bitmap, unsigned mode)
{
  switch (mode)
  {
    case Pull_none:
      _regs[Pull_enable] = _regs[Pull_enable] & ~bitmap;
      break;
    case Pull_up:
      _regs[Pull_enable] = _regs[Pull_enable] | bitmap;
      break;
    default:
      // Invalid pull-up/down mode for pin.
      throw -L4_EINVAL;
  }
}

// Pin function configuration.

void
Gpio_jz4730_chip::config_pad(unsigned pin, unsigned func, unsigned value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  _config_pad(_pin_bit(pin), func, value);
}

void
Gpio_jz4730_chip::_config_pad(unsigned bitmap, unsigned func, unsigned value)
{
  uint32_t upper_mask = 0, lower_mask = 0, upper = 0, lower = 0;

  if (value > 3)
    throw -L4_EINVAL;

  switch (func)
  {
    case Hw::Gpio_chip::Function_alt:
      _get_bitmaps(bitmap, value, &upper, &lower);
      // Fall through to the next case to complete the operation.

    case Hw::Gpio_chip::Function_gpio:
      _get_bitmaps(bitmap, 3, &upper_mask, &lower_mask);
      _regs[Port_function_upper] = (_regs[Port_function_upper] & ~upper_mask) | upper;
      _regs[Port_function_lower] = (_regs[Port_function_lower] & ~lower_mask) | lower;
      break;

    default:
      throw -L4_EINVAL;
  }
}

// Obtain a pin's configuration from a register in the supplied value.

void
Gpio_jz4730_chip::config_get(unsigned pin, unsigned reg, unsigned *value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  *value = (_regs[reg] >> _pin_shift(pin)) & 1;
}

// Obtain an IRQ abstraction for a pin.

Hw::Gpio_irq_pin *
Gpio_jz4730_chip::get_irq(unsigned pin)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  return new Gpio_jz4730_irq_pin(pin, _regs);
}

// Pull-up function configuration for multiple pins.

void
Gpio_jz4730_chip::multi_config_pull(Pin_slice const &mask, unsigned mode)
{
  _config_pull(mask.mask << mask.offset, mode);
}

// Pin function configuration for multiple pins.

void
Gpio_jz4730_chip::multi_config_pad(Pin_slice const &mask, unsigned func, unsigned val)
{
  _config_pad(mask.mask << mask.offset, func, val);
}

// Set up multiple pins with the given mode.

void
Gpio_jz4730_chip::multi_setup(Pin_slice const &mask, unsigned mode, unsigned outvalues)
{
  _config(mask.mask << mask.offset, mode);

  if (mode == Output)
    multi_set(mask, outvalues);
}



// C language interface functions.

void *jz4730_gpio_init(l4_addr_t start, l4_addr_t end, unsigned pins)
{
  return (void *) new Gpio_jz4730_chip(start, end, pins);
}

void jz4730_gpio_setup(void *gpio, unsigned pin, unsigned mode, int value)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->setup(pin, mode, value);
}

void jz4730_gpio_config_pull(void *gpio, unsigned pin, unsigned mode)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->config_pull(pin, mode);
}

void jz4730_gpio_config_pad(void *gpio, unsigned pin, unsigned func, unsigned value)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->config_pad(pin, func, value);
}

void jz4730_gpio_config_get(void *gpio, unsigned pin, unsigned reg, unsigned *value)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->config_get(pin, reg, value);
}

void jz4730_gpio_multi_setup(void *gpio, Pin_slice const *mask, unsigned mode, unsigned outvalues)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->multi_setup(*mask, mode, outvalues);
}

void jz4730_gpio_multi_config_pull(void *gpio, Pin_slice const *mask, unsigned mode)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->multi_config_pull(*mask, mode);
}

void jz4730_gpio_multi_config_pad(void *gpio, Pin_slice const *mask, unsigned func, unsigned value)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->multi_config_pad(*mask, func, value);
}

void jz4730_gpio_multi_set(void *gpio, Pin_slice const *mask, unsigned data)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->multi_set(*mask, data);
}

unsigned jz4730_gpio_multi_get(void *gpio, unsigned offset)
{
  return static_cast<Gpio_jz4730_chip *>(gpio)->multi_get(offset);
}

int jz4730_gpio_get(void *gpio, unsigned pin)
{
  return static_cast<Gpio_jz4730_chip *>(gpio)->get(pin);
}

void jz4730_gpio_set(void *gpio, unsigned pin, int value)
{
  static_cast<Gpio_jz4730_chip *>(gpio)->set(pin, value);
}

void *jz4730_gpio_get_irq(void *gpio, unsigned pin)
{
  return (void *) static_cast<Gpio_jz4730_chip *>(gpio)->get_irq(pin);
}

bool jz4730_gpio_irq_set_mode(void *gpio_irq, unsigned mode)
{
  return static_cast<Hw::Gpio_irq_pin *>(gpio_irq)->do_set_mode(mode);
}
