/*
 * GPIO driver for Ingenic JZ4740.
 * (See below for additional copyright and licensing notices.)
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 *
 *
 * Subject to other copyrights, being derived from the bcm2835.cc and
 * omap.cc GPIO driver implementations.
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */

#include <l4/sys/icu.h>
#include <l4/util/util.h>
#include <l4/devices/hw_mmio_register_block.h>

#include "gpio-jz4740.h"

// GPIO register offsets (x in A..D).

enum Regs
{
  Pin_level           = 0x000,  // PxPIN (read-only)
  Port_data           = 0x010,  // PxDAT (read-only)
  Port_data_set       = 0x014,  // PxDATS
  Port_data_clear     = 0x018,  // PxDATC
  Irq_mask            = 0x020,  // PxIM (read-only)
  Irq_mask_set        = 0x024,  // PxIMS
  Irq_mask_clear      = 0x028,  // PxIMC
  Pull_disable        = 0x030,  // PxPE (read-only)
  Pull_disable_set    = 0x034,  // PxPES
  Pull_disable_clear  = 0x038,  // PxPEC
  Port_function       = 0x040,  // PxFUN (read-only)
  Port_function_set   = 0x044,  // PxFUNS
  Port_function_clear = 0x048,  // PxFUNC
  Port_select         = 0x050,  // PxSEL (read-only)
  Port_select_set     = 0x054,  // PxSELS
  Port_select_clear   = 0x058,  // PxSELC
  Port_dir            = 0x060,  // PxDIR (read-only)
  Port_dir_set        = 0x064,  // PxDIRS
  Port_dir_clear      = 0x068,  // PxDIRC
  Port_trigger        = 0x070,  // PxTRG (read-only)
  Port_trigger_set    = 0x074,  // PxTRGS
  Port_trigger_clear  = 0x078,  // PxTRGC
  Irq_flag            = 0x080,  // PxFLG (read-only)
  Irq_flag_clear      = 0x084,  // PxFLGC
};



// IRQ control for each GPIO pin.

Gpio_jz4740_irq_pin::Gpio_jz4740_irq_pin(unsigned pin, Hw::Register_block<32> const &regs)
: _pin(pin), _regs(regs)
{}

void
Gpio_jz4740_irq_pin::write_reg_pin(unsigned reg)
{
  // Write the pin bit to the register, setting or clearing the pin
  // depending on the register chosen.

  _regs[reg] = _pin_bit(_pin);
}

void Gpio_jz4740_irq_pin::do_mask()
{
  // Set the interrupt bit in the PxIM register.

  write_reg_pin(Irq_mask_set);
}

void Gpio_jz4740_irq_pin::do_unmask()
{
  // Clear the interrupt bit in the PxIM register, first also clearing the
  // flag bit in the PxFLG register to allow interrupts to be delivered.

  write_reg_pin(Irq_flag_clear);
  write_reg_pin(Irq_mask_clear);
}

bool Gpio_jz4740_irq_pin::do_set_mode(unsigned mode)
{
  // Standard comment found for this method:
  // this operation touches multiple mmio registers and is thus
  // not atomic, that's why we first mask the IRQ and if it was
  // enabled we unmask it after we have changed the mode

  if (enabled())
    do_mask();

  // Do the PxTRG, PxFUN, PxSEL and PxDIR configuration.

  switch(mode)
  {
    case L4_IRQ_F_LEVEL_HIGH:
      write_reg_pin(Port_trigger_clear);
      write_reg_pin(Port_function_clear);
      write_reg_pin(Port_select_set);
      write_reg_pin(Port_dir_set);
      break;
    case L4_IRQ_F_LEVEL_LOW:
      write_reg_pin(Port_trigger_clear);
      write_reg_pin(Port_function_clear);
      write_reg_pin(Port_select_set);
      write_reg_pin(Port_dir_clear);
      break;
    case L4_IRQ_F_POS_EDGE:
      write_reg_pin(Port_trigger_set);
      write_reg_pin(Port_function_clear);
      write_reg_pin(Port_select_set);
      write_reg_pin(Port_dir_set);
      break;
    case L4_IRQ_F_NEG_EDGE:
      write_reg_pin(Port_trigger_set);
      write_reg_pin(Port_function_clear);
      write_reg_pin(Port_select_set);
      write_reg_pin(Port_dir_clear);
      break;

    default:
      return false;
  }

  if (enabled())
    do_unmask();

  return true;
}

int Gpio_jz4740_irq_pin::clear()
{
  // Obtain the flag status for the pin, clearing it if set.

  l4_uint32_t e = _regs[Irq_flag] & (1UL << _pin);
  if (e)
    _regs[Irq_flag_clear] = e;

  return (e >> _pin);
}

bool Gpio_jz4740_irq_pin::enabled()
{
  return true;
}



// Initialise the GPIO controller.

Gpio_jz4740_chip::Gpio_jz4740_chip(l4_addr_t start, l4_addr_t end,
                                   unsigned nr_pins)
: _start(start), _end(end),
  _nr_pins(nr_pins)
{
  _regs = new Hw::Mmio_register_block<32>(_start);
}

// Return the value of a pin.

int
Gpio_jz4740_chip::get(unsigned pin)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  l4_uint32_t val = _regs[Pin_level];
  return (val >> _pin_shift(pin)) & 1;
}

// Return multiple pin values.

unsigned
Gpio_jz4740_chip::multi_get(unsigned offset)
{
  _reg_offset_check(offset);
  return _regs[Pin_level];
}

// Set the value of a pin.

void
Gpio_jz4740_chip::set(unsigned pin, int value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  l4_uint32_t reg_set = value ? Port_data_set : Port_data_clear;
  _regs[reg_set] = _pin_bit(pin);
}

// Set multiple pin values.

void
Gpio_jz4740_chip::multi_set(Pin_slice const &mask, unsigned data)
{
  _reg_offset_check(mask.offset);
  if (mask.mask & data)
    _regs[Port_data_set] = (mask.mask & data);
  if (mask.mask & ~data)
    _regs[Port_data_clear] = (mask.mask & ~data);
}

// Set a pin up with the given mode and value (if appropriate).

void
Gpio_jz4740_chip::setup(unsigned pin, unsigned mode, int value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  config(pin, mode);

  if (mode == Output)
    set(pin, value);
}

// Configuration of a pin using the generic input/output/IRQ mode.

void
Gpio_jz4740_chip::config(unsigned pin, unsigned mode)
{
  _config(_pin_bit(pin), mode);
}

void
Gpio_jz4740_chip::_config(unsigned bitmap, unsigned mode)
{
  switch (mode)
  {
    case Input:
      _regs[Port_function_clear] = bitmap;
      _regs[Port_select_clear] = bitmap;
      _regs[Port_dir_clear] = bitmap;
      break;
    case Output:
      _regs[Port_function_clear] = bitmap;
      _regs[Port_select_clear] = bitmap;
      _regs[Port_dir_set] = bitmap;
      break;
    case Irq:
      _regs[Port_function_clear] = bitmap;
      _regs[Port_select_set] = bitmap;
      // The direction depends on the actual trigger mode.
      break;
    default:
      break;
  }
}

// Pull-up configuration for a pin.

void
Gpio_jz4740_chip::config_pull(unsigned pin, unsigned mode)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  _config_pull(_pin_bit(pin), mode);
}

void
Gpio_jz4740_chip::_config_pull(unsigned bitmap, unsigned mode)
{
  switch (mode)
  {
    case Pull_none:
      _regs[Pull_disable_set] = bitmap;
      break;
    case Pull_up:
      _regs[Pull_disable_clear] = bitmap;
      break;
    default:
      // Invalid pull-up/down mode for pin.
      throw -L4_EINVAL;
  }
}

// Pin function configuration.

void
Gpio_jz4740_chip::config_pad(unsigned pin, unsigned func, unsigned value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  _config_pad(_pin_bit(pin), func, value);
}

void
Gpio_jz4740_chip::_config_pad(unsigned bitmap, unsigned func, unsigned value)
{
  if (value > 1)
    throw -L4_EINVAL;

  switch (func)
  {
    case Hw::Gpio_chip::Function_gpio:
      _regs[Port_function_clear] = bitmap;
      break;

    // Support two different device functions.

    case Hw::Gpio_chip::Function_alt:
      _regs[Port_function_set] = bitmap;
      _regs[value ? Port_select_set : Port_select_clear] = bitmap;
      break;
    default:
      throw -L4_EINVAL;
  }
}

// Obtain a pin's configuration from a register in the supplied value.

void
Gpio_jz4740_chip::config_get(unsigned pin, unsigned reg, unsigned *value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  *value = (_regs[reg] >> _pin_shift(pin)) & 1;
}

// Obtain an IRQ abstraction for a pin.

Hw::Gpio_irq_pin *
Gpio_jz4740_chip::get_irq(unsigned pin)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  return new Gpio_jz4740_irq_pin(pin, _regs);
}

// Pull-up function configuration for multiple pins.

void
Gpio_jz4740_chip::multi_config_pull(Pin_slice const &mask, unsigned mode)
{
  _config_pull(mask.mask << mask.offset, mode);
}

// Pin function configuration for multiple pins.

void
Gpio_jz4740_chip::multi_config_pad(Pin_slice const &mask, unsigned func, unsigned val)
{
  _config_pad(mask.mask << mask.offset, func, val);
}

// Set up multiple pins with the given mode.

void
Gpio_jz4740_chip::multi_setup(Pin_slice const &mask, unsigned mode, unsigned outvalues)
{
  _config(mask.mask << mask.offset, mode);

  if (mode == Output)
    multi_set(mask, outvalues);
}



// C language interface functions.

void *jz4740_gpio_init(l4_addr_t start, l4_addr_t end, unsigned pins)
{
  return (void *) new Gpio_jz4740_chip(start, end, pins);
}

void jz4740_gpio_setup(void *gpio, unsigned pin, unsigned mode, int value)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->setup(pin, mode, value);
}

void jz4740_gpio_config_pull(void *gpio, unsigned pin, unsigned mode)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->config_pull(pin, mode);
}

void jz4740_gpio_config_pad(void *gpio, unsigned pin, unsigned func, unsigned value)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->config_pad(pin, func, value);
}

void jz4740_gpio_config_get(void *gpio, unsigned pin, unsigned reg, unsigned *value)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->config_get(pin, reg, value);
}

void jz4740_gpio_multi_setup(void *gpio, Pin_slice const *mask, unsigned mode, unsigned outvalues)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->multi_setup(*mask, mode, outvalues);
}

void jz4740_gpio_multi_config_pull(void *gpio, Pin_slice const *mask, unsigned mode)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->multi_config_pull(*mask, mode);
}

void jz4740_gpio_multi_config_pad(void *gpio, Pin_slice const *mask, unsigned func, unsigned value)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->multi_config_pad(*mask, func, value);
}

void jz4740_gpio_multi_set(void *gpio, Pin_slice const *mask, unsigned data)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->multi_set(*mask, data);
}

unsigned jz4740_gpio_multi_get(void *gpio, unsigned offset)
{
  return static_cast<Gpio_jz4740_chip *>(gpio)->multi_get(offset);
}

int jz4740_gpio_get(void *gpio, unsigned pin)
{
  return static_cast<Gpio_jz4740_chip *>(gpio)->get(pin);
}

void jz4740_gpio_set(void *gpio, unsigned pin, int value)
{
  static_cast<Gpio_jz4740_chip *>(gpio)->set(pin, value);
}

void *jz4740_gpio_get_irq(void *gpio, unsigned pin)
{
  return (void *) static_cast<Gpio_jz4740_chip *>(gpio)->get_irq(pin);
}

bool jz4740_gpio_irq_set_mode(void *gpio_irq, unsigned mode)
{
  return static_cast<Hw::Gpio_irq_pin *>(gpio_irq)->do_set_mode(mode);
}
