/*
 * GPIO driver for Ingenic JZ4780.
 * (See below for additional copyright and licensing notices.)
 *
 * Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 *
 *
 * Subject to other copyrights, being derived from the bcm2835.cc and
 * omap.cc GPIO driver implementations.
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */

#include <l4/sys/icu.h>
#include <l4/util/util.h>
#include <l4/devices/hw_mmio_register_block.h>

#include "gpio-jz4780.h"

/*
GPIO register offsets (x in A..F).

Register summary:

PxINT         0 (function/GPIO)                                             1 (interrupt)
PxMSK         0 (function)                        1 (GPIO)                  0 (IRQ enable)/1 (IRQ disable)
PxPAT1        0 (function 0/1)  1 (function 2/3)  0 (output)    1 (input)   0 (level trigger) 1 (edge trigger)
PxPAT0        0 (function 0)    0 (function 2)    0 (output value 0)        0 (low level)     0 (falling edge)
              1 (function 1)    1 (function 3)    1 (output value 1)        1 (high level)    1 (rising edge)
*/

enum Regs
{ 
  Pin_level           = 0x000,  // PxPIN (read-only)
  
  Port_int            = 0x010,  // PxINT
  Port_int_set        = 0x014,  // PxINTS
  Port_int_clear      = 0x018,  // PxINTC
  
  Irq_mask            = 0x020,  // PxMSK (for PxINT == 1)
  Irq_mask_set        = 0x024,  // PxMSKS
  Irq_mask_clear      = 0x028,  // PxMSKC
  Port_gpio           = 0x020,  // PxMSK (for PxINT == 0)
  Port_gpio_set       = 0x024,  // PxMSKS
  Port_gpio_clear     = 0x028,  // PxMSKC
  
  Port_trigger        = 0x030,  // PxPAT1 (for PxINT == 1)
  Port_trigger_set    = 0x034,  // PxPAT1S
  Port_trigger_clear  = 0x038,  // PxPAT1C
  Port_dir            = 0x030,  // PxPAT1 (for PxINT == 0, PxMSK == 1)
  Port_dir_set        = 0x034,  // PxPAT1S
  Port_dir_clear      = 0x038,  // PxPAT1C
  Port_group1         = 0x030,  // PxPAT1 (for PxINT == 0, PxMSK == 0)
  Port_group1_set     = 0x034,  // PxPAT1S
  Port_group1_clear   = 0x038,  // PxPAT1C
  
  Port_level          = 0x040,  // PxPAT0 (for PxINT == 1)
  Port_level_set      = 0x044,  // PxPAT0S
  Port_level_clear    = 0x048,  // PxPAT0C
  Port_data           = 0x040,  // PxPAT0 (for PxINT == 0, PxMSK == 1, PxPAT1 == 0)
  Port_data_set       = 0x044,  // PxPAT0S
  Port_data_clear     = 0x048,  // PxPAT0C
  Port_group0         = 0x040,  // PxPAT0 (for PxINT == 0, PxMSK == 0)
  Port_group0_set     = 0x044,  // PxPAT0S
  Port_group0_clear   = 0x048,  // PxPAT0C
  
  Irq_flag            = 0x050,  // PxFLG (read-only)
  Irq_flag_clear      = 0x058,  // PxFLGC
  
  Pull_disable        = 0x070,  // PxPE
  Pull_disable_set    = 0x074,  // PxPES
  Pull_disable_clear  = 0x078,  // PxPEC
};



// IRQ control for each GPIO pin.

Gpio_jz4780_irq_pin::Gpio_jz4780_irq_pin(unsigned pin, Hw::Register_block<32> const &regs)
: _pin(pin), _regs(regs)
{}

void
Gpio_jz4780_irq_pin::write_reg_pin(unsigned reg)
{
  // Write the pin bit to the register, setting or clearing the pin
  // depending on the register chosen.

  _regs[reg] = _pin_bit(_pin);
}

void Gpio_jz4780_irq_pin::do_mask()
{
  // Set the interrupt bit in the PxIM register.

  write_reg_pin(Irq_mask_set);
}

void Gpio_jz4780_irq_pin::do_unmask()
{
  // Clear the interrupt bit in the PxIM register, first also clearing the
  // flag bit in the PxFLG register to allow interrupts to be delivered.

  write_reg_pin(Irq_flag_clear);
  write_reg_pin(Irq_mask_clear);
}

bool Gpio_jz4780_irq_pin::do_set_mode(unsigned mode)
{
  // Standard comment found for this method:
  // this operation touches multiple mmio registers and is thus
  // not atomic, that's why we first mask the IRQ and if it was
  // enabled we unmask it after we have changed the mode

  if (enabled())
    do_mask();

  // Do the PxINT, PxPAT1 and PxPAT0 configuration.

  switch(mode)
  {
    case L4_IRQ_F_LEVEL_HIGH:
      write_reg_pin(Port_int_set);
      write_reg_pin(Port_trigger_clear);
      write_reg_pin(Port_level_set);
      break;
    case L4_IRQ_F_LEVEL_LOW:
      write_reg_pin(Port_int_set);
      write_reg_pin(Port_trigger_clear);
      write_reg_pin(Port_level_clear);
      break;
    case L4_IRQ_F_POS_EDGE:
      write_reg_pin(Port_int_set);
      write_reg_pin(Port_trigger_set);
      write_reg_pin(Port_level_set);
      break;
    case L4_IRQ_F_NEG_EDGE:
      write_reg_pin(Port_int_set);
      write_reg_pin(Port_trigger_set);
      write_reg_pin(Port_level_clear);
      break;

    default:
      return false;
  }

  if (enabled())
    do_unmask();

  return true;
}

int Gpio_jz4780_irq_pin::clear()
{
  // Obtain the flag status for the pin, clearing it if set.

  l4_uint32_t e = _regs[Irq_flag] & (1UL << _pin);
  if (e)
    _regs[Irq_flag_clear] = e;

  return (e >> _pin);
}

bool Gpio_jz4780_irq_pin::enabled()
{
  return true;
}



// Initialise the GPIO controller.

Gpio_jz4780_chip::Gpio_jz4780_chip(l4_addr_t start, l4_addr_t end,
                                   unsigned nr_pins,
                                   l4_uint32_t pull_ups, l4_uint32_t pull_downs)
: _start(start), _end(end),
  _nr_pins(nr_pins),
  _pull_ups(pull_ups), _pull_downs(pull_downs)
{
  _regs = new Hw::Mmio_register_block<32>(_start);
}

// Return the value of a pin.

int
Gpio_jz4780_chip::get(unsigned pin)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  l4_uint32_t val = _regs[Pin_level];
  return (val >> _pin_shift(pin)) & 1;
}

// Return multiple pin values.

unsigned
Gpio_jz4780_chip::multi_get(unsigned offset)
{
  _reg_offset_check(offset);
  return _regs[Pin_level];
}

// Set the value of a pin.

void
Gpio_jz4780_chip::set(unsigned pin, int value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  l4_uint32_t reg_set = value ? Port_data_set : Port_data_clear;
  _regs[reg_set] = _pin_bit(pin);
}

// Set multiple pin values.

void
Gpio_jz4780_chip::multi_set(Pin_slice const &mask, unsigned data)
{
  _reg_offset_check(mask.offset);
  if (mask.mask & data)
    _regs[Port_data_set] = (mask.mask & data);
  if (mask.mask & ~data)
    _regs[Port_data_clear] = (mask.mask & ~data);
}

// Set a pin up with the given mode and value (if appropriate).

void
Gpio_jz4780_chip::setup(unsigned pin, unsigned mode, int value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  config(pin, mode);

  if (mode == Output)
    set(pin, value);
}

// Configuration of a pin using the generic input/output/IRQ mode.

void
Gpio_jz4780_chip::config(unsigned pin, unsigned mode)
{
  switch (mode)
  {
    case Input:
      _regs[Port_int_clear] = _pin_bit(pin);
      _regs[Port_gpio_set] = _pin_bit(pin);
      _regs[Port_dir_set] = _pin_bit(pin);
      break;
    case Output:
      _regs[Port_int_clear] = _pin_bit(pin);
      _regs[Port_gpio_set] = _pin_bit(pin);
      _regs[Port_dir_clear] = _pin_bit(pin);
      break;
    case Irq:
      _regs[Port_int_set] = _pin_bit(pin);
      // Other details depend on the actual trigger mode.
      break;
    default:
      break;
  }
}

// Pull-up/down configuration for a pin.

void
Gpio_jz4780_chip::config_pull(unsigned pin, unsigned mode)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  switch (mode)
  {
    case Pull_none:
      _regs[Pull_disable_set] = _pin_bit(pin);
      break;
    case Pull_down:
      if (_pin_bit(pin) & _pull_downs)
        _regs[Pull_disable_clear] = _pin_bit(pin);
      break;
    case Pull_up:
      if (_pin_bit(pin) & _pull_ups)
      _regs[Pull_disable_clear] = _pin_bit(pin);
      break;
    default:
      // Invalid pull-up/down mode for pin.
      throw -L4_EINVAL;
  }
}

// Pin function configuration.

void
Gpio_jz4780_chip::config_pad(unsigned pin, unsigned func, unsigned value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  if (value > 1)
    throw -L4_EINVAL;

  switch (func)
  {
    // Support two different outputs.

    case Hw::Gpio_chip::Function_gpio:
      _regs[Port_int_clear] = _pin_bit(pin);
      _regs[Port_gpio_set] = _pin_bit(pin);
      _regs[value & 1 ? Port_data_set : Port_data_clear] = _pin_bit(pin);
      break;

    // Support four different device functions.

    case Hw::Gpio_chip::Function_alt:
      _regs[Port_int_clear] = _pin_bit(pin);
      _regs[Port_gpio_clear] = _pin_bit(pin);
      _regs[value & 2 ? Port_group1_set : Port_group1_clear] = _pin_bit(pin);
      _regs[value & 1 ? Port_group0_set : Port_group0_clear] = _pin_bit(pin);
      break;
    default:
      throw -L4_EINVAL;
  }
}

// Obtain a pin's configuration from a register in the supplied value.

void
Gpio_jz4780_chip::config_get(unsigned pin, unsigned reg, unsigned *value)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  *value = (_regs[reg] >> _pin_shift(pin)) & 1;
}

// Obtain an IRQ abstraction for a pin.

Hw::Gpio_irq_pin *
Gpio_jz4780_chip::get_irq(unsigned pin)
{
  if (pin >= _nr_pins)
    throw -L4_EINVAL;

  return new Gpio_jz4780_irq_pin(pin, _regs);
}

// Pin function configuration for multiple pins.

void
Gpio_jz4780_chip::multi_config_pad(Pin_slice const &mask, unsigned func, unsigned val)
{
  unsigned m = mask.mask;
  for (unsigned pin = mask.offset; pin < _nr_pins; ++pin, m >>= 1)
    if (m & 1)
      config_pad(pin, func, val);
}

// Set up multiple pins with the given mode.

void
Gpio_jz4780_chip::multi_setup(Pin_slice const &mask, unsigned mode, unsigned outvalues)
{
  unsigned m = mask.mask;
  for (unsigned pin = mask.offset; pin < _nr_pins; ++pin, m >>= 1, outvalues >>= 1)
    if (m & 1)
      setup(pin, mode, outvalues & 1);
}



// C language interface functions.

void *jz4780_gpio_init(l4_addr_t start, l4_addr_t end, unsigned pins,
                       l4_uint32_t pull_ups, l4_uint32_t pull_downs)
{
  return (void *) new Gpio_jz4780_chip(start, end, pins, pull_ups, pull_downs);
}

void jz4780_gpio_setup(void *gpio, unsigned pin, unsigned mode, int value)
{
  static_cast<Gpio_jz4780_chip *>(gpio)->setup(pin, mode, value);
}

void jz4780_gpio_config_pull(void *gpio, unsigned pin, unsigned mode)
{
  static_cast<Gpio_jz4780_chip *>(gpio)->config_pull(pin, mode);
}

void jz4780_gpio_config_pad(void *gpio, unsigned pin, unsigned func, unsigned value)
{
  static_cast<Gpio_jz4780_chip *>(gpio)->config_pad(pin, func, value);
}

void jz4780_gpio_config_get(void *gpio, unsigned pin, unsigned reg, unsigned *value)
{
  static_cast<Gpio_jz4780_chip *>(gpio)->config_get(pin, reg, value);
}

void jz4780_gpio_multi_setup(void *gpio, Pin_slice const *mask, unsigned mode, unsigned outvalues)
{
  static_cast<Gpio_jz4780_chip *>(gpio)->multi_setup(*mask, mode, outvalues);
}

void jz4780_gpio_multi_config_pad(void *gpio, Pin_slice const *mask, unsigned func, unsigned value)
{
  static_cast<Gpio_jz4780_chip *>(gpio)->multi_config_pad(*mask, func, value);
}

void jz4780_gpio_multi_set(void *gpio, Pin_slice const *mask, unsigned data)
{
  static_cast<Gpio_jz4780_chip *>(gpio)->multi_set(*mask, data);
}

unsigned jz4780_gpio_multi_get(void *gpio, unsigned offset)
{
  return static_cast<Gpio_jz4780_chip *>(gpio)->multi_get(offset);
}

int jz4780_gpio_get(void *gpio, unsigned pin)
{
  return static_cast<Gpio_jz4780_chip *>(gpio)->get(pin);
}

void jz4780_gpio_set(void *gpio, unsigned pin, int value)
{
  static_cast<Gpio_jz4780_chip *>(gpio)->set(pin, value);
}

void *jz4780_gpio_get_irq(void *gpio, unsigned pin)
{
  return (void *) static_cast<Gpio_jz4780_chip *>(gpio)->get_irq(pin);
}

bool jz4780_gpio_irq_set_mode(void *gpio_irq, unsigned mode)
{
  return static_cast<Hw::Gpio_irq_pin *>(gpio_irq)->do_set_mode(mode);
}
