/*
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/i2c-jz4780.h>
#include <l4/devices/hw_mmio_register_block.h>

#include <l4/sys/icu.h>
#include <l4/util/util.h>
#include <sys/time.h>

#include <cstdio>

/*
I2C pins:

I2C0: PD30/SMB0_SDA,        PD31/SMB0_SCK
I2C1: PD30/SMB1_SDA,        PD31/SMB1_SCK (pulled up on CI20)
I2C2: PF16/SMB2_SDA,        PF17/SMB2_SCK
 PWM: PD10/SMB3_SDA,        PD11/SMB3_SCK
 PWM:  PE3/SMB4_SDA,         PE4/SMB4_SCK
I2C4: PE12/SMB4_SDA,        PE13/SMB4_SCK
HDMI: PF25/SMB4_SDA/DDCSDA, PF24/SMB4_SCK/DDCSCK

See: http://mipscreator.imgtec.com/CI20/hardware/board/ci20_jz4780_v2.0.pdf
*/

enum Regs
{ 
  Smb_control         = 0x000,  // SMBCON
  Smb_target_address  = 0x004,  // SMBTAR
  Smb_slave_address   = 0x008,  // SMBSAR
  Smb_data_command    = 0x010,  // SMBDC
  Std_high_count      = 0x014,  // SMBSHCNT
  Std_low_count       = 0x018,  // SMBSLCNT
  Fast_high_count     = 0x01c,  // SMBFHCNT
  Fast_low_count      = 0x020,  // SMBFLCNT
  Int_status          = 0x02c,  // SMBINTST (read-only)
  Int_mask            = 0x030,  // SMBINTM

  Rx_fifo_thold       = 0x038,  // SMBRXTL
  Tx_fifo_thold       = 0x03c,  // SMBTXTL
  Int_combined_clear  = 0x040,  // SMBCINT (read-only)
  Int_rx_uf_clear     = 0x044,  // SMBCRXUF (read-only)
  Int_rx_of_clear     = 0x048,  // SMBCRXOF (read-only)
  Int_tx_of_clear     = 0x04c,  // SMBCTXOF (read-only)
  Int_rd_req_clear    = 0x050,  // SMBCRXREQ (read-only)
  Int_tx_abort_clear  = 0x054,  // SMBCTXABT (read-only)
  Int_rx_done_clear   = 0x058,  // SMBCRXDN (read-only)
  Int_activity_clear  = 0x05c,  // SMBCACT (read-only)
  Int_stop_clear      = 0x060,  // SMBCSTP (read-only)
  Int_start_clear     = 0x064,  // SMBCSTT (read-only)
  Int_call_clear      = 0x068,  // SMBCGC (read-only)
  Smb_enable          = 0x06c,  // SMBENB
  Smb_status          = 0x070,  // SMBST (read-only)

  Tx_fifo_count       = 0x074,  // SMBTXFLR (read-only)
  Rx_fifo_count       = 0x078,  // SMBRXFLR (read-only)

  Trans_abort_status0 = 0x080,  // SMBABTSRC (read-only)
  Trans_abort_status1 = 0x084,  // ... (read-only)

  Smb_dma_ctrl        = 0x088,  // SMBDMACR
  Smb_trans_data_lvl  = 0x08c,  // SMBDMATDLR
  Smb_recv_data_lvl   = 0x090,  // SMBDMARDLR
  Smb_sda_setup_time  = 0x094,  // SMBSDASU
  Smb_ack_call        = 0x098,  // SMBACKGC

  Smb_enable_status   = 0x09c,  // SMBENBST (read-only)
  Smb_sda_hold_time   = 0x0d0,  // SMBSDAHD

  Smb_block_offset    = 0x1000
};

enum Smb_control_bits : unsigned
{
  Smb_no_stop_empty   = 0x80,   // STPHLD (no STP condition when queue empty)
  Smb_disable_slave   = 0x40,   // SLVDIS (slave disabled)
  Smb_enable_restart  = 0x20,   // REST
  Smb_enable_master   = 0x01,   // MD (master enabled)
  Smb_speed_bit       = 1,      // SPD
};

enum Smb_enable_bits : unsigned
{
  Smb_enable_enabled  = 0x01,   // SMBEN
};

enum Smb_status_bits : unsigned
{
  Smb_status_master_act = 0x20,   // MSTACT (master active)
  Smb_status_rx_nempty  = 0x08,   // RFNE (read queue not empty)
  Smb_status_tx_empty   = 0x04,   // TFE (write queue empty)
  Smb_status_tx_nfull   = 0x02,   // TFNF (write queue not full)
  Smb_status_active     = 0x01,   // ACT (device active as master or slave)
};

enum Smb_target_bits : unsigned
{
  Smb_target_7bits    = 0x7f,
};

enum Smb_hold_control_bits : unsigned
{
  Smb_hold_enable     = 0x100,  // HDENB
  Smb_hold_disable    = 0x000,  // HDENB
  Smb_hold_mask       = 0x1ff,
};

enum Smb_setup_control_bits : unsigned
{
  Smb_setup_mask      = 0x0ff,
};

enum Smb_command_bits : unsigned
{
  Smb_command_read    = 0x100,  // CMD
  Smb_command_write   = 0x000,  // CMD
};

enum Smb_fifo_bits : unsigned
{
  Smb_fifo_limit      = 16,
};

enum Int_bits : unsigned
{
  Int_call       = 0x800,       // IGC (general call received)
  Int_start      = 0x400,       // ISTT (start/restart condition occurred)
  Int_stop       = 0x200,       // ISTP (stop condition occurred)
  Int_activity   = 0x100,       // IACT (bus activity interrupt)
  Int_rx_done    = 0x080,       // RXDN (read from master device done)
  Int_tx_abort   = 0x040,       // TXABT (transmit abort)
  Int_rd_req     = 0x020,       // RDREQ (read request from master device)
  Int_tx_empty   = 0x010,       // TXEMP (threshold reached or passed)
  Int_tx_of      = 0x008,       // TXOF (overflow when writing to queue)
  Int_rx_full    = 0x004,       // RXFL (threshold reached or exceeded)
  Int_rx_of      = 0x002,       // RXOF (overflow from device)
  Int_rx_uf      = 0x001,       // RXUF (underflow when reading from queue)
};



// Initialise a channel.

I2c_jz4780_channel::I2c_jz4780_channel(l4_addr_t start,
                                       Cpm_jz4780_chip *cpm,
                                       uint32_t frequency) 
: _cpm(cpm), _frequency(frequency)
{
  _regs = new Hw::Mmio_register_block<32>(start);
}

// Enable the channel.

void
I2c_jz4780_channel::enable()
{
  _regs[Smb_control] = _regs[Smb_control] | Smb_no_stop_empty;

  _regs[Smb_enable] = Smb_enable_enabled;
  while (!(_regs[Smb_enable_status] & Smb_enable_enabled));
}

// Disable the channel.

void
I2c_jz4780_channel::disable()
{
  _regs[Smb_enable] = 0;
  while (_regs[Smb_enable_status] & Smb_enable_enabled);
}

// Set the frequency-related peripheral parameters.

void
I2c_jz4780_channel::set_frequency()
{
  // The APB clock (PCLK) is used to drive I2C transfers. Its value must be
  // obtained from the CPM unit. It is known as SMB_CLK here and is scaled to
  // kHz in order to keep the numbers easily representable, as is the bus
  // frequency.

  uint32_t smb_clk = _cpm->get_pclock_frequency() / 1000;
  uint32_t i2c_clk = _frequency / 1000;
  unsigned speed = (i2c_clk <= 100) ? 1 : 2;

  _regs[Smb_control] = _regs[Smb_control] | (speed << Smb_speed_bit) |
                                            Smb_disable_slave  |
                                            Smb_enable_restart |
                                            Smb_enable_master;

  // According to the programming manual, if the PCLK period is T{SMB_CLK}
  // then the I2C clock period is...

  // T{SCL} = T{SCL_high} + T{SCL_low}

  // Where...

  // T{SCL_low} = T{SMB_CLK} * (#cycles for low signal)
  // T{SCL_high} = T{SMB_CLK} * (#cycles for high signal)

  // Since, with minimum periods being defined...

  // T{SCL} >= T{min_SCL}
  // T{SCL_low} >= T{min_SCL_low}
  // T{SCL_high} >= T{min_SCL_high}
  // T{min_SCL} = T{min_SCL_low} + T{min_SCL_high}

  // Then the following applies...

  // T{SMB_CLK} * (#cycles for low signal)) >= T{min_SCL_low}
  // T{SMB_CLK} * (#cycles for high signal) >= T{min_SCL_high}

  // To work with different clock speeds while maintaining the low-to-high
  // ratios:

  // T{min_SCL_low} = T{min_SCL} * T{min_SCL_low} / T{min_SCL}
  //                = T{min_SCL} * (T{min_SCL_low} / (T{min_SCL_low} + T{min_SCL_high}))

  // T{min_SCL_high} = T{min_SCL} * T{min_SCL_high} / T{min_SCL}
  //                 = T{min_SCL} * (T{min_SCL_high} / (T{min_SCL_low} + T{min_SCL_high}))

  // Constraints are given with respect to the high and low count registers.

  // #cycles for high signal = SMBxHCNT + 8
  // #cycles for low signal = SMBxLCNT + 1

  // From earlier, this yields...

  // T{SMB_CLK} * (SMBxLCNT + 1) >= T{min_SCL_low}
  // T{SMB_CLK} * (SMBxHCNT + 8) >= T{min_SCL_high}

  // Rearranging...

  // SMBxLCNT >= (T{min_SCL_low} / T{SMB_CLK}) - 1
  //          >= T{min_SCL_low} * SMB_CLK - 1

  // SMBxHCNT >= (T{min_SCL_high} / T{SMB_CLK}) - 8
  //          >= T{min_SCL_high} * SMB_CLK - 8

  // Introducing the definitions for the high and low periods...

  // SMBxLCNT >= T{min_SCL} * (T{min_SCL_low} / (T{min_SCL_low} + T{min_SCL_high})) * SMB_CLK - 1
  //          >= (T{min_SCL_low} / T{min_SCL}) * SMB_CLK / I2C_CLK - 1

  // SMBxHCNT >= T{min_SCL} * (T{min_SCL_high} / (T{min_SCL_low} + T{min_SCL_high})) * SMB_CLK - 8
  //          >= (T{min_SCL_high} / T{min_SCL}) * SMB_CLK / I2C_CLK - 8

  uint32_t high_reg, low_reg;
  uint32_t high_count, low_count;
  uint32_t hold_count, setup_count;

  // Level hold times:

  //              Standard  Fast
  // SCL low      4.7us     1.3us
  // SCL high     4.0us     0.6us +
  // SCL period   8.7us     1.9us =

  if (i2c_clk <= 100) // 100 kHz
  {
    low_count = (smb_clk * 47) / (i2c_clk * 87) - 1;
    high_count = (smb_clk * 40) / (i2c_clk * 87) - 8;
    low_reg = Std_low_count;
    high_reg = Std_high_count;
  }
  else
  {
    low_count = (smb_clk * 13) / (i2c_clk * 19) - 1;
    high_count = (smb_clk * 6) / (i2c_clk * 19) - 8;
    low_reg = Fast_low_count;
    high_reg = Fast_high_count;
  }

  // Minimum counts are 8 and 6 for low and high respectively.

  _regs[low_reg] = low_count < 8 ? 8 : low_count;
  _regs[high_reg] = high_count < 6 ? 6 : high_count;

  // Data hold and setup times:

  //              Standard  Fast
  // t{HD;DAT}    300ns     300ns
  // t{SU;DAT}    250ns     100ns

  // T{delay} = (SMBSDAHD + 1) * T{SMB_CLK}
  // SMBSDAHD = T{delay} / T{SMB_CLK} - 1
  // SMBSDAHD = SMB_CLK * T{delay} - 1

  hold_count = (smb_clk * 300) / 1000000 - 1;

  _regs[Smb_sda_hold_time] = (_regs[Smb_sda_hold_time] & ~Smb_hold_mask) |
                             (hold_count > 0 ? Smb_hold_enable : Smb_hold_disable) |
                             (hold_count < 255 ? hold_count : 255);

  // T{delay} = (SMBSDASU - 1) * T{SMB_CLK}
  // SMBSDASU = T{delay} / T{SMB_CLK} + 1
  // SMBSDASU = SMB_CLK * T{delay} + 1

  if (i2c_clk <= 100)
    setup_count = (smb_clk * 250) / 1000000 + 1;
  else
    setup_count = (smb_clk * 100) / 1000000 + 1;

  _regs[Smb_sda_setup_time] = (_regs[Smb_sda_setup_time] & ~Smb_setup_mask) |
                              (setup_count < 255 ? setup_count : 255);
}

// Set the target address and enable transfer.
// NOTE: Only supporting 7-bit addresses currently.

void
I2c_jz4780_channel::set_target(uint8_t address)
{
  disable();
  set_frequency();
  _regs[Smb_target_address] = address & Smb_target_7bits;
  enable();
  init_parameters();
  queued = 0;
}



// Reset interrupt flags upon certain conditions.

void
I2c_jz4780_channel::reset_flags()
{
  volatile uint32_t r;

  _regs[Int_mask] = 0;

  // Read from the register to clear interrupts.

  r = _regs[Int_combined_clear];
  (void) r;
}

// Initialise interrupt flags and queue thresholds for reading and writing.

void
I2c_jz4780_channel::init_parameters()
{
  // Handle read queue conditions for data, write queue conditions for commands.

  reset_flags();

  _regs[Int_mask] = Int_rx_full  | // read condition (reading needed)
                    Int_rx_of    | // abort condition
                    Int_tx_empty | // write condition (writing needed)
                    Int_tx_abort;  // abort condition

  _regs[Tx_fifo_thold] = 0; // write when 0 in queue

  // Make sure that the stop condition does not occur automatically.

  _regs[Smb_control] = _regs[Smb_control] | Smb_no_stop_empty;
}



// Return whether the device is active.

int
I2c_jz4780_channel::active()
{
  return _regs[Smb_status] & Smb_status_master_act;
}

// Return whether data is available to receive.

int
I2c_jz4780_channel::have_input()
{
  return _regs[Smb_status] & Smb_status_rx_nempty;
}

// Return whether data is queued for sending.

int
I2c_jz4780_channel::have_output()
{
  return !(_regs[Smb_status] & Smb_status_tx_empty);
}

// Return whether data can be queued for sending.

int
I2c_jz4780_channel::can_send()
{
  return _regs[Smb_status] & Smb_status_tx_nfull;
}

// Return whether a receive operation has failed.

int
I2c_jz4780_channel::read_failed()
{
  return _regs[Int_status] & Int_rx_of;
}

// Return whether a send operation has failed.

int
I2c_jz4780_channel::write_failed()
{
  return _regs[Int_status] & Int_tx_abort;
}



// Send read commands for empty queue entries.

void
I2c_jz4780_channel::queue_reads()
{
  unsigned remaining = _total - _reqpos;
  unsigned can_queue = Smb_fifo_limit - queued + 1;

  if (!remaining)
  {
    _regs[Smb_data_command] = Smb_command_read;
    return;
  }

  // Keep the number of reads in progress below the length of the read queue.

  if (!can_queue)
  {
    //printf("remaining=%d queued=%d\n", remaining, queued);
    return;
  }

  if (remaining < can_queue)
    can_queue = remaining;

  // Queue read requests for any remaining queue entries.

  for (unsigned i = 0; i < can_queue; i++)
    _regs[Smb_data_command] = Smb_command_read;

  // Track queued messages and requested positions.

  queued += can_queue;
  _reqpos += can_queue;

  if (_total == _reqpos)
    printf("Written %d read requests.\n", _reqpos);
}

// Send write commands for empty queue entries.

void
I2c_jz4780_channel::queue_writes(uint8_t buf[], unsigned *pos, unsigned total)
{
  // Queue write requests for any remaining queue entries.

  while ((*pos < total) && can_send() && !write_failed())
  {
    _regs[Smb_data_command] = Smb_command_write | buf[*pos];
    (*pos)++;
  }
}

// Store read command results from the queue.

void
I2c_jz4780_channel::store_reads()
{
  int have_read = 0;

  // Read any input and store it in the buffer.

  while (have_input() && (_pos < _total))
  {
    _buf[_pos] = _regs[Smb_data_command] & 0xff;
    queued--;
    _pos++;

    have_read = 1;
  }

  // Update the threshold to be notified of any reduced remaining amount.

  if (have_read)
    set_read_threshold();

  //printf("Read to %d, still %d queued.\n", _pos, queued);
}

// Read from the target device.

void
I2c_jz4780_channel::start_read(uint8_t buf[], unsigned total)
{
  printf("intst=%x\n", (uint32_t) _regs[Int_status]);

  _buf = buf;
  _total = total;
  _pos = 0;
  _reqpos = 0;
  _fail = 0;
  _stop = 0;

  set_read_threshold();
}

void
I2c_jz4780_channel::set_read_threshold()
{
  unsigned remaining = _total - _pos;

  if (!remaining)
    return;

  if (remaining <= Smb_fifo_limit)
    _regs[Rx_fifo_thold] = remaining - 1; // read all remaining
  else
    _regs[Rx_fifo_thold] = Smb_fifo_limit - 1; // read to limit

  //printf("rxthold=%d\n", (uint32_t) _regs[Rx_fifo_thold] + 1);
}

void
I2c_jz4780_channel::read()
{
  if (read_failed() || write_failed())
  {
    printf("intst*=%x\n", (uint32_t) _regs[Int_status]);
    printf("rxfifo=%d\n", (uint32_t) _regs[Rx_fifo_count] & 0x3f);
    printf("smbabtsrc=%x\n", (uint32_t) _regs[Trans_abort_status0]);
    _fail = 1;
    return;
  }

  //printf("rxfifo=%d\n", (uint32_t) _regs[Rx_fifo_count]);
  //printf("intst=%x\n", (uint32_t) _regs[Int_status]);

  if (_pos < _total)
  {
    if (_regs[Int_status] & Int_rx_full)
      store_reads();
    if (_regs[Int_status] & Int_tx_empty)
      queue_reads();
  }
  else
  {
    stop();
    _regs[Int_mask] = 0;
  }
}

int
I2c_jz4780_channel::read_done()
{
  return _pos == _total;
}

unsigned
I2c_jz4780_channel::have_read()
{
  return _pos;
}

int
I2c_jz4780_channel::read_incomplete()
{
  return _fail;
}

// Write to the target device.

void
I2c_jz4780_channel::write(uint8_t buf[], unsigned total)
{
  unsigned pos = 0;

  while ((pos < total) && !write_failed())
  {
    queue_writes(buf, &pos, total);
  }
}

// Explicitly stop communication.

void
I2c_jz4780_channel::stop()
{
  if (_stop)
    return;

  _regs[Smb_control] = _regs[Smb_control] & ~Smb_no_stop_empty;
  _stop = 1;
  printf("Stop sent.\n");
}



// Initialise the I2C controller.

I2c_jz4780_chip::I2c_jz4780_chip(l4_addr_t start, l4_addr_t end,
                                 Cpm_jz4780_chip *cpm,
                                 uint32_t frequency)
: _start(start), _end(end), _cpm(cpm), _frequency(frequency)
{
}

// Obtain a channel object.

I2c_jz4780_channel *
I2c_jz4780_chip::get_channel(uint8_t channel)
{
  l4_addr_t block = _start + channel * Smb_block_offset;

  if (block < _end)
  {
    _cpm->start_i2c(channel);
    return new I2c_jz4780_channel(block, _cpm, _frequency);
  }
  else
    throw -L4_EINVAL;
}



// C language interface functions.

void *jz4780_i2c_init(l4_addr_t start, l4_addr_t end, void *cpm, uint32_t frequency)
{
  return (void *) new I2c_jz4780_chip(start, end, static_cast<Cpm_jz4780_chip *>(cpm), frequency);
}

void *jz4780_i2c_get_channel(void *i2c, uint8_t channel)
{
  return static_cast<I2c_jz4780_chip *>(i2c)->get_channel(channel);
}

void jz4780_i2c_set_target(void *i2c_channel, uint8_t addr)
{
  static_cast<I2c_jz4780_channel *>(i2c_channel)->set_target(addr);
}

void jz4780_i2c_start_read(void *i2c_channel, uint8_t buf[], unsigned total)
{
  static_cast<I2c_jz4780_channel *>(i2c_channel)->start_read(buf, total);
}

void jz4780_i2c_read(void *i2c_channel)
{
  static_cast<I2c_jz4780_channel *>(i2c_channel)->read();
}

int jz4780_i2c_read_done(void *i2c_channel)
{
  return static_cast<I2c_jz4780_channel *>(i2c_channel)->read_done();
}

unsigned jz4780_i2c_have_read(void *i2c_channel)
{
  return static_cast<I2c_jz4780_channel *>(i2c_channel)->have_read();
}

int jz4780_i2c_read_incomplete(void *i2c_channel)
{
  return static_cast<I2c_jz4780_channel *>(i2c_channel)->read_incomplete();
}

void jz4780_i2c_write(void *i2c_channel, uint8_t buf[], unsigned total)
{
  static_cast<I2c_jz4780_channel *>(i2c_channel)->write(buf, total);
}

void jz4780_i2c_stop(void *i2c_channel)
{
  static_cast<I2c_jz4780_channel *>(i2c_channel)->stop();
}
