/*
 * LCD peripheral support for the JZ4740 and related SoCs.
 *
 * Copyright (C) Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) 2015, 2016, 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include <l4/sys/cache.h>
#include <l4/sys/types.h>

#include "lcd-jz4740.h"
#include "lcd-jz4740-config.h"

#include <stdint.h>

enum Regs : unsigned
{
  Lcd_config          = 0x000,  // LCD_CFG
  Lcd_vsync           = 0x004,  // LCD_VSYNC
  Lcd_hsync           = 0x008,  // LCD_HSYNC
  Vertical_area       = 0x00c,  // LCD_VAT
  Display_hlimits     = 0x010,  // LCD_DAH
  Display_vlimits     = 0x014,  // LCD_DAV
  Lcd_ps              = 0x018,  // LCD_PS
  Lcd_cls             = 0x01c,  // LCD_CLS
  Lcd_spl             = 0x020,  // LCD_SPL
  Lcd_rev             = 0x024,  // LCD_REV
  Lcd_control         = 0x030,  // LCD_CTRL
  Lcd_status          = 0x034,  // LCD_STATE
  Lcd_irq_id          = 0x038,  // LCD_IID
  Desc_address_0      = 0x040,  // LCD_DA0
  Source_address_0    = 0x044,  // LCD_SA0
  Frame_id_0          = 0x048,  // LCD_FID0
  Command_0           = 0x04c,  // LCD_CMD0
  Desc_address_1      = 0x050,  // LCD_DA1
  Source_address_1    = 0x054,  // LCD_SA1
  Frame_id_1          = 0x058,  // LCD_FID1
  Command_1           = 0x05c,  // LCD_CMD1
};

// Lcd_config descriptions.

enum Config_values : unsigned
{
  Config_stn_pins_mask  = 0x3,
  Config_mode_mask      = 0xf,
};

// Field positions for registers employing two values, with the first typically
// being the start value and the second being an end value.

enum Value_pair_bits : unsigned
{
  Value_first           = 16,
  Value_second          = 0,
};

// Vertical_area bits.

enum Vertical_area_values : unsigned
{
  Vertical_area_horizontal_size   = Value_first,  // sum of display and blank regions (dot/pixel clock periods)
  Vertical_area_vertical_size     = Value_second, // sum of display and blank regions (line periods)
};

// Lcd_control descriptions.

enum Control_bits : unsigned
{
  Control_burst_length            = 28, // BST (burst length selection)
  Control_rgb_mode                = 27, // RGB (RGB mode)
  Control_out_underrun            = 26, // OFUP (output FIFO underrun protection)
  Control_frc_algorithm           = 24, // FRC (STN FRC algorithm selection)
  Control_palette_delay           = 16, // PDD (load palette delay counter)
  Control_frame_end_irq_mask      = 13, // EOFM (end of frame interrupt mask)
  Control_frame_start_irq_mask    = 12, // SOFM (start of frame interrupt mask)
  Control_out_underrun_irq_mask   = 11, // OFUM (output FIFO underrun interrupt mask)
  Control_in0_underrun_irq_mask   = 10, // IFUM0 (input FIFO 0 underrun interrupt mask)
  Control_in1_underrun_irq_mask   = 9,  // IFUM1 (input FIFO 1 underrun interrupt mask)
  Control_disabled_irq_mask       = 8,  // LDDM (LCD disable done interrupt mask)
  Control_quick_disabled_irq_mask = 7,  // QDM (LCD quick disable done interrupt mask)
  Control_endian_select           = 6,  // BEDN (endian selection)
  Control_bit_order               = 5,  // PEDN (bit order in bytes)
  Control_disable                 = 4,  // DIS (disable controller)
  Control_enable                  = 3,  // ENA (enable controller)
  Control_bpp                     = 0,  // BPP (bits per pixel)
};

enum Burst_length_values : unsigned
{
  Burst_length_4          = 0,  // 4 word
  Burst_length_8          = 1,  // 8 word
  Burst_length_16         = 2,  // 16 word

  // JZ4780 extensions.

  Burst_length_32         = 3,  // 32 word
  Burst_length_64         = 4,  // 64 word
  Burst_length_mask       = 0x7,
};

enum Rgb_mode_values : unsigned
{
  Rgb_mode_565            = 0,
  Rgb_mode_555            = 1,
  Rgb_mode_mask           = 0x1,
};

enum Frc_algorithm_values : unsigned
{
  Frc_greyscales_16       = 0,
  Frc_greyscales_4        = 1,
  Frc_greyscales_2        = 2,
  Frc_greyscales_mask     = 0x3,
};

enum Control_bpp_values : unsigned
{
  Control_bpp_1bpp        = 0,
  Control_bpp_2bpp        = 1,
  Control_bpp_4bpp        = 2,
  Control_bpp_8bpp        = 3,
  Control_bpp_15bpp       = 4,
  Control_bpp_16bpp       = 4,
  Control_bpp_18bpp       = 5,
  Control_bpp_24bpp       = 5,
  Control_bpp_24bpp_comp  = 6,
  Control_bpp_30bpp       = 7,
  Control_bpp_32bpp       = 7,
  Control_bpp_mask        = 0x7,
};

// Command descriptions.

enum Command_bits : unsigned
{
  Command_frame_start_irq   = 31, // SOFINT (start of frame interrupt)
  Command_frame_end_irq     = 30, // EOFINT (end of frame interrupt)
  Command_lcm_command       = 29, // JZ4780: CMD (LCM command/data via DMA0)
  Command_palette_buffer    = 28, // PAL (descriptor references palette, not display data)
  Command_frame_compressed  = 27, // JZ4780: COMPEN (16/24bpp compression enabled)
  Command_frame_enable      = 26, // JZ4780: FRM_EN
  Command_field_even        = 25, // JZ4780: FIELD_SEL (interlace even field)
  Command_16x16_block       = 24, // JZ4780: 16x16BLOCK (fetch data by 16x16 block)
  Command_buffer_length     = 0,  // LEN
};

enum Command_values : unsigned
{
  Command_buffer_length_mask  = 0x00ffffff,
};



// Utility functions.

// Round values up according to the resolution.

static uint32_t align(uint32_t value, uint32_t resolution)
{
  return (value + (resolution - 1)) & ~(resolution - 1);
}

// Value pair encoding.

static uint32_t encode_pair(uint32_t start, uint32_t end)
{
  return (start << Value_first) | (end << Value_second);
}

// RGB conversions.

static uint16_t rgb8_to_rgb16(uint8_t rgb)
{
  return ((((rgb & 0xe0) >> 5) * 4) << 11) | ((((rgb & 0x1c) >> 2) * 9) << 6) | ((rgb & 0x03) * 10);
}

static uint16_t rgb4_to_rgb16(uint8_t rgb)
{
  return ((((rgb & 8) >> 3) * 0x1f) << 11) | ((((rgb & 6) >> 1) * 0x15) << 5) | ((rgb & 1) * 0x1f);
}




// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Lcd_jz4740_chip::Lcd_jz4740_chip(l4_addr_t addr, Jz4740_lcd_panel *panel)
: _panel(panel)
{
  _regs = new Hw::Mmio_register_block<32>(addr);
  _burst_size = 16; // 16-word burst size

  // add_cid("lcd");
  // add_cid("lcd-jz4740");
}

struct Jz4740_lcd_panel *
Lcd_jz4740_chip::get_panel()
{
  return _panel;
}

void
Lcd_jz4740_chip::disable()
{
  // Set the disable bit for normal shutdown.

  _regs[Lcd_control] = _regs[Lcd_control] | (1 << Control_disable);
}

void
Lcd_jz4740_chip::disable_quick()
{
  // Clear the enable bit for quick shutdown.

  _regs[Lcd_control] = _regs[Lcd_control] & ~(1 << Control_enable);
}

void
Lcd_jz4740_chip::enable()
{
  // Clear the disable bit and set the enable bit.

  _regs[Lcd_control] = (_regs[Lcd_control] & ~(1 << Control_disable)) | (1 << Control_enable);
}

// Calculate and return the pixel clock frequency.

int
Lcd_jz4740_chip::get_pixel_clock()
{
  int pclk, multiplier;

  // Serial mode: 3 pixel clock cycles per pixel (one per channel).
  // Parallel mode: 1 pixel clock cycle per pixel.

  multiplier = have_serial_tft() ? 3 : 1;

  // Derive pixel clock rate from frame rate.
  // This multiplies the number of pixel periods in a line by the number of
  // lines in a frame, thus obtaining the number of such periods in a frame.
  // Multiplying this result with the frame rate yields the pixel frequency.

  pclk = _panel->frame_rate *
         (_panel->width * multiplier +
          _panel->hsync + _panel->line_start + _panel->line_end) *
         (_panel->height +
          _panel->vsync + _panel->frame_start + _panel->frame_end);

  // STN panel adjustments.

  if (have_stn_panel())
  {
    // Colour STN panels apparently need to be driven at three times the rate.

    if (have_colour_stn()) pclk = (pclk * 3);

    // Reduce the rate according to the width of the STN connection.
    // Since the pins setting employs log2(pins), a shift by this value is
    // equivalent to a division by the number of pins.

    pclk = pclk >> ((_panel->config & Config_stn_pins_mask) >> Jz4740_lcd_config_stn_pins);

    // Divide the rate by the number of panels.

    pclk /= get_panels();
  }

  return pclk;
}



// Return the panel mode.

uint32_t
Lcd_jz4740_chip::_mode()
{
  return _panel->config & Config_mode_mask;
}

// Return the number of panels available.

int
Lcd_jz4740_chip::get_panels()
{
  uint32_t mode = _mode();

  return (mode == Jz4740_lcd_mode_stn_dual_colour) ||
         (mode == Jz4740_lcd_mode_stn_dual_mono) ? 2 : 1;
}

// Return whether the panel is STN.

int
Lcd_jz4740_chip::have_stn_panel()
{
  uint32_t mode = _mode();

  return ((mode == Jz4740_lcd_mode_stn_single_colour) ||
          (mode == Jz4740_lcd_mode_stn_dual_colour) ||
          (mode == Jz4740_lcd_mode_stn_single_mono) ||
          (mode == Jz4740_lcd_mode_stn_dual_mono));
}

// Return whether the panel is colour STN.

int
Lcd_jz4740_chip::have_colour_stn()
{
  uint32_t mode = _mode();

  return ((mode == Jz4740_lcd_mode_stn_single_colour) ||
          (mode == Jz4740_lcd_mode_stn_dual_colour));
}

// Return whether the panel is colour STN.

int
Lcd_jz4740_chip::have_serial_tft()
{
  return _mode() == Jz4740_lcd_mode_tft_serial;
}



// Return the line memory size.

l4_size_t
Lcd_jz4740_chip::get_line_size()
{
  // Lines must be aligned to a word boundary.

  return align((_panel->width * _panel->bpp) / 8, sizeof(uint32_t));
}

// Return the screen memory size.

l4_size_t
Lcd_jz4740_chip::get_screen_size()
{
  return get_line_size() * _panel->height;
}

// Return the aligned size for the DMA transfer.

l4_size_t 
Lcd_jz4740_chip::get_aligned_size()
{
  return align(get_screen_size(), _burst_size * sizeof(uint32_t));
}

// Return the size of the palette.

l4_size_t
Lcd_jz4740_chip::get_palette_size()
{
  // No palette for modes with more than eight bits per pixel.

  if (_panel->bpp > 8) return 0;

  // Get the size of a collection of two-byte entries, one per colour.

  return (1 << (_panel->bpp)) * sizeof(uint16_t);
}

// Return the aligned size of the palette for the DMA transfer.

l4_size_t
Lcd_jz4740_chip::get_aligned_palette_size()
{
  return align(get_palette_size(), _burst_size * sizeof(uint32_t));
}

// Return the total memory requirements of the framebuffers and palette.

l4_size_t
Lcd_jz4740_chip::get_total_size()
{
  return get_aligned_size() * get_panels() + get_aligned_palette_size();
}

// Return the total memory requirements of any DMA descriptors.

l4_size_t
Lcd_jz4740_chip::get_descriptors_size()
{
   return 3 * sizeof(struct Jz4740_lcd_descriptor);
}



// Functions returning addresses of each data region.
// The base parameter permits the retrieval of virtual or physical addresses.

l4_addr_t
Lcd_jz4740_chip::get_palette(l4_addr_t base)
{       
  // Use memory at the end of the allocated region for the palette.
        
  return base + (get_panels() * get_aligned_size()) - get_aligned_palette_size();
}

l4_addr_t
Lcd_jz4740_chip::get_framebuffer(int panel, l4_addr_t base)
{
  // Framebuffers for panels are allocated at the start of the region.

  return base + (panel * get_aligned_size());
}



// Palette initialisation.

void
Lcd_jz4740_chip::init_palette(l4_addr_t palette)
{
  uint8_t colours = 1 << (_panel->bpp);
  uint16_t *entry = (uint16_t *) palette;
  uint16_t *end = entry + colours;
  uint8_t value = 0;

  while (entry < end)
  {
    switch (_panel->bpp)
    {
      case 4:
      *entry = rgb4_to_rgb16(value);
      break;

      case 8:
      *entry = rgb8_to_rgb16(value);
      break;

      default:
      break;
    }

    value++;
    entry++;
  }
}



// Return colour depth control value.
// NOTE: Not supporting JZ4780 options.

uint32_t
Lcd_jz4740_chip::_control_bpp()
{
  switch (_panel->bpp)
  {
    case 1:         return Control_bpp_1bpp;
    case 2:         return Control_bpp_2bpp;
    case 3 ... 4:   return Control_bpp_4bpp;
    case 5 ... 8:   return Control_bpp_8bpp;
    case 9 ... 15:  return Control_bpp_15bpp | (Rgb_mode_555 << Control_rgb_mode);
    case 17 ... 18: return Control_bpp_18bpp;
    case 19 ... 32: return Control_bpp_24bpp;
    case 16:
    default:        return Control_bpp_16bpp;
  }
}

// Return a panel-related control value.

uint32_t
Lcd_jz4740_chip::_control_panel()
{
  if (have_stn_panel())
    return _control_stn_frc();
  else
    return 0;
}

// Return a STN-related control value.

uint32_t
Lcd_jz4740_chip::_control_stn_frc()
{
  if (_panel->bpp <= 2)
    return Frc_greyscales_2;
  if (_panel->bpp <= 4)
    return Frc_greyscales_4;
  return Frc_greyscales_16;
}

// Return a transfer-related control value.

uint32_t
Lcd_jz4740_chip::_control_transfer()
{
  uint32_t length;

  switch (_burst_size)
  {
    case 4:  length = Burst_length_4;  break;
    case 8:  length = Burst_length_8;  break;
    case 32: length = Burst_length_32; break;
    case 64: length = Burst_length_64; break;
    case 16:
    default: length = Burst_length_16; break;
  }

  return (length << Control_burst_length) | (1 << Control_out_underrun);
}

// STN panel-specific initialisation.

void
Lcd_jz4740_chip::_init_stn()
{
  // Divide the height by the number of panels.

  uint32_t height = _panel->height / get_panels();

  // Since the value is log2(pins), 1 << value yields the number of pins.

  int pins = 1 << ((_panel->config & Config_stn_pins_mask) >> Jz4740_lcd_config_stn_pins);

  // Round parameters up to a multiple of the number of pins.

  uint32_t hsync = align(_panel->hsync, pins);
  uint32_t line_start = align(_panel->line_start, pins);
  uint32_t line_end = align(_panel->line_end, pins);

  // Define the start and end positions of visible data on a line and in a frame.
  // Visible frame data is anchored at line zero, with the start region
  // preceding this line (and thus appearing at the end of the preceding frame).

  uint32_t line_start_pos = line_start;
  uint32_t line_end_pos = line_start_pos + _panel->width;
  uint32_t frame_start_pos = 0;
  uint32_t frame_end_pos = frame_start_pos + height;

  // Define sync pulse locations, with hsync occurring after the visible data.

  _regs[Lcd_hsync] = encode_pair(line_end_pos, line_end_pos + hsync);
  _regs[Lcd_vsync] = encode_pair(0, _panel->vsync);

  // Set the display area and limits.

  _regs[Vertical_area] = encode_pair(line_end_pos + hsync + line_end,
                                     frame_end_pos + _panel->vsync + _panel->frame_end + _panel->frame_start);

  _regs[Display_hlimits] = encode_pair(line_start_pos, line_end_pos);
  _regs[Display_vlimits] = encode_pair(frame_start_pos, frame_end_pos);

  // Set the AC bias signal.

  _regs[Lcd_ps] = encode_pair(0, _panel->frame_start + height + _panel->vsync + _panel->frame_end);
}

// TFT panel-specific initialisation.

void
Lcd_jz4740_chip::_init_tft()
{
  // Define the start and end positions of visible data on a line and in a frame.

  uint32_t line_start_pos = _panel->line_start + _panel->hsync;
  uint32_t line_end_pos = line_start_pos + _panel->width;
  uint32_t frame_start_pos = _panel->frame_start + _panel->vsync;
  uint32_t frame_end_pos = frame_start_pos + _panel->height;

  // Define sync pulse locations, with pulses appearing before visible data.

  _regs[Lcd_hsync] = encode_pair(0, _panel->hsync);
  _regs[Lcd_vsync] = encode_pair(0, _panel->vsync);

  // Set the display area and limits.

  _regs[Vertical_area] = encode_pair(line_end_pos + _panel->line_end,
                                     frame_end_pos + _panel->frame_end);

  _regs[Display_hlimits] = encode_pair(line_start_pos, line_end_pos);
  _regs[Display_vlimits] = encode_pair(frame_start_pos, frame_end_pos);
}

// Initialise the panel.
// NOTE: Only generic STN and TFT panels are supported.

void
Lcd_jz4740_chip::_init_panel()
{
  if (have_stn_panel())
    _init_stn();
  else
    switch (_mode())
    {
      case Jz4740_lcd_mode_tft_generic:
      case Jz4740_lcd_mode_tft_casio:
      case Jz4740_lcd_mode_tft_serial: _init_tft();

      default: break;
    }
}

// Initialise a DMA descriptor.

void
Lcd_jz4740_chip::_set_descriptor(struct Jz4740_lcd_descriptor &desc,
                                 l4_addr_t source, l4_size_t size,
                                 struct Jz4740_lcd_descriptor *next,
                                 uint32_t flags)
{
  // In the command, indicate the number of words from the source for transfer.

  desc.next = next;
  desc.source = source;
  desc.identifier = 0;
  desc.command = ((size / sizeof(uint32_t)) & Command_buffer_length_mask) | flags;
}



// Initialise the LCD controller with the memory, panel and framebuffer details.
// Any palette must be initialised separately using get_palette and init_palette.

void
Lcd_jz4740_chip::config(struct Jz4740_lcd_descriptor *desc_vaddr,
                        struct Jz4740_lcd_descriptor *desc_paddr,
                        l4_addr_t fb_paddr)
{
  int have_palette = (_panel->bpp <= 8);

  // Provide the first framebuffer descriptor in single and dual modes.
  // Flip back and forth between any palette and the framebuffer.

  _set_descriptor(desc_vaddr[0], get_framebuffer(0, fb_paddr),
                                 get_screen_size(),
                                 have_palette ? desc_paddr + 2 : desc_paddr);

  // Provide the second framebuffer descriptor only in dual-panel mode.
  // Only employ this descriptor in the second DMA channel.

  if (get_panels() == 2)
    _set_descriptor(desc_vaddr[1], get_framebuffer(1, fb_paddr),
                                   get_screen_size(),
                                   desc_paddr + 1);

  // Initialise palette descriptor details for lower colour depths.

  if (have_palette)
    _set_descriptor(desc_vaddr[2], get_palette(fb_paddr),
                                   get_palette_size(),
                                   desc_paddr,
                                   Command_palette_buffer);

  // Flush cached structure data.

  l4_cache_clean_data((unsigned long) desc_vaddr,
                      (unsigned long) desc_vaddr + get_descriptors_size());

  // Configure DMA by setting frame descriptor addresses.

  // Provide the palette descriptor address first, if employed.

  _regs[Desc_address_0] = (uint32_t) (have_palette ? desc_paddr + 2 : desc_paddr);

  // Provide a descriptor for the second DMA channel in dual-panel mode.

  if (get_panels() == 2)
    _regs[Desc_address_1] = (uint32_t) (desc_paddr + 1);

  // Initialise panel-related registers.

  _init_panel();

  // Initialise the control and configuration registers.

  _regs[Lcd_control] = _control_panel() | _control_bpp() | _control_transfer();
  _regs[Lcd_config] = _panel->config;
}



// C language interface functions.

void *
jz4740_lcd_init(l4_addr_t lcd_base, struct Jz4740_lcd_panel *panel)
{
  return (void *) new Lcd_jz4740_chip(lcd_base, panel);
}

void
jz4740_lcd_config(void *lcd, struct Jz4740_lcd_descriptor *desc_vaddr,
                             struct Jz4740_lcd_descriptor *desc_paddr,
                             l4_addr_t fb_paddr)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->config(desc_vaddr, desc_paddr, fb_paddr);
}

void
jz4740_lcd_disable(void *lcd)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->disable();
}

void
jz4740_lcd_disable_quick(void *lcd)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->disable_quick();
}

void
jz4740_lcd_enable(void *lcd)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->enable();
}

int
jz4740_lcd_get_pixel_clock(void *lcd)
{
  return static_cast<Lcd_jz4740_chip *>(lcd)->get_pixel_clock();
}

l4_addr_t
jz4740_lcd_get_palette(void *lcd, l4_addr_t base)
{
  return static_cast<Lcd_jz4740_chip *>(lcd)->get_palette(base);
}

void
jz4740_lcd_init_palette(void *lcd, l4_addr_t palette)
{
  static_cast<Lcd_jz4740_chip *>(lcd)->init_palette(palette);
}
