/*
 * Pulse width modulation peripheral support.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "pwm-jz4730.h"



enum Regs : unsigned
{
  Pwm_control         = 0x000,  // CTR
  Pwm_period          = 0x004,  // PER
  Pwm_duty            = 0x008,  // DUT
};

enum Control_bits : unsigned
{
  Pwm_enable          = 7,
  Pwm_abrupt_shutdown = 6,
  Pwm_prescale_mask   = 0x3f,
};

enum Duty_bits : unsigned
{
  Pwm_duty_full       = 10,
  Pwm_duty_mask       = 0x3ff,
};

enum Period_bits : unsigned
{
  Pwm_period_mask     = 0x3ff,
};



// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Pwm_jz4730_chip::Pwm_jz4730_chip(l4_addr_t start, l4_addr_t end)
: _start(start), _end(end)
{
  _regs = new Hw::Mmio_register_block<32>(_start);

  // add_cid("pwm");
  // add_cid("pwm-jz4730");
}

void
Pwm_jz4730_chip::set_control(uint8_t control)
{
  // This allows the setting of the actual control bits.

  _regs[Pwm_control] = control;
}

void
Pwm_jz4730_chip::disable()
{
  _regs[Pwm_control] = _regs[Pwm_control] & ~(1 << Pwm_enable);
}

void
Pwm_jz4730_chip::enable()
{
  _regs[Pwm_control] = _regs[Pwm_control] | (1 << Pwm_enable);
}

void
Pwm_jz4730_chip::set_shutdown_mode(enum Jz4730_pwm_shutdown_mode mode)
{
  if (mode == Jz4730_pwm_shutdown_graceful)
    _regs[Pwm_control] = _regs[Pwm_control] & ~(1 << Pwm_abrupt_shutdown);
  else
    _regs[Pwm_control] = _regs[Pwm_control] | (1 << Pwm_abrupt_shutdown);
}

void
Pwm_jz4730_chip::set_duty(uint16_t duty)
{
  // This deselects full duty, if it were set.
  // NOTE: A way of selecting full duty is required.

  _regs[Pwm_duty] = (_regs[Pwm_duty] & ~(Pwm_duty_mask | (1 << Pwm_duty_full))) |
                    (duty & Pwm_duty_mask);
}

void
Pwm_jz4730_chip::set_period(uint16_t period)
{
  _regs[Pwm_period] = (_regs[Pwm_period] & ~Pwm_period_mask) |
                      (period & Pwm_period_mask);
}




// C language interface functions.

void *
jz4730_pwm_init(l4_addr_t start, l4_addr_t end)
{
  return (void *) new Pwm_jz4730_chip(start, end);
}

void
jz4730_pwm_disable(void *pwm)
{
  static_cast<Pwm_jz4730_chip *>(pwm)->disable();
}

void
jz4730_pwm_enable(void *pwm)
{
  static_cast<Pwm_jz4730_chip *>(pwm)->enable();
}

void
jz4730_pwm_set_shutdown_mode(void *pwm, enum Jz4730_pwm_shutdown_mode mode)
{
  static_cast<Pwm_jz4730_chip *>(pwm)->set_shutdown_mode(mode);
}

void
jz4730_pwm_set_control(void *pwm, uint8_t control)
{
  static_cast<Pwm_jz4730_chip *>(pwm)->set_control(control);
}

void
jz4730_pwm_set_duty(void *pwm, uint16_t duty)
{
  static_cast<Pwm_jz4730_chip *>(pwm)->set_duty(duty);
}

void
jz4730_pwm_set_period(void *pwm, uint16_t period)
{
  static_cast<Pwm_jz4730_chip *>(pwm)->set_period(period);
}
