/*
 * Export a JZ4730 PWM peripheral as a server.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4730.h>
#include <l4/devices/pwm-jz4730.h>

#include <l4/cxx/ipc_server>
#include <l4/re/env>
#include <l4/re/util/object_registry>

#include <stdint.h>
#include <stdlib.h>

#include "pwm-ops.h"
#include "memory.h"

/* Virtual addresses for the GPIO and PWM register blocks. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;
static l4_addr_t pwm_virt_base = 0, pwm_virt_base_end = 0;



static int setup_memory(void)
{
  if (get_memory("jz4730-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  if (get_memory("jz4730-pwm", &pwm_virt_base, &pwm_virt_base_end))
    return 1;

  return 0;
}



/* PWM peripheral device. */

class Pwm_device_server : public L4::Server_object_t<L4::Kobject>
{
  Pwm_jz4730_chip *_device = 0;
  int _duty, _period, _prescale;

public:
  /* Associate the device with a particular memory region. */

  explicit Pwm_device_server(Pwm_jz4730_chip *device, int duty, int period, int prescale)
  : _device(device), _duty(duty), _period(period), _prescale(prescale)
  {
  }

  /* Dispatch incoming requests. */

  int dispatch(l4_umword_t obj, L4::Ipc::Iostream &ios)
  {
    l4_msgtag_t tag;
    int arg;

    (void) obj;
    ios >> tag;

    switch (tag.label())
    {
      case Pwm_op_disable:
        disable();
        return L4_EOK;

      case Pwm_op_enable:
        enable();
        return L4_EOK;

      case Pwm_op_set_control:
        ios >> arg;
        set_control(arg);
        return L4_EOK;

      case Pwm_op_set_duty:
        ios >> arg;
        set_duty(arg);
        return L4_EOK;

      case Pwm_op_set_period:
        ios >> arg;
        set_period(arg);
        return L4_EOK;

      default:
        return -L4_EBADPROTO;
    }
  }

  void disable()
  {
    _device->disable();
  }

  void enable()
  {
    set_duty(_duty);
    set_period(_period);
    set_control(0x80 | _prescale); /* enable | prescale */
  }

  /* Set the control register. */

  void set_control(uint8_t control)
  {
    _device->set_control(control);
  }

  /* Set the PWM duty cycle. */

  void set_duty(uint16_t duty)
  {
    _duty = duty;
    _device->set_duty(duty);
  }

  /* Set the PWM period. */

  void set_period(uint16_t period)
  {
    _period = period;
    _device->set_period(period);
  }
};

static L4Re::Util::Registry_server<> server;



/* Initialise devices and start the server. */

static void run(int number, int duty, int period, int prescale)
{
  Pwm_jz4730_chip pwm_device(pwm_virt_base + number * 0x1000,
                             pwm_virt_base + (number + 1) * 0x1000);

  Gpio_jz4730_chip gpio_port_c(gpio_virt_base + 2 * 0x30,
                               gpio_virt_base + 3 * 0x30, 32);

  /* Enable the PWM output for PC30 or PC31. */

  gpio_port_c.config_pad(30 + number, Hw::Gpio_chip::Function_alt, 1);

  /* Initialise and register a new server object. */

  Pwm_device_server server_obj(&pwm_device, duty, period, prescale);
  server.registry()->register_obj(&server_obj, "pwm");

  /* Enter the IPC server loop. */

  server.loop();
}



/* Arguments: <PWM device number> <duty> <period> <prescale> */

int main(int argc, char *argv[])
{
  int number, duty, period, prescale;

  if (argc < 5) return 1;

  /* Interpret and restrict the device number. */

  number = atoi(argv[1]);
  if ((number < 0) || (number > 1)) return 1;

  /* Interpret and restrict the initial parameters. */

  duty = atoi(argv[2]);
  period = atoi(argv[3]);
  prescale = atoi(argv[4]);
  if ((duty < 0) || (period < 0) || (prescale < 0)) return 1;

  /* Obtain access to peripheral memory. */

  if (setup_memory()) return 1;

  run(number, duty, period, prescale);
  return 0;
}
