/*
 * Export JZ4740 GPIO pins as a SPI server.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4740.h>

#include <l4/cxx/ipc_server>
#include <l4/re/env>
#include <l4/re/util/object_registry>

#include <errno.h>
#include <stdint.h>
#include <stdlib.h>

#include "spi-ops.h"
#include "memory.h"

/* Virtual addresses for the GPIO register block. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;



static int setup_memory(void)
{
  if (get_memory("jz4740-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  return 0;
}



/* SPI peripheral device. */

class Spi_device_server : public L4::Server_object_t<L4::Kobject>
{
  Gpio_jz4740_chip *_clock_device = 0, *_data_device = 0, *_enable_device = 0;
  int _clock_pin, _data_pin, _enable_pin;

public:
  /* Associate the device with a particular memory region. */

  explicit Spi_device_server(Gpio_jz4740_chip *clock_device,
                             Gpio_jz4740_chip *data_device,
                             Gpio_jz4740_chip *enable_device,
                             int clock_pin, int data_pin, int enable_pin)
  : _clock_device(clock_device),
    _data_device(data_device),
    _enable_device(enable_device),
    _clock_pin(clock_pin), _data_pin(data_pin), _enable_pin(enable_pin)
  {
  }

  /* Dispatch incoming requests. */

  int dispatch(l4_umword_t obj, L4::Ipc::Iostream &ios)
  {
    l4_msgtag_t tag;
    int bits, data;

    (void) obj;
    ios >> tag;

    switch (tag.label())
    {
      case Spi_op_send:
        ios >> bits;
        ios >> data;
        send(bits, data);
        return L4_EOK;

      default:
        return -L4_EBADPROTO;
    }
  }

  /* Send a SPI command. */

  void send(int bits, int data)
  {
    uint32_t mask = 1 << (bits - 1);
    int bit;

    /* Initialise pin levels. */

    _enable_device->set(_enable_pin, 1);
    _clock_device->set(_clock_pin, 1);
    _data_device->set(_data_pin, 0);

    /* Enter the transmission state. */

    _enable_device->set(_enable_pin, 0);
    
    /* Clock data using the clock and data outputs. */
    
    for (bit = 0; bit < bits; bit++)
    { 
      _clock_device->set(_clock_pin, 0);
      _data_device->set(_data_pin, data & mask ? 1 : 0);
      _clock_device->set(_clock_pin, 1);
      mask >>= 1;
    }

    _enable_device->set(_enable_pin, 1);
  }
};

static L4Re::Util::Registry_server<> server;



/*
Parse a string of the form "<port><pin>" where <port> is a single character
whose position in the ports string indicates the port number, and where <pin> is
either a base 10 number, a base 8 number preceded by "0", or a base 16 number
preceded by "0x" or "0X", defined by the strtol library function.
*/

static int parse_pin(const char *s, const char *ports, int *port, int *pin)
{
  int i = 0;

  if (!s || !(*s)) return 0;

  /* Parse prefix in character range. */

  while (*ports)
  {
    if (s[0] == *ports)
    {
      *port = i;

      /* Parse pin number. */

      *pin = (int) strtol(s+1, NULL, 0);
      return !errno;
    }
    ports++; i++;
  }

  return 0;
}

/* Arguments: <SPI clock pin> <SPI data pin> <SPI enable pin> */

int main(int argc, char *argv[])
{
  int clock_port, clock_pin, data_port, data_pin, enable_port, enable_pin;

  if (argc < 4) return 1;

  /* Interpret the pin details. */

  if (!parse_pin(argv[1], "ABCD", &clock_port, &clock_pin)) return 1;
  if (!parse_pin(argv[2], "ABCD", &data_port, &data_pin)) return 1;
  if (!parse_pin(argv[3], "ABCD", &enable_port, &enable_pin)) return 1;

  /* Obtain access to peripheral memory. */

  if (setup_memory()) return 1;

  /* Configure the clock pin. */

  Gpio_jz4740_chip gpio_port_clock(gpio_virt_base + clock_port * 0x100,
                                   gpio_virt_base + (clock_port + 1) * 0x100, 32);

  gpio_port_clock.setup(clock_pin, Hw::Gpio_chip::Output, 0);

  /* Configure the data pin. */

  Gpio_jz4740_chip gpio_port_data(gpio_virt_base + data_port * 0x100,
                                  gpio_virt_base + (data_port + 1) * 0x100, 32);

  gpio_port_data.setup(data_pin, Hw::Gpio_chip::Output, 0);

  /* Configure the enable pin. */

  Gpio_jz4740_chip gpio_port_enable(gpio_virt_base + enable_port * 0x100,
                                    gpio_virt_base + (enable_port + 1) * 0x100, 32);

  gpio_port_enable.setup(enable_pin, Hw::Gpio_chip::Output, 0);

  /* Initialise and register a new server object. */

  Spi_device_server server_obj(&gpio_port_clock, &gpio_port_data, &gpio_port_enable,
                               clock_pin, data_pin, enable_pin);

  server.registry()->register_obj(&server_obj, "spi");

  /* Enter the IPC server loop. */

  server.loop();
  return 0;
}
