/*
 * Generic event loop functionality.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#ifdef __cplusplus

#include <pthread.h>
#include <pthread-l4.h>

/* Event loop abstraction. */

class Event_loop
{
protected:
  /* Thread properties. */

  int _priority;
  pthread_t _pthread;

public:
  /* Initialise the event loop with a thread priority. */

  explicit Event_loop(int priority=0x20)
  : _priority(priority)
  {
  }

  /* Static main loop function for event-handling thread. */

  static void *event_loop(void *data);

  /* Event handler method. */

  virtual void handle() = 0;

  /* Initiation function. */

  virtual void start();
};



/* Event loop with handler. */

template <class T>
class Event_handler_loop : public Event_loop
{
protected:
  /* Event type. */

  typedef T Event_type;

  /* Handler function type. */

  typedef void (*Event_handler)(Event_type &);

  /* External handler function. */

  Event_handler _handler;

public:
  /* Initialise the event loop with a thread priority. */

  explicit Event_handler_loop(int priority=0x20)
  : Event_loop(priority)
  {
  }

  /* Attach a handler to the loop. */

  virtual void attach(Event_handler handler)
  {
    _handler = handler; 
  }
};

/* Event loop with handler and null event type. */

template <>
class Event_handler_loop<void> : public Event_loop
{
protected:
  /* Event type. */

  typedef void Event_type;

  /* Handler function type. */

  typedef void (*Event_handler)();

  /* External handler function. */

  Event_handler _handler;

public:
  /* Initialise the event loop with a thread priority. */

  explicit Event_handler_loop(int priority=0x20)
  : Event_loop(priority)
  {
  }

  /* Attach a handler to the loop. */

  virtual void attach(Event_handler handler)
  {
    _handler = handler; 
  }
};

#endif
