/*
 * (c) 2008-2009 Adam Lackorzynski <adam@os.inf.tu-dresden.de>
 *     economic rights: Technische Universität Dresden (Germany)
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
/*
 * Access the I2C peripherals on the MIPS Creator CI20 board.
 */

#include <l4/devices/cpm-jz4780.h>
#include <l4/devices/gpio-jz4780.h>
#include <l4/devices/i2c-jz4780.h>
#include <l4/io/io.h>
#include <l4/libedid/edid.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/vbus/vbus.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>



enum {
  DDCSCL = 24,  /* via PORTF */
  DDCSDA = 25,  /* via PORTF */
};



/* Device and resource discovery. */

static char const *resource_type(enum l4io_resource_types_t type)
{
  switch (type)
  {
    case L4VBUS_RESOURCE_INVALID:
    return "INVALID";

    case L4VBUS_RESOURCE_IRQ:
    return "IRQ";

    case L4VBUS_RESOURCE_MEM:
    return "MEMORY";

    default:
    return "OTHER";
  }
}

static int vbus_get_device(char const *hid, l4io_device_handle_t *dh, l4io_resource_handle_t *rh)
{
  int result = l4io_lookup_device(hid, dh, 0, rh);

  if (result < 0)
    printf("Could not access '%s': %s\n", hid, result == -L4_ENOENT ? "no such device" : "no device");

  return result;
}

static int vbus_get_resource(l4io_device_handle_t dh, l4io_resource_t *res,
                            enum l4io_resource_types_t type)
{
  int current = 0, result = 0;
  l4_cap_idx_t vbus = l4re_env_get_cap("vbus");

  do
  {
    result = l4vbus_get_resource(vbus, dh, current, res);

    if (result)
      printf("Could not access resource of type %s.\n", resource_type(type));
    else
      printf("Resource %d: type %s, start=%lx, end=%lx\n", res->id,
        resource_type(res->type), res->start, res->end);

    current++;
  }
  while ((!result) && (res->type != type));

  return result;
}

static int vbus_get_irq(char const *hid, l4_uint32_t *start, l4_uint32_t *end)
{
  l4io_device_handle_t dh;
  l4io_resource_handle_t rh;
  l4io_resource_t res;
  int result;

  result = vbus_get_device(hid, &dh, &rh);

  if (result < 0)
    return result;

  result = vbus_get_resource(dh, &res, L4IO_RESOURCE_IRQ);

  if (result)
    return result;

  *start = res.start;
  *end = res.end;

  return result;
}

static int vbus_get_memory(char const *hid, l4_addr_t *start, l4_addr_t *end)
{
  l4io_device_handle_t dh;
  l4io_resource_handle_t rh;
  l4io_resource_t res;
  int result;

  result = vbus_get_device(hid, &dh, &rh);

  if (result < 0)
    return result;

  result = vbus_get_resource(dh, &res, L4IO_RESOURCE_MEM);

  if (result)
    return result;

  if ((result = l4io_request_iomem(res.start, res.end - res.start + 1,
                                  L4IO_MEM_NONCACHED, start)))
  {
    printf("Could not get address for '%s'.\n", hid);
    return result;
  }

  printf("Resource at 0x%lx...0x%lx.\n", res.start, res.end);

  *end = *start + (res.end - res.start + 1);

  return 0;
}

static long item_in_range(long start, long end, long index)
{
  if (start < end)
    return start + index;
  else
    return start - index;
}

static void show_deviceid(uint8_t buf[])
{
  printf("Manufacturer %x part %x revision %x\n",
         (buf[0] << 4) | (buf[1] >> 4),
         ((buf[1] & 0xf) << 5) | (buf[2] >> 3),
         buf[2] & 0x7);
}

static void show_data(uint8_t buf[], unsigned pos)
{
  unsigned i;

  if (pos)
  {
    printf("Read %d bytes from bus.\n", pos);
    for (i = 0; i < pos; i++)
      printf(" %02x", buf[i]);
    printf("\n");
  }
  else
    printf("No reply from bus.\n");
}

int main(void)
{
  l4_addr_t gpio_base = 0, gpio_base_end = 0;
  l4_addr_t port_f, port_f_end;
  l4_addr_t i2c_base = 0, i2c_base_end = 0, cpm_base = 0, cpm_base_end;
  void *gpio_port_f;
  void *i2c, *i2c_channel, *cpm;
  l4_uint32_t i2c_irq_start = 0, i2c_irq_end = 0, i2c_irq;
  l4_cap_idx_t irqcap, icucap;
  l4_msgtag_t tag;
  long err;
  int result = 0;
  uint8_t buf[256];
  unsigned pos;
  unsigned width = 0, height = 0;

  /* Obtain capabilities for the interrupt controller and an interrupt. */

  irqcap = l4re_util_cap_alloc();
  icucap = l4re_env_get_cap("icu");

  if (l4_is_invalid_cap(icucap))
  {
    printf("No 'icu' capability available in the virtual bus.\n");
    return 1;
  }

  if (l4_is_invalid_cap(irqcap))
  {
    printf("No capability available for the interrupt.\n");
    return 1;
  }

  /* Obtain resource details describing the interrupt for I2C channel 4. */

  printf("Access IRQ...\n");

  if ((result = vbus_get_irq("jz4780-i2c", &i2c_irq_start, &i2c_irq_end)) < 0)
    return 1;

  i2c_irq = item_in_range(i2c_irq_start, i2c_irq_end, 4);
  printf("IRQ range at %d...%d.\n", i2c_irq_start, i2c_irq_end);
  printf("I2C IRQ at %d.\n", i2c_irq);

  /* Obtain resource details describing I/O memory. */

  printf("Access GPIO...\n");

  if ((result = vbus_get_memory("jz4780-gpio", &gpio_base, &gpio_base_end)) < 0)
    return 1;

  printf("GPIO at 0x%lx...0x%lx.\n", gpio_base, gpio_base_end);

  printf("Access CPM...\n");

  if ((result = vbus_get_memory("jz4780-cpm", &cpm_base, &cpm_base_end)) < 0)
    return 1;

  printf("CPM at 0x%lx...0x%lx.\n", cpm_base, cpm_base_end);

  printf("Access I2C...\n");

  if ((result = vbus_get_memory("jz4780-i2c", &i2c_base, &i2c_base_end)) < 0)
    return 1;

  printf("I2C at 0x%lx...0x%lx.\n", i2c_base, i2c_base_end);

  /* Create an interrupt object. */

  if ((err = l4_error(tag = l4_factory_create_irq(l4re_global_env->factory, irqcap))))
  {
    printf("Could not create IRQ object: %lx\n", err);
    return 1;
  }

  /* Bind the interrupt object to the IRQ number. */

  if ((err = l4_error(l4_icu_bind(icucap, i2c_irq, irqcap))))
  {
    printf("Could not bind IRQ %d to the ICU: %ld\n", i2c_irq, err);
    return 1;
  }

  /* Attach ourselves to the interrupt handler. */

  tag = l4_irq_attach(irqcap, 0xDEAD, l4re_env()->main_thread);

  if ((err = l4_error(tag)))
  {
    printf("Could not attach to IRQ %d: %ld\n", i2c_irq, err);
    return 1;
  }

  /* Configure pins. */

  port_f = gpio_base + 0x500;
  port_f_end = port_f + 0x100;

  printf("PORTF at 0x%lx...0x%lx.\n", port_f, port_f_end);

  gpio_port_f = jz4780_gpio_init(port_f, port_f_end, 32, 0xffa7f00f, 0x00580ff0);

  printf("Set up GPIO pins...\n");

  jz4780_gpio_config_pad(gpio_port_f, DDCSCL, Function_alt, 1);
  jz4780_gpio_config_pad(gpio_port_f, DDCSDA, Function_alt, 1);

  /* Obtain CPM and I2C objects. */

  cpm = jz4780_cpm_init(cpm_base);

  /* Attempt to set the PCLK source to SCLK_A. */

  jz4780_cpm_set_pclock_source(cpm, 1);
  printf("Peripheral clock: %d\n", jz4780_cpm_get_pclock_frequency(cpm));

  /* Obtain I2C reference. */

  i2c = jz4780_i2c_init(i2c_base, i2c_base_end, cpm, 100000); /* 100 kHz */

  printf("Trying DDC on I2C4...\n");

  i2c_channel = jz4780_i2c_get_channel(i2c, 4);

  /* Attempt to read from address 0x50 for DDC.
     See: drivers/video/fbdev/core/fb_ddc.c */

  jz4780_i2c_set_target(i2c_channel, 0x50);
  buf[0] = 0;
  jz4780_i2c_write(i2c_channel, buf, 1);
  jz4780_i2c_start_read(i2c_channel, buf, 128);

  printf("Waiting\n");

  while (!jz4780_i2c_read_done(i2c_channel))
  {
    if (jz4780_i2c_read_incomplete(i2c_channel))
    {
      printf("Failed\n");
      break;
    }

    tag = l4_irq_receive(irqcap, L4_IPC_NEVER);
    
    if ((err = l4_ipc_error(tag, l4_utcb())))
    {
      printf("Error on IRQ receive: %ld\n", err);
      continue;
    }

    jz4780_i2c_read(i2c_channel);
  }

  pos = jz4780_i2c_have_read(i2c_channel);
  show_data(buf, pos);

  /* Attempt to decode EDID information. */

  libedid_prefered_resolution(buf, &width, &height);
  printf("Preferred resolution: %d x %d\n", width, height);

  libedid_dump_standard_timings(buf);

  /* Detach from the interrupt. */

  tag = l4_irq_detach(irqcap);

  if ((err = l4_error(tag)))
    printf("Error detaching from IRQ: %ld\n", err);

  return 0;
}
