/*
 * (c) 2008-2009 Adam Lackorzynski <adam@os.inf.tu-dresden.de>
 *     economic rights: Technische Universität Dresden (Germany)
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
/*
 * Access the LED and other GPIOs on the MIPS Creator CI20 board.
 * This example shows how to use the following GPIOs:
 *
 * PD17 - SW1 button
 * PF15 - blue/red LED (high -> red, low -> blue)
 */

#include <l4/devices/gpio-jz4780.h>
#include <l4/io/io.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/vbus/vbus.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>

enum {
  LED = 15, // via PORTF
  SW1 = 17, // via PORTD
};



/* Device and resource discovery. */

static char const *resource_type(enum l4io_resource_types_t type)
{
  switch (type)
  {
    case L4VBUS_RESOURCE_INVALID:
    return "INVALID";

    case L4VBUS_RESOURCE_IRQ:
    return "IRQ";

    case L4VBUS_RESOURCE_MEM:
    return "MEMORY";

    default:
    return "OTHER";
  }
}

static int gpio_get_device(char const *hid, l4io_device_handle_t *dh, l4io_resource_handle_t *rh)
{
  int result = l4io_lookup_device(hid, dh, 0, rh);

  if (result < 0)
    printf("Could not access '%s': %s\n", hid, result == -L4_ENOENT ? "no such device" : "no device");

  return result;
}

static int gpio_get_resource(l4io_device_handle_t dh, l4io_resource_t *res,
                             enum l4io_resource_types_t type)
{
  int current = 0, result = 0;
  l4_cap_idx_t vbus = l4re_env_get_cap("vbus");

  do
  {
    result = l4vbus_get_resource(vbus, dh, current, res);

    if (result)
      printf("Could not access resource of type %s.\n", resource_type(type));
    else
      printf("Resource %d: type %s, start=%lx, end=%lx\n", res->id,
        resource_type(res->type), res->start, res->end);

    current++;
  }
  while ((!result) && (res->type != type));

  return result;
}

static int gpio_get_irq(char const *hid, l4_uint32_t *start, l4_uint32_t *end)
{
  l4io_device_handle_t dh;
  l4io_resource_handle_t rh;
  l4io_resource_t res;
  int result;

  result = gpio_get_device(hid, &dh, &rh);

  if (result < 0)
    return result;

  result = gpio_get_resource(dh, &res, L4IO_RESOURCE_IRQ);

  if (result)
    return result;

  *start = res.start;
  *end = res.end;

  return result;
}

static int gpio_get_memory(char const *hid, l4_addr_t *start, l4_addr_t *end)
{
  l4io_device_handle_t dh;
  l4io_resource_handle_t rh;
  l4io_resource_t res;
  int result;

  result = gpio_get_device(hid, &dh, &rh);

  if (result < 0)
    return result;

  result = gpio_get_resource(dh, &res, L4IO_RESOURCE_MEM);

  if (result)
    return result;

  if ((result = l4io_request_iomem(res.start, res.end - res.start + 1,
                                  L4IO_MEM_NONCACHED, start)))
  {
    printf("Could not get address for '%s'.\n", hid);
    return result;
  }

  printf("Resource at 0x%lx...0x%lx.\n", res.start, res.end);

  *end = *start + (res.end - res.start + 1);

  return 0;
}

static long item_in_range(long start, long end, long index)
{
  if (start < end)
    return start + index;
  else
    return start - index;
}

int main(void)
{
  l4_addr_t gpio_base = 0, gpio_base_end = 0, port_d, port_d_end, port_f, port_f_end;
  l4_uint32_t gpio_irq_start = 0, gpio_irq_end = 0, gpio_irq;
  l4_cap_idx_t irqcap, icucap;
  l4_msgtag_t tag;
  long err;
  void *gpio_port_d, *gpio_port_f, *gpio_irq_pin;
  int result = 0;
  int led = 0;

  /* Obtain capabilities for the interrupt controller and an interrupt. */

  irqcap = l4re_util_cap_alloc();
  icucap = l4re_env_get_cap("icu");

  if (l4_is_invalid_cap(icucap))
  {
    printf("No 'icu' capability available in the virtual bus.\n");
    return 1;
  }

  if (l4_is_invalid_cap(irqcap))
  {
    printf("No capability available for the interrupt.\n");
    return 1;
  }

  /* Obtain resource details describing the interrupt and I/O memory. */

  printf("Access IRQ...\n");

  if ((result = gpio_get_irq("jz4780-gpio", &gpio_irq_start, &gpio_irq_end)) < 0)
    return 1;

  gpio_irq = item_in_range(gpio_irq_start, gpio_irq_end, 3);
  printf("IRQ range at %d...%d.\n", gpio_irq_start, gpio_irq_end);
  printf("PORTD IRQ at %d.\n", gpio_irq);

  printf("Access GPIO...\n");

  if ((result = gpio_get_memory("jz4780-gpio", &gpio_base, &gpio_base_end)) < 0)
    return 1;

  port_d = gpio_base + 0x300;
  port_d_end = port_d + 0x100;
  port_f = gpio_base + 0x500;
  port_f_end = port_f + 0x100;

  printf("GPIO at 0x%lx...0x%lx.\n", gpio_base, gpio_base_end);
  printf("PORTD at 0x%lx...0x%lx.\n", port_d, port_d_end);
  printf("PORTF at 0x%lx...0x%lx.\n", port_f, port_f_end);

  gpio_port_d = jz4780_gpio_init(port_d, port_d_end, 32, 0xffff4fff, 0x0000b000);
  gpio_port_f = jz4780_gpio_init(port_f, port_f_end, 32, 0xffa7f00f, 0x00580ff0);

  /* Create an interrupt object. */

  if ((err = l4_error(tag = l4_factory_create_irq(l4re_global_env->factory, irqcap))))
  {
    printf("Could not create IRQ object: %lx\n", err);
    return 1;
  }

  /* Bind the interrupt object to the IRQ number. */

  if ((err = l4_error(l4_icu_bind(icucap, gpio_irq, irqcap))))
  {
    printf("Could not bind IRQ %d to the ICU: %ld\n", gpio_irq, err);
    return 1;
  }

  /* Attach ourselves to the interrupt handler. */

  tag = l4_irq_attach(irqcap, 0xDEAD, l4re_env()->main_thread);

  if ((err = l4_error(tag)))
  {
    printf("Could not attach to IRQ %d: %ld\n", gpio_irq, err);
    return 1;
  }

  /* Set the GPIO pins up. */

  printf("Set up GPIO pins...\n");

  jz4780_gpio_setup(gpio_port_d, SW1, L4VBUS_GPIO_SETUP_IRQ, 1);
  jz4780_gpio_setup(gpio_port_f, LED, L4VBUS_GPIO_SETUP_OUTPUT, 0);

  gpio_irq_pin = jz4780_gpio_get_irq(gpio_port_d, SW1);
  jz4780_gpio_irq_set_mode(gpio_irq_pin, L4_IRQ_F_LEVEL_HIGH);

  printf("Press SW1 to invert LED.\n");

  while (1)
  {
    tag = l4_irq_receive(irqcap, L4_IPC_NEVER);

    if ((err = l4_ipc_error(tag, l4_utcb())))
    {
      printf("Error on IRQ receive: %ld\n", err);
      continue;
    }

    printf("SW1: %x\n", jz4780_gpio_get(gpio_port_d, SW1));

    if (jz4780_gpio_get(gpio_port_d, SW1))
    {
      printf("Invert LED\n");
      jz4780_gpio_set(gpio_port_f, LED, led);
      led = 1 - led;
    }
  }

  /* Detach from the interrupt. */

  tag = l4_irq_detach(irqcap);

  if ((err = l4_error(tag)))
    printf("Error detaching from IRQ: %ld\n", err);

  return 0;
}
