/*
 * Access an input device via a capability.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 *
 *
 * Font definitions and licence (see unifont.tff for bitmap data derived from
 * GNU Unifont's unifont.hex file):
 *
 * Copyright (C) 1998-2003 Roman Czyborra (http://czyborra.com/)
 *
 *   All glyphs are released under the GNU General Public License
 *   (GPL) version 2 or (at your option) a later version, with the
 *   GNU font embedding exception:
 *
 *        ** GPL v2.0 license with font embedding exception:
 *
 *        As a special exception, if you create a document which
 *        uses this font, and embed this font or unaltered portions
 *        of this font into the document, this font does not by
 *        itself cause the resulting document to be covered by
 *        the GNU General Public License. This exception does not
 *        however invalidate any other reasons why the document
 *        might be covered by the GNU General Public License.
 *        If you modify this font, you may extend this exception
 *        to your version of the font, but you are not obligated
 *        to do so. If you do not wish to do so, delete this
 *        exception statement from your version.
 */

#include <l4/devices/input-event-loop.h>

#include <pthread.h>
#include <pthread-l4.h>

#include <l4/re/c/util/video/goos_fb.h>
#include <l4/re/c/video/view.h>

#include <l4/re/env>
#include <l4/re/event>
#include <l4/re/event_enums.h>
#include <l4/re/util/cap_alloc>
#include <l4/sys/icu.h>
#include <l4/util/util.h>

#include <l4/mag-gfx/canvas>
#include <l4/mag-gfx/font>
#include <l4/mag-gfx/geometry>
#include <l4/mag-gfx/gfx_colors>
#include <l4/mag-gfx/mem_factory>

#include <stdint.h>
#include <string.h>

/* Video abstractions. */

static l4re_util_video_goos_fb_t gfb;
static l4re_video_view_info_t view_info;
static void *fb = 0;

/* Bundled font data. */

extern char const _binary_unifont_tff_start[];

/* Screen abstractions. */

using namespace Mag_gfx;

static Font *_font = 0;
static Canvas *_screen = 0;



/* Factories for certain pixel formats. */

static Mem::Factory<Rgb16> _rgb16;
static Mem::Factory<Rgb32> _rgb32;



/* Key to character conversion function. */

static const char *keys_to_strings[] = {

  0, "Escape", "1", "2", "3", "4", "5", "6", "7", "8",

  "9", "0", "-", "=", "Backspace", "Tab", "Q", "W", "E", "R",

  "T", "Y", "U", "I", "O", "P", "[", "]", "Enter", "Left Ctrl",

  "A", "S", "D", "F", "G", "H", "J", "K", "L", ";",

  "'", "`", "Left Shift", "\\", "Z", "X", "C", "V", "B", "N",

  "M", ",", ".", "/", "Right Shift", "Keypad *", "Left Alt", "Space",
    "Caps Lock", "F1",

  "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10", "Num Lock",

  "Scroll Lock", "Keypad 7", "Keypad 8", "Keypad 9", "Keypad -", "Keypad 4",
    "Keypad 5", "Keypad 6", "Keypad +", "Keypad 1",

  "Keypad 2", "Keypad 3", "Keypad 0", "Keypad .", 0, 0, "#102", "F11", "F12", 0,

  0, 0, 0, 0, 0, 0, "Keypad Enter", "Right Ctrl", "Keypad /", "SysRq",

  "Right Alt", "Line Feed", "Home", "Up", "Page Up", "Left", "Right", "End",
    "Down", "Page Down",

  "Insert", "Delete", "Macro", "Mute", "Volume Down", "Volume Up", "Power",
    "Keypad =", "Keypad +-", "Pause",
};

static const char null_string[] = "Unknown";

const int keys_to_strings_length = 120;

static const char *key_to_string(int key)
{
  return key < keys_to_strings_length ? keys_to_strings[key] : 0;
}



/* Show the keypad event status on the display. */

static uint8_t row = 0;
static uint32_t text_x = 0, text_y = 0, next_y = 0;

static void show_key_code(L4Re::Event_buffer::Event &event)
{
  uint32_t colsize = view_info.width / 10,
           rowsize = view_info.height / 20;
  uint8_t column;
  uint16_t mask;

  /* Convert the key code into a bit pattern. */

  for (column = 0, mask = (1 << 9); column < 10; column++, mask >>= 1)
    _screen->draw_box(Rect(Point(column * colsize, row * rowsize),
                           Area(colsize, rowsize)),
                      event.payload.code & mask ? (
                        event.payload.value ? Rgb32::Color(0, 255, 0)
                                            : Rgb32::Color(255, 0, 0))
                        : Rgb32::Color(0, 0, 0));

  /* Advance to the next row, wrapping around. */

  row = (row + 1) % 20;

  /* Refresh the display. */

  l4re_util_video_goos_fb_refresh(&gfb, 0, 0, view_info.width, view_info.height);
}

static void show_key_label(L4Re::Event_buffer::Event &event)
{
  if (event.payload.value)
  {
    const char *s = key_to_string(event.payload.code);
    Rgba32::Color col;

    if (!s)
    {
      s = null_string;
      col = Rgba32::Color(255, 0, 0, Rgba32::Color::Amax);
    }
    else
      col = Rgba32::Color(255, 255, 255, Rgba32::Color::Amax);

    Area box = _font->str_sz(s, strlen(s));

    /* Test for enough space horizontally. */

    if (text_x + box.w() > view_info.width)
    {
      text_x = 0;
      text_y = next_y;
      next_y = text_y + box.h();
    }

    /* Expand the line height, if appropriate. */

    else if (text_y + box.h() > next_y)
      next_y += box.h();

    /* Test for enough space vertically. */

    if (next_y > view_info.height)
    {
      text_x = 0;
      text_y = 0;
      next_y = box.h();

      /* Clear the screen. */

      _screen->draw_box(Rect(Point(0, 0), Area(view_info.width, view_info.height)), Rgb32::Color(0, 0, 0));
    }

    Point p(text_x, text_y);

    _screen->draw_box(Rect(p, box), Rgb32::Color(0, 0, 0));
    _screen->draw_string(p, _font, col, s, strlen(s));

    /* Move to the next position. */

    text_x += box.w();
  }

  /* Refresh the display. */

  l4re_util_video_goos_fb_refresh(&gfb, 0, 0, view_info.width, view_info.height);
}



/* Arguments: [ chars ] */

int main(int argc, char *argv[])
{
  if (l4re_util_video_goos_fb_setup_name(&gfb, "fb"))
    return 1;

  if (l4re_util_video_goos_fb_view_info(&gfb, &view_info))
    return 1;

  if (!(fb = l4re_util_video_goos_fb_attach_buffer(&gfb)))
    return 1;

  /* Obtain a canvas for the framebuffer. */

  Factory *factory;

  if (view_info.pixel_info.bytes_per_pixel == 2)
    factory = &_rgb16;
  else
    factory = &_rgb32;

  Canvas *screen = factory->create_canvas((void *) ((unsigned long) fb + view_info.buffer_offset),
      Area(view_info.width, view_info.height),
      view_info.bytes_per_line);

  Font font(&_binary_unifont_tff_start[0]);

  _screen = screen;
  _font = &font;

  /* Obtain a reference to the event source. It should also be possible to use
     the Input_event_interface for the core operations. */

  L4::Cap<L4Re::Event> event_server = L4Re::Env::env()->get_cap<L4Re::Event>("ev");
  if (!event_server.is_valid()) return 1;

  /* Obtain a capability for the event buffer. */

  L4::Cap<L4Re::Dataspace> mem = L4Re::Util::cap_alloc.alloc<L4Re::Dataspace>();
  if (!mem.is_valid()) return 1;
  if (event_server->get_buffer(mem)) return 1;

  /* Attach the event buffer to this task. */

  void *evmem = 0;

  if (L4Re::Env::env()->rm()->attach(&evmem, mem->size(), L4Re::Rm::Search_addr,
                                     L4::Ipc::make_cap_rw(mem)))
    return 1;

  L4Re::Event_buffer event_buffer = L4Re::Event_buffer(evmem, mem->size());

  /* Obtain an interrupt capability. */

  L4::Cap<L4::Irq> irq = L4Re::Util::cap_alloc.alloc<L4::Irq>();
  if (!irq.is_valid()) return 1;

  /* Create an interrupt object. */

  if (l4_error(L4Re::Env::env()->factory()->create(irq))) return 1;

  /* Bind the interrupt to the event capability. */

  if (l4_error(event_server->bind(0, irq))) return 1;

  /* Create an event handler. */

  Input_event_loop loop(event_buffer, irq);

  /* Attach the handler function. */

  if ((argc > 1) && (!strcmp(argv[1], "chars")))
    loop.attach(show_key_label);
  else
    loop.attach(show_key_code);

  /* Wait for events. */

  loop.start();
  l4_sleep_forever();

  return 0;
}
