/*
 * Display the keypad matrix using the specified dimensions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/keypad-client.h>

#include <l4/cxx/ipc_server>
#include <l4/re/dataspace>
#include <l4/re/env>
#include <l4/re/rm>
#include <l4/re/util/cap_alloc>

#include <l4/re/c/util/video/goos_fb.h>
#include <l4/re/c/video/view.h>

#include <l4/util/util.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>

/* Video abstractions. */

static l4re_util_video_goos_fb_t gfb;
static l4re_video_view_info_t fbi;
static void *fb;

/* Keypad status and dimensions. */

uint32_t *keypad = 0;
void *keymem = 0;
int columns, rows;



static uint32_t bitmask(uint32_t size)
{
  return (1 << size) - 1;
}

static uint32_t truncate_channel(uint32_t value, uint32_t size)
{
  return (value >> (8 - size)) & bitmask(size);
}

static void set_pixel(uint32_t pos, uint8_t bpp, uint32_t value)
{
  if (bpp <= 8) *(uint8_t *) pos = value;
  else if (bpp <= 16) *(uint16_t *) pos = value;
  else *(uint32_t *) pos = value;
}

/* Show the state of a key on the display. */

static void show_keystate(uint32_t x, uint32_t y, uint32_t w, uint32_t h, uint32_t rgb)
{
  uint8_t bpp = l4re_video_bits_per_pixel(&fbi.pixel_info);
  uint8_t bytes_per_pixel = fbi.pixel_info.bytes_per_pixel;
  uint32_t bytes_per_line = fbi.bytes_per_line;
  uint32_t pos;
  uint32_t col, row;

  rgb = (truncate_channel(rgb >> 16, fbi.pixel_info.r.size) << fbi.pixel_info.r.shift) |
        (truncate_channel(rgb >> 8, fbi.pixel_info.g.size) << fbi.pixel_info.g.shift) |
        (truncate_channel(rgb, fbi.pixel_info.b.size) << fbi.pixel_info.b.shift);

  for (row = 0; row < h; row++)
  {
    pos = (uint32_t) fb + (y + row) * bytes_per_line + x * bytes_per_pixel;

    for (col = 0; col < w; col++)
    {
      set_pixel(pos, bpp, rgb);
      pos += bytes_per_pixel;
    }
  }
}

/* Show the keypad status on the display. */

static void show_keypad(void)
{
  uint32_t colsize = fbi.width / columns,
           rowsize = fbi.height / rows;
  uint8_t column, row;
  uint32_t mask;

  for (column = 0; column < columns; column++)

    for (row = 0, mask = 1 << (rows - 1);
         row < rows;
         row++, mask >>= 1)

      show_keystate(column * colsize, row * rowsize, colsize, rowsize,
                    keypad[column] & mask ? 0xffffff : 0);

  /* Refresh the display. */

  l4re_util_video_goos_fb_refresh(&gfb, 0, 0, fbi.width, fbi.height);
}



int main(int argc, char *argv[])
{
  L4::Cap<Keypad_device_interface> keypad_server;
  L4::Cap<L4Re::Dataspace> mem;

  /* Obtain the keypad matrix dimensions. */

  if (argc < 3)
    return 1;

  columns = atoi(argv[1]);
  rows = atoi(argv[2]);

  if (l4re_util_video_goos_fb_setup_name(&gfb, "fb"))
    return 1;

  if (l4re_util_video_goos_fb_view_info(&gfb, &fbi))
    return 1;

  if (!(fb = l4re_util_video_goos_fb_attach_buffer(&gfb)))
    return 1;

  /* Obtain a reference to the keypad. */

  keypad_server = L4Re::Env::env()->get_cap<Keypad_device_interface>("keypad");
  if (!keypad_server.is_valid()) return 1;

  /* Obtain a capability for the keypad data. */

  mem = L4Re::Util::cap_alloc.alloc<L4Re::Dataspace>();
  if (!mem.is_valid()) return 1;

  /* Obtain a reference to the keypad data. */

  if (keypad_server->get_keypad_data(mem)) return 1;

  /* Attach the keypad data to a region in this task. */

  if (L4Re::Env::env()->rm()->attach(&keymem, mem->size(), L4Re::Rm::Search_addr,
                                     L4::Ipc::make_cap_rw(mem)))
    return 1;

  /* Show the keypad state. */

  keypad = (uint32_t *) keymem;

  while (1) show_keypad();

  return 0;
}
