/*
 * Access the keypad to modify the backlight on the Letux 400.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/backlight-client.h>
#include <l4/devices/input-keypad-client.h>

#include <l4/re/env>
#include <l4/sys/capability>

#include <l4/re/event_enums.h>
#include <l4/util/util.h>

/* Backlight level. */

enum Letux400_backlight_levels
{
  Letux400_backlight_min_level = 0,
  Letux400_backlight_max_level = 300,
  Letux400_backlight_step = 25,
};

static int backlight_level = 250;

/* Key state. */

static int modifier_set = 0;

/* Backlight device abstractions. */

static L4::Cap<Backlight_device_interface> backlight_device;



/* Input event handler. */

static void handler(Input_event event, void *priv)
{
  (void) priv;

  /* Track the state of the modifier key. */

  if (event.code == L4RE_KEY_SLEEP)
    modifier_set = event.value;

  if (!event.value || !modifier_set)
    return;

  /* Upon keypress events, test controls and update the backlight. */

  switch (event.code)
  {
    case L4RE_KEY_DOWN:
      if (backlight_level < Letux400_backlight_min_level + Letux400_backlight_step)
        backlight_level = Letux400_backlight_min_level;
      else
        backlight_level -= Letux400_backlight_step;
      break;

    case L4RE_KEY_UP:
      if (backlight_level > Letux400_backlight_max_level - Letux400_backlight_step)
        backlight_level = Letux400_backlight_max_level;
      else
        backlight_level += Letux400_backlight_step;
      break;

    default: return;
  }

  /* Use the backlight device to update the backlight level. */

  backlight_device->set_brightness(backlight_level);
}



int main(void)
{
  Keypad *keypad = Keypad::get_keypad();
  Input_keypad_client client(keypad);

  /* Obtain a reference to the PWM device. */

  backlight_device = L4Re::Env::env()->get_cap<Backlight_device_interface>("backlight");
  if (!backlight_device.is_valid()) return 1;

  client.attach(handler, 0);

  l4_sleep_forever();

  return 0;
}
