/*
 * Access the keypad GPIOs on the Letux 400 notebook computer.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4730.h>
#include <l4/re/c/util/video/goos_fb.h>
#include <l4/re/c/video/view.h>
#include <l4/sys/cache.h>
#include <l4/util/util.h>

#include <stdio.h>
#include <stdint.h>
#include <unistd.h>

#include "memory.h"

enum Jz4730_keypad_gpio
{
  Jz4730_keypad_gpio_inputs_count = 8,
  Jz4730_keypad_gpio_outputs_count = 17,
};

/* Port A input pins. */

const uint8_t Jz4730_keypad_inputs[Jz4730_keypad_gpio_inputs_count] = {
  0, 1, 2, 3, 4, 5, 6, 7
};

const Pin_slice Jz4730_keypad_inputs_mask = {.mask=0x000000ff, .offset=0};

/* Port D output pins. */

const uint8_t Jz4730_keypad_outputs[Jz4730_keypad_gpio_outputs_count] = {
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 29
};

const Pin_slice Jz4730_keypad_outputs_mask = {.mask=0x2000ffff, .offset=0};

/* Video abstractions. */

static l4re_util_video_goos_fb_t gfb;
static l4re_video_view_info_t fbi;
static void *fb;

/* Peripheral memory regions. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;

/* GPIO abstractions. */

static void *gpio_port_a, *gpio_port_d;

/* Keypad status. */

uint32_t keypad[Jz4730_keypad_gpio_outputs_count];



/* Initialise the pins for scanning the keypad. */

static void init_keyscan(void)
{
  jz4730_gpio_multi_setup(gpio_port_a, &Jz4730_keypad_inputs_mask, Fix_input, 0);
  jz4730_gpio_multi_config_pull(gpio_port_a, &Jz4730_keypad_inputs_mask, Pull_up);
  jz4730_gpio_multi_setup(gpio_port_d, &Jz4730_keypad_outputs_mask, Fix_input, 0);
}

/*
Scan the keypad by enabling each output column and inspecting each input row.
Store each column bitmap in the keypad array.
*/

static void scan_keypad(void)
{
  uint8_t column, row;
  uint32_t value;

  for (column = 0; column < Jz4730_keypad_gpio_outputs_count; column++)
  {
    jz4730_gpio_setup(gpio_port_d, Jz4730_keypad_outputs[column], Fix_output, 0);
    l4_sleep(1);

    value = 0;

    for (row = 0; row < Jz4730_keypad_gpio_inputs_count; row++)
      value = (value << 1) | (jz4730_gpio_get(gpio_port_a, Jz4730_keypad_inputs[row]) ? 0 : 1);

    keypad[column] = value;

    jz4730_gpio_setup(gpio_port_d, Jz4730_keypad_outputs[column], Fix_input, 0);
  }
}

static uint32_t bitmask(uint32_t size)
{
  return (1 << size) - 1;
}

static uint32_t truncate_channel(uint32_t value, uint32_t size)
{
  return (value >> (8 - size)) & bitmask(size);
}

static void set_pixel(uint32_t pos, uint8_t bpp, uint32_t value)
{
  if (bpp <= 8) *(uint8_t *) pos = value;
  else if (bpp <= 16) *(uint16_t *) pos = value;
  else *(uint32_t *) pos = value;
}

/* Show the state of a key on the display. */

static void show_keystate(uint32_t x, uint32_t y, uint32_t w, uint32_t h, uint32_t rgb)
{
  uint8_t bpp = l4re_video_bits_per_pixel(&fbi.pixel_info);
  uint8_t bytes_per_pixel = fbi.pixel_info.bytes_per_pixel;
  uint32_t bytes_per_line = fbi.bytes_per_line;
  uint32_t pos;
  uint32_t col, row;

  rgb = (truncate_channel(rgb >> 16, fbi.pixel_info.r.size) << fbi.pixel_info.r.shift) |
        (truncate_channel(rgb >> 8, fbi.pixel_info.g.size) << fbi.pixel_info.g.shift) |
        (truncate_channel(rgb, fbi.pixel_info.b.size) << fbi.pixel_info.b.shift);

  for (row = 0; row < h; row++)
  {
    pos = (uint32_t) fb + (y + row) * bytes_per_line + x * bytes_per_pixel;

    for (col = 0; col < w; col++)
    {
      set_pixel(pos, bpp, rgb);
      pos += bytes_per_pixel;
    }
  }
}

/* Show the keypad status on the display. */

static void show_keypad(void)
{
  uint32_t colsize = fbi.width / Jz4730_keypad_gpio_outputs_count,
           rowsize = fbi.height / Jz4730_keypad_gpio_inputs_count;
  uint8_t column, row;
  uint32_t mask;

  for (column = 0; column < Jz4730_keypad_gpio_outputs_count; column++)

    for (row = 0, mask = 1 << (Jz4730_keypad_gpio_inputs_count - 1);
         row < Jz4730_keypad_gpio_inputs_count;
         row++, mask >>= 1)

      show_keystate(column * colsize, row * rowsize, colsize, rowsize,
                    keypad[column] & mask ? 0xffffff : 0);

  /* Refresh the display. */

  l4re_util_video_goos_fb_refresh(&gfb, 0, 0, fbi.width, fbi.height);
}

int main(void)
{
  int result;

  if ((result = get_memory("jz4730-gpio", &gpio_virt_base, &gpio_virt_base_end)) < 0)
    return 1;

  if (l4re_util_video_goos_fb_setup_name(&gfb, "fb"))
    return 1;

  if (l4re_util_video_goos_fb_view_info(&gfb, &fbi))
    return 1;

  if (!(fb = l4re_util_video_goos_fb_attach_buffer(&gfb)))
    return 1;

  gpio_port_a = jz4730_gpio_init(gpio_virt_base, gpio_virt_base + 0x30, 32);
  gpio_port_d = jz4730_gpio_init(gpio_virt_base + 0x90, gpio_virt_base + 0xc0, 32);

  init_keyscan();

  while (1)
  {
    scan_keypad();
    show_keypad();
  }

  return 0;
}
