/*
 * Memory allocation utility functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 * Subject to other copyrights, being derived from the existing L4Re
 * LCD driver implementations.
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */

#include <stdio.h>

#include "memory.h"
#include <l4/io/io.h>
#include <l4/re/env.h>
#include <l4/re/c/mem_alloc.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/util/util.h>
#include <l4/vbus/vbus.h>

/* Internal memory allocation/mapping functions. */

static char const *resource_type(enum l4io_resource_types_t type)
{ 
  switch (type)
  { 
    case L4VBUS_RESOURCE_INVALID:
    return "INVALID";
    
    case L4VBUS_RESOURCE_IRQ:
    return "IRQ";
    
    case L4VBUS_RESOURCE_MEM:
    return "MEMORY";
    
    default:
    return "OTHER";
  }
}

int get_device(char const *hid, l4io_device_handle_t *dh, l4io_resource_handle_t *rh)
{
  int result = l4io_lookup_device(hid, dh, 0, rh);

  if (result < 0)
    printf("Could not access '%s': %s\n", hid, result == -L4_ENOENT ? "no such device" : "no device");

  return result;
}

int get_resource(l4io_device_handle_t dh, l4io_resource_t *res,
                             enum l4io_resource_types_t type)
{
  int current = 0, result = 0;
  l4_cap_idx_t vbus = l4re_env_get_cap("vbus");

  do
  {
    result = l4vbus_get_resource(vbus, dh, current, res);

    if (result)
      printf("Could not access resource of type %s.\n", resource_type(type));
    else
      printf("Resource %d: type %s, start=%lx, end=%lx\n", res->id,
        resource_type(res->type), res->start, res->end);

    current++;
  }
  while ((!result) && (res->type != type));

  return result;
}

int get_memory(char const *hid, l4_addr_t *start, l4_addr_t *end)
{
  l4io_device_handle_t dh;
  l4io_resource_handle_t rh;
  l4io_resource_t res;
  int result;

  result = get_device(hid, &dh, &rh);

  if (result < 0)
    return result;

  result = get_resource(dh, &res, L4IO_RESOURCE_MEM);

  if (result)
    return result;

  if ((result = l4io_request_iomem(res.start, res.end - res.start + 1,
                                  L4IO_MEM_NONCACHED, start)))
  {
    printf("Could not get address for '%s'.\n", hid);
    return result;
  }

  printf("Resource at 0x%lx...0x%lx.\n", res.start, res.end);

  *end = *start + (res.end - res.start + 1);

  return 0;
}
