/*
 * Access the LED and PWM GPIOs on the Letux 400 notebook computer.
 * This example shows how to use the following GPIOs:
 *
 * PA27 - Caps Lock
 * PC22 - Num Lock
 * PC30 - PWM backlight
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4730.h>
#include <l4/devices/pwm-jz4730.h>

#include <l4/io/io.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/factory.h>
#include <l4/sys/ipc.h>
#include <l4/util/util.h>
#include <l4/vbus/vbus.h>

#include <stdio.h>
#include <stdint.h>
#include <unistd.h>

enum {
  CAPS = 27, // via PORTA
  NUM = 22, // via PORTC
  PWM = 30, // via PORTC
};



/* Device and resource discovery. */

static char const *resource_type(enum l4io_resource_types_t type)
{
  switch (type)
  {
    case L4VBUS_RESOURCE_INVALID:
    return "INVALID";

    case L4VBUS_RESOURCE_IRQ:
    return "IRQ";

    case L4VBUS_RESOURCE_MEM:
    return "MEMORY";

    default:
    return "OTHER";
  }
}

static int get_device(char const *hid, l4io_device_handle_t *dh, l4io_resource_handle_t *rh)
{
  int result = l4io_lookup_device(hid, dh, 0, rh);

  if (result < 0)
    printf("Could not access '%s': %s\n", hid, result == -L4_ENOENT ? "no such device" : "no device");

  return result;
}

static int get_resource(l4io_device_handle_t dh, l4io_resource_t *res,
                             enum l4io_resource_types_t type)
{
  int current = 0, result = 0;
  l4_cap_idx_t vbus = l4re_env_get_cap("vbus");

  do
  {
    result = l4vbus_get_resource(vbus, dh, current, res);

    if (result)
      printf("Could not access resource of type %s.\n", resource_type(type));
    else
      printf("Resource %d: type %s, start=%lx, end=%lx\n", res->id,
        resource_type(res->type), res->start, res->end);

    current++;
  }
  while ((!result) && (res->type != type));

  return result;
}

static int get_memory(char const *hid, l4_addr_t *start, l4_addr_t *end)
{
  l4io_device_handle_t dh;
  l4io_resource_handle_t rh;
  l4io_resource_t res;
  int result;

  result = get_device(hid, &dh, &rh);

  if (result < 0)
    return result;

  result = get_resource(dh, &res, L4IO_RESOURCE_MEM);

  if (result)
    return result;

  if ((result = l4io_request_iomem(res.start, res.end - res.start + 1,
                                  L4IO_MEM_NONCACHED, start)))
  {
    printf("Could not get address for '%s'.\n", hid);
    return result;
  }

  printf("Resource at 0x%lx...0x%lx.\n", res.start, res.end);

  *end = *start + (res.end - res.start + 1);

  return 0;
}

int main(void)
{
  l4_addr_t gpio_base = 0, gpio_base_end = 0, port_a, port_a_end, port_c, port_c_end;
  l4_addr_t pwm_base = 0, pwm_base_end = 0;
  void *gpio_port_a, *gpio_port_c, *pwm0_device;
  int result = 0;
  int caps = 0, num = 1, pwm = 300, pwmdir = -50;

  /* Obtain resource details describing the I/O memory. */

  printf("Access GPIO...\n");

  if ((result = get_memory("jz4730-gpio", &gpio_base, &gpio_base_end)) < 0)
    return 1;

  port_a = gpio_base;
  port_a_end = port_a + 0x30;
  port_c = gpio_base + 0x60;
  port_c_end = port_c + 0x30;

  printf("GPIO at 0x%lx...0x%lx.\n", gpio_base, gpio_base_end);
  printf("PORTA at 0x%lx...0x%lx.\n", port_a, port_a_end);
  printf("PORTC at 0x%lx...0x%lx.\n", port_c, port_c_end);

  gpio_port_a = jz4730_gpio_init(port_a, port_a_end, 32);
  gpio_port_c = jz4730_gpio_init(port_c, port_c_end, 32);

  printf("Access PWM...\n");

  if ((result = get_memory("jz4730-pwm", &pwm_base, &pwm_base_end)) < 0)
    return 1;

  pwm0_device = jz4730_pwm_init(pwm_base, pwm_base + 0x1000);

  /* Set the GPIO pins up. */

  printf("Set up GPIO pins...\n");

  jz4730_gpio_setup(gpio_port_a, CAPS, Fix_output, caps);
  jz4730_gpio_setup(gpio_port_c, NUM, Fix_output, num);
  jz4730_gpio_config_pad(gpio_port_c, PWM, Function_alt, 1);

  /* Set the PWM device up. */

  printf("Set up PWM...\n");

  jz4730_pwm_set_duty(pwm0_device, pwm);
  jz4730_pwm_set_period(pwm0_device, 299);
  jz4730_pwm_set_control(pwm0_device, 0x80 | 0x3f);

  while (1)
  {
    caps = 1 - caps;
    jz4730_gpio_set(gpio_port_a, CAPS, caps);
    num = 1 - num;
    jz4730_gpio_set(gpio_port_c, NUM, num);
    pwm += pwmdir;
    jz4730_pwm_set_duty(pwm0_device, pwm);
    if (pwm == 0) pwmdir = 50; else if (pwm == 300) pwmdir = -50;
    l4_sleep(1000); // 1000ms
  }

  return 0;
}
