/*
 * Access the keypad to modify the backlight on the Ben NanoNote.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/backlight-client.h>
#include <l4/devices/input-keypad-client.h>

#include <l4/re/env>
#include <l4/sys/capability>

#include <l4/re/event_enums.h>
#include <l4/util/util.h>

/* Backlight level. */

enum Qi_lb60_backlight_levels
{
  Qi_lb60_backlight_min_level = 0,
  Qi_lb60_backlight_max_level = 255,
  Qi_lb60_backlight_step = 10,
};

static int backlight_level = 64;

/* Key state. */

static int modifier_set = 0;
static int enable_modifier_set = 0;

/* Backlight device abstractions. */

static L4::Cap<Backlight_device_interface> backlight_device;



/* Input event handler. */

static void handler(Input_event event, void *priv)
{
  (void) priv;

  /* Track the state of the modifier keys. */

  if (event.code == L4RE_KEY_FN)
    modifier_set = event.value;
  else if (event.code == L4RE_KEY_RIGHTALT)
    enable_modifier_set = event.value;

  if (!event.value || !(modifier_set || enable_modifier_set))
    return;

  /* Upon keypress events, test controls and update the backlight. */

  switch (event.code)
  {
    case L4RE_KEY_VOLUMEDOWN:
      if (enable_modifier_set)
      {
        backlight_device->disable();
        return;
      }
      else if (backlight_level < Qi_lb60_backlight_min_level + Qi_lb60_backlight_step)
        backlight_level = Qi_lb60_backlight_min_level;
      else
        backlight_level -= Qi_lb60_backlight_step;
      break;

    case L4RE_KEY_VOLUMEUP:
      if (enable_modifier_set)
      {
        backlight_device->enable();
        return;
      }
      else if (backlight_level > Qi_lb60_backlight_max_level - Qi_lb60_backlight_step)
        backlight_level = Qi_lb60_backlight_max_level;
      else
        backlight_level += Qi_lb60_backlight_step;
      break;

    default: return;
  }

  /* Use the backlight device to update the backlight level. */

  backlight_device->set_brightness(backlight_level);
}



int main(void)
{
  Keypad *keypad = Keypad::get_keypad();
  Input_keypad_client client(keypad);

  /* Obtain a reference to the backlight device. */

  backlight_device = L4Re::Env::env()->get_cap<Backlight_device_interface>("backlight");
  if (!backlight_device.is_valid()) return 1;

  client.attach(handler, 0);

  l4_sleep_forever();

  return 0;
}
