/*
 * Display the physical keypad matrix layout for the Ben NanoNote.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/rm.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/re/env.h>
#include <l4/util/util.h>
#include <l4/sys/ipc.h>

#include <l4/re/c/util/video/goos_fb.h>
#include <l4/re/c/video/view.h>

#include <l4/util/util.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>

enum Jz4740_keypad_gpio
{
  Jz4740_keypad_gpio_inputs_count = 8,
  Jz4740_keypad_gpio_outputs_count = 8,
};

/* Video abstractions. */

static l4re_util_video_goos_fb_t gfb;
static l4re_video_view_info_t fbi;
static void *fb;

/* Keypad status and dimensions. */

uint32_t *keypad = 0;
void *keymem = 0;
int columns = Jz4740_keypad_gpio_outputs_count, rows = Jz4740_keypad_gpio_inputs_count;

/* Position units. */

enum {
  KEY = 12, DIRKEY = 8,
  WIDTH = 120, HEIGHT = 72, ROWS = 6,
};

/*
        F1  F2  F3  F4  F5  F6  F7  F8
Q   W   E   R   T   Y   U   I   O   P
A   S   D   F   G   H   J   K   L   Bsp
Esc Z   X   C   V   B   N   M   =   Enter
Tab Cap \   '   ,   .   /      Up   V-Up
LSh LAl Fn  Sym Spc Qi Ctr Lt Dn Rt V-Dn
*/

enum {
  SIZE_KEY_F1 = KEY, SIZE_KEY_F2 = KEY, SIZE_KEY_F3 = KEY, SIZE_KEY_F4 = KEY,
  SIZE_KEY_F5 = KEY, SIZE_KEY_F6 = KEY, SIZE_KEY_F7 = KEY, SIZE_KEY_F8 = KEY,

  SIZE_KEY_Q = KEY, SIZE_KEY_W = KEY, SIZE_KEY_E = KEY, SIZE_KEY_R = KEY,
  SIZE_KEY_T = KEY, SIZE_KEY_Y = KEY, SIZE_KEY_U = KEY, SIZE_KEY_I = KEY,
  SIZE_KEY_O = KEY, SIZE_KEY_P = KEY,

  SIZE_KEY_A = KEY, SIZE_KEY_S = KEY, SIZE_KEY_D = KEY, SIZE_KEY_F = KEY,
  SIZE_KEY_G = KEY, SIZE_KEY_H = KEY, SIZE_KEY_J = KEY, SIZE_KEY_K = KEY,
  SIZE_KEY_L = KEY, SIZE_KEY_BACKSPACE = KEY,

  SIZE_KEY_ESCAPE = KEY, SIZE_KEY_Z = KEY, SIZE_KEY_X = KEY, SIZE_KEY_C = KEY,
  SIZE_KEY_V = KEY, SIZE_KEY_B = KEY, SIZE_KEY_N = KEY, SIZE_KEY_M = KEY,
  SIZE_KEY_EQUAL = KEY, SIZE_KEY_ENTER = KEY,

  SIZE_KEY_TAB = KEY, SIZE_KEY_DOWNSHIFT = KEY, SIZE_KEY_BACKSLASH = KEY, SIZE_KEY_APOSTROPHE = KEY,
  SIZE_KEY_COMMA = KEY, SIZE_KEY_DOT = KEY, SIZE_KEY_SLASH = KEY,
  SIZE_KEY_UP = DIRKEY, SIZE_KEY_VOLUMEUP = KEY,

  SIZE_KEY_LEFTSHIFT = KEY, SIZE_KEY_LEFTALT = KEY, SIZE_KEY_FN = KEY, SIZE_KEY_SYM = KEY,
  SIZE_KEY_SPACE = KEY, SIZE_KEY_QI = KEY, SIZE_KEY_RIGHTCTRL = KEY, SIZE_KEY_LEFT = DIRKEY,
  SIZE_KEY_DOWN = DIRKEY, SIZE_KEY_RIGHT = DIRKEY, SIZE_KEY_VOLUMEDOWN = KEY,
};

enum {
  XPOS_KEY_F1 = 2*KEY, XPOS_KEY_F2 = 3*KEY, XPOS_KEY_F3 = 4*KEY, XPOS_KEY_F4 = 5*KEY,
  XPOS_KEY_F5 = 6*KEY, XPOS_KEY_F6 = 7*KEY, XPOS_KEY_F7 = 8*KEY, XPOS_KEY_F8 = 9*KEY,

  XPOS_KEY_Q = 0, XPOS_KEY_W = KEY, XPOS_KEY_E = 2*KEY, XPOS_KEY_R = 3*KEY,
  XPOS_KEY_T = 4*KEY, XPOS_KEY_Y = 5*KEY, XPOS_KEY_U = 6*KEY, XPOS_KEY_I = 7*KEY,
  XPOS_KEY_O = 8*KEY, XPOS_KEY_P = 9*KEY,

  XPOS_KEY_A = 0, XPOS_KEY_S = KEY, XPOS_KEY_D = 2*KEY, XPOS_KEY_F = 3*KEY,
  XPOS_KEY_G = 4*KEY, XPOS_KEY_H = 5*KEY, XPOS_KEY_J = 6*KEY, XPOS_KEY_K = 7*KEY,
  XPOS_KEY_L = 8*KEY, XPOS_KEY_BACKSPACE = 9*KEY,

  XPOS_KEY_ESCAPE = 0, XPOS_KEY_Z = KEY, XPOS_KEY_X = 2*KEY, XPOS_KEY_C = 3*KEY,
  XPOS_KEY_V = 4*KEY, XPOS_KEY_B = 5*KEY, XPOS_KEY_N = 6*KEY, XPOS_KEY_M = 7*KEY,
  XPOS_KEY_EQUAL = 8*KEY, XPOS_KEY_ENTER = 9*KEY,

  XPOS_KEY_TAB = 0, XPOS_KEY_DOWNSHIFT = KEY, XPOS_KEY_BACKSLASH = 2*KEY, XPOS_KEY_APOSTROPHE = 3*KEY,
  XPOS_KEY_COMMA = 4*KEY, XPOS_KEY_DOT = 5*KEY, XPOS_KEY_SLASH = 6*KEY,
  XPOS_KEY_UP = 7*KEY+DIRKEY, XPOS_KEY_VOLUMEUP = 9*KEY,

  XPOS_KEY_LEFTSHIFT = 0, XPOS_KEY_LEFTALT = KEY, XPOS_KEY_FN = 2*KEY, XPOS_KEY_SYM = 3*KEY,
  XPOS_KEY_SPACE = 4*KEY, XPOS_KEY_QI = 5*KEY, XPOS_KEY_RIGHTCTRL = 6*KEY, XPOS_KEY_LEFT = 7*KEY,
  XPOS_KEY_DOWN = 7*KEY+DIRKEY, XPOS_KEY_RIGHT = 7*KEY+2*DIRKEY, XPOS_KEY_VOLUMEDOWN = 9*KEY,
};

enum {
  YPOS_KEY_F1 = 0, YPOS_KEY_F2 = 0, YPOS_KEY_F3 = 0, YPOS_KEY_F4 = 0,
  YPOS_KEY_F5 = 0, YPOS_KEY_F6 = 0, YPOS_KEY_F7 = 0, YPOS_KEY_F8 = 0,

  YPOS_KEY_Q = 1, YPOS_KEY_W = 1, YPOS_KEY_E = 1, YPOS_KEY_R = 1,
  YPOS_KEY_T = 1, YPOS_KEY_Y = 1, YPOS_KEY_U = 1, YPOS_KEY_I = 1,
  YPOS_KEY_O = 1, YPOS_KEY_P = 1,

  YPOS_KEY_A = 2, YPOS_KEY_S = 2, YPOS_KEY_D = 2, YPOS_KEY_F = 2,
  YPOS_KEY_G = 2, YPOS_KEY_H = 2, YPOS_KEY_J = 2, YPOS_KEY_K = 2,
  YPOS_KEY_L = 2, YPOS_KEY_BACKSPACE = 2,

  YPOS_KEY_ESCAPE = 3, YPOS_KEY_Z = 3, YPOS_KEY_X = 3, YPOS_KEY_C = 3,
  YPOS_KEY_V = 3, YPOS_KEY_B = 3, YPOS_KEY_N = 3, YPOS_KEY_M = 3,
  YPOS_KEY_EQUAL = 3, YPOS_KEY_ENTER = 3,

  YPOS_KEY_TAB = 4, YPOS_KEY_DOWNSHIFT = 4, YPOS_KEY_BACKSLASH = 4, YPOS_KEY_APOSTROPHE = 4,
  YPOS_KEY_COMMA = 4, YPOS_KEY_DOT = 4, YPOS_KEY_SLASH = 4,
  YPOS_KEY_UP = 4, YPOS_KEY_VOLUMEUP = 4,

  YPOS_KEY_LEFTSHIFT = 5, YPOS_KEY_LEFTALT = 5, YPOS_KEY_FN = 5, YPOS_KEY_SYM = 5,
  YPOS_KEY_SPACE = 5, YPOS_KEY_QI = 5, YPOS_KEY_RIGHTCTRL = 5, YPOS_KEY_LEFT = 5,
  YPOS_KEY_DOWN = 5, YPOS_KEY_RIGHT = 5, YPOS_KEY_VOLUMEDOWN = 5,
};

/* Keypad matrix mapping. */

#define PHYS_KEY(X) {XPOS_KEY_##X, YPOS_KEY_##X, SIZE_KEY_##X}
#define PHYS_NULL   {0, 0, 0}

static const uint32_t keypos[Jz4740_keypad_gpio_inputs_count][Jz4740_keypad_gpio_outputs_count][3] = {

{PHYS_KEY(F1), PHYS_KEY(F2), PHYS_KEY(F3), PHYS_KEY(F4), PHYS_KEY(F5), PHYS_KEY(F6),
 PHYS_KEY(F7), PHYS_NULL},

{PHYS_KEY(Q), PHYS_KEY(W), PHYS_KEY(E), PHYS_KEY(R), PHYS_KEY(T), PHYS_KEY(Y),
 PHYS_KEY(U), PHYS_KEY(I)},

{PHYS_KEY(A), PHYS_KEY(S), PHYS_KEY(D), PHYS_KEY(F), PHYS_KEY(G), PHYS_KEY(H),
 PHYS_KEY(J), PHYS_KEY(K)},

{PHYS_KEY(ESCAPE), PHYS_KEY(Z), PHYS_KEY(X), PHYS_KEY(C), PHYS_KEY(V), PHYS_KEY(B),
 PHYS_KEY(N), PHYS_KEY(M)},

{PHYS_KEY(TAB), PHYS_KEY(DOWNSHIFT) /* CAPSLOCK */, PHYS_KEY(BACKSLASH),
 PHYS_KEY(APOSTROPHE), PHYS_KEY(COMMA), PHYS_KEY(DOT), PHYS_KEY(SLASH), PHYS_KEY(UP)},

{PHYS_KEY(O), PHYS_KEY(L), PHYS_KEY(EQUAL), PHYS_KEY(SYM) /* RIGHTALT */,
 PHYS_KEY(SPACE), PHYS_KEY(QI) /* F13 */, PHYS_KEY(RIGHTCTRL), PHYS_KEY(LEFT)},

{PHYS_KEY(F8), PHYS_KEY(P), PHYS_KEY(BACKSPACE), PHYS_KEY(ENTER), PHYS_KEY(VOLUMEUP),
 PHYS_KEY(VOLUMEDOWN), PHYS_KEY(DOWN), PHYS_KEY(RIGHT)},

{PHYS_KEY(LEFTSHIFT), PHYS_KEY(LEFTALT), PHYS_KEY(FN), PHYS_NULL, PHYS_NULL,
 PHYS_NULL, PHYS_NULL, PHYS_NULL}
};



static uint32_t bitmask(uint32_t size)
{
  return (1 << size) - 1;
}

static uint32_t truncate_channel(uint32_t value, uint32_t size)
{
  return (value >> (8 - size)) & bitmask(size);
}

static void set_pixel(uint32_t pos, uint8_t bpp, uint32_t value)
{
  if (bpp <= 8) *(uint8_t *) pos = value;
  else if (bpp <= 16) *(uint16_t *) pos = value;
  else *(uint32_t *) pos = value;
}

/* Show the state of a key on the display. */

static void fill_rectangle(uint32_t x, uint32_t y, uint32_t w, uint32_t h, uint32_t rgb)
{
  uint8_t bpp = l4re_video_bits_per_pixel(&fbi.pixel_info);
  uint8_t bytes_per_pixel = fbi.pixel_info.bytes_per_pixel;
  uint32_t bytes_per_line = fbi.bytes_per_line;
  uint32_t pos;
  uint32_t col, row;

  rgb = (truncate_channel(rgb >> 16, fbi.pixel_info.r.size) << fbi.pixel_info.r.shift) |
        (truncate_channel(rgb >> 8, fbi.pixel_info.g.size) << fbi.pixel_info.g.shift) |
        (truncate_channel(rgb, fbi.pixel_info.b.size) << fbi.pixel_info.b.shift);

  for (row = 0; row < h; row++)
  {
    pos = (uint32_t) fb + (y + row) * bytes_per_line + x * bytes_per_pixel;

    for (col = 0; col < w; col++)
    {
      set_pixel(pos, bpp, rgb);
      pos += bytes_per_pixel;
    }
  }
}

/* Show the keypad status on the display. */

static void show_keypad(void)
{
  uint32_t rowsize = fbi.height / ROWS, colsize;
  uint8_t column, row;
  uint32_t mask;
  const uint32_t *pos;

  for (column = 0; column < columns; column++)
 
    for (row = 0, mask = 1 << (rows - 1);
         row < rows;
         row++, mask >>= 1)
    {
      /* Obtain the physical position. */

      pos = keypos[row][column];

      /* Obtain the width of the key. */

      colsize = (pos[2] * fbi.width) / WIDTH;

      /* Plot the rectangle for the key. */

      fill_rectangle((pos[0] * fbi.width) / WIDTH, pos[1] * rowsize, colsize, rowsize,
                     keypad[column] & mask ? 0xffffff : 0);
    }

  /* Refresh the display. */

  l4re_util_video_goos_fb_refresh(&gfb, 0, 0, fbi.width, fbi.height);
}



int main(void)
{
  l4_cap_idx_t keypad_server;
  l4re_ds_t mem;
  l4_msgtag_t tag;

  if (l4re_util_video_goos_fb_setup_name(&gfb, "fb"))
    return 1;

  if (l4re_util_video_goos_fb_view_info(&gfb, &fbi))
    return 1;

  if (!(fb = l4re_util_video_goos_fb_attach_buffer(&gfb)))
    return 1;

  /* Obtain a reference to the keypad. */

  keypad_server = l4re_env_get_cap("keypad");
  if (l4_is_invalid_cap(keypad_server)) return 1;

  /* Obtain a capability for the keypad data. */

  mem = l4re_util_cap_alloc();
  if (l4_is_invalid_cap(mem)) return 1;

  /* Obtain a reference to the keypad data. */

  l4_utcb_br()->bdr = 0;
  l4_utcb_br()->br[0] = L4_RCV_ITEM_SINGLE_CAP | mem;

  tag = l4_ipc_call(keypad_server, l4_utcb(),
                    l4_msgtag(0, 0, 0, 0), /* label zero, zero words, zero *sent* items */
                    L4_IPC_NEVER);

  if (l4_ipc_error(tag, l4_utcb())) return 1;

  /* Attach the keypad data to a region in this task. */

  if (l4re_rm_attach(&keymem, l4re_ds_size(mem), L4RE_RM_SEARCH_ADDR, mem, 0,
                     L4_PAGESHIFT))
    return 1;

  /* Show the keypad state. */

  keypad = (uint32_t *) keymem;

  while (1) show_keypad();

  return 0;
}
